!*******************************************************************
!*                         penEasy                                 *
!*                                                                 *
!* Short description:                                              *
!*   General-purpose main program for the PENELOPE system.         *
!*   Please refer to the manual for detailed instructions.         *
!*                                                                 *
!*   For a list of dependencies from PENELOPE, see                 *
!*   ~/documentation/dependencies.txt                              *
!*                                                                 *
!* Version:                                                        *
!*   See main program output, written below.                       *
!*                                                                 *
!* Josep Sempau                                                    *
!* email: josep.sempau@upc.es                                      *
!* Universitat Politecnica de Catalunya, Barcelona, Spain          *
!* SEE THE MANUAL FOR COPYRIGHT NOTICE                             *
!*******************************************************************

! PENELOPE library:
#include "penelope.f"
#include "pengeom.f"
#include "rita.f"
#include "penvared.f"
! Protons. Next lines will be compiled only if compiler is invoked with option -DPROTONS:
#ifdef PROTONS
#include "penh.f"
#endif

! penEasy library:
#include "penvr.F"
#include "penaux.F"
#include "penvox.F"
#include "timing.F"

! Source models:
#include "sourcePennuc.F"
#include "sourceBoxIsotropicGaussSpectrum.F"
#include "sourcePhaseSpaceFile.F"

! Tallies:
#include "tallyCylindricalDoseDistrib.F"
#include "tallyEnergyDeposition.F"
#include "tallyFluenceTrackLength.F"
#include "tallyParticleCurrentSpectrum.F"
#include "tallyParticleTrackStructure.F"
#include "tallyPhaseSpaceFile.F"
#include "tallyPhotonFluencePoint.F"
#include "tallyPixelImageDetect.F"
#include "tallyPulseHeightSpectrum.F"
#include "tallySpatialDoseDistrib.F"
#include "tallySphericalDoseDistrib.F"
#include "tallyVoxelDoseDistrib.F"


!*******************************************************************
!*    MAIN                                                         *
!*******************************************************************
      program main
      use TRACK_mod, only : mat,e
      use ctrsimMod, only : nhist
      use dsmaxMod
      implicit none
      logical endsim,absorb
      integer*4 ncross,icol,left
      real*8 ds,dsef,de

      write(*,'(a)') '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      write(*,'(a)') '>>>> This is penEasy v.2019-09-21 >>>>>>>>'
      write(*,'(a)') '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'

      call init                            ! Initialize the PENELOPE/penEasy system

      history: do                          ! Each iteration simulates a new history
        nhist = nhist+1.0d0                ! Update history counter
        call cleans                        ! Empty the stack, just in case
        call source                        ! Push initial particles to the stack
        call tally(1,nhist)                ! The simulation of this history begins

        particle: do                       ! Each iteration simulates a new particle
          call secpar(left)                ! Retrieve a particle from the stack
          if (left.eq.0) exit particle     ! Stack was empty
          call locatevox                   ! Locate in voxels geometry
          call tally(-99,-e)               ! The simulation of this particle begins
          if (absorb()) cycle particle     ! Check particle absorption
          call start                       ! Reset transport mechanics
          call forcing                     ! Set interaction forcing status

          interact: do                     ! Each iteration simulates an interaction
            if (absorb()) cycle particle   ! Check particle absorption
            call jumpx(dsmax(mat),ds)      ! Get distance to next interaction

            call stepx(ds,dsef,ncross)     ! Advance up to interaction point or interface
            if (ncross.eq.0) then          ! Moved a distance DS, no interface crossed
              call tally(3,ds)             ! Notify tallies
            else                           ! Moved a distance DSEF, interface found
              call tally(4,dsef)           ! Notify tallies
              if (mat.eq.0) cycle particle ! New material is vacuum => gone
              call start                   ! New material => reset transport mechanics
              call forcing                 ! Set interaction forcing status
              call russian                 ! Apply Russian roulette
              call splitting               ! Apply particle splitting
              cycle interact               ! Go to next interaction
            endif

            call knockx(de,icol)           ! Simulate an interaction
            call tally(-int(icol),de)      ! Tally kinetic energy released
          enddo interact
        enddo particle

        call tally(6,nhist)                ! End-of-history bookkeeping
        if (endsim()) exit history         ! Simulation is finished
      enddo history

      call report(nhist)                   ! Write final report
      end


      subroutine init
!*******************************************************************
!*    Initializes the simulation system.                           *
!*                                                                 *
!*    Comments:                                                    *
!*      -> The history counter is set to >0 in case of a restart   *
!*      -> The order in which init routines are called must not be *
!*         changed, since some of them provide data for others and *
!*         write data to a dump file. This file must be written in *
!*         the same order in which is will be read afterwards.     *
!*******************************************************************
      use ctrsimMod, only : nhist
      use RSEEDcommonMod
      implicit none

      integer restartf
      integer*4 nmatgeo
      real*8 emax,realtime,cputime,rtime,utime,dmem
      real*8 :: mem=0.0            ! Memory usage

      call initime                 ! Write date on the screen
      call treset                  ! Reset simulation timer to compute init() timing

      call iniconfig(restartf)     ! Simulation config
      nhist = 0.0d0                ! Clear number of simulated histories
      if (restartf.ge.0) read(restartf) nhist,seed1,seed2  ! Simulation restart
      call inisource(emax,dmem)    ! Source models
      mem = mem+dmem               ! Used memory
      call inigeo(nmatgeo,dmem)    ! Geometry: PENGEOM & penVox
      mem = mem+dmem
      call inipen(emax,nmatgeo)    ! PENELOPE init
      call inimassvox(dmem)        ! Init voxels masses (cannot do before PENELOPE inits)
      mem = mem+dmem
      call initally(emax,dmem)     ! Tallies; tallies read from restart file if needed
      mem = mem+dmem
      call iniforce(emax)          ! Interaction forcing
      call inisplit                ! Particle splitting
      call inirussia               ! Russian roulette

      write(*,*) ''
      write(*,*) ''
      write(*,'(a)')
     & '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      write(*,'(a)') 'init: INITIALIZATION ENDED'
      write(*,'(a,f9.2,a)') 'Elapsed real time:',realtime(),' s'
      write(*,'(a,f9.2,a)') 'Elapsed CPU time :',cputime(),' s'
      write(*,'(a,es7.1,a)') "Memory used by penEasy arrays: ",
     &  mem*1.0e-6,' MB'

      call treset                  ! Reset simulation timers

      if (restartf.ge.0) then      ! This is a sim restart
        read(restartf) rtime,utime
        close(restartf)
        call trestart(rtime,utime) ! Re-reset CPU and realtime timers in case of a restart
        write(*,*) ''
        write(*,'(a)') 'Simulation restarted with the following '//
     &                 'data from the previous run:'
        write(*,'(a,f18.0)') '  No. of histories: ',nhist
        write(*,'(a,2(1x,i0))') '  Seed1,Seed2:',seed1,seed2
        write(*,'(a,2(1x,es12.5))') '  RealTime(s),CPUtime(s):',
     &    rtime,utime
      endif

      write(*,'(a)')
     & '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'

      flush(6)                    ! Flush stdout, prevent OS waiting until simulation ends
      end


      subroutine report(n)
!*******************************************************************
!*    Reports final results.                                       *
!*                                                                 *
!*    Input:                                                       *
!*      n -> no. of histories simulated.                           *
!*******************************************************************
      use RSEEDcommonMod
      implicit none
      real*8 n

      integer unc
      real*8 cputime,realtime,nowcpu

      nowcpu = cputime()                 ! Set a reference to measure report timing
      call tallyreport(-1,n,nowcpu,unc)  ! EOS report, each tally will report its data

      write(*,*) ''
      write(*,*) ''
      write(*,'(a)')
     & '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      write(*,'(a)') 'report: SIMULATION ENDED'
      write(*,'(a)')
     & 'Results have been written to the corresponding DAT files.'
      select case(unc)
      case(0)
        write(*,'(a)')
     &   'The requested uncertainty has NOT been reached.'
      case(1)
        continue  ! Uncertainty limit not defined, no message written
      case default
        write(*,'(a)')
     &   'The requested uncertainty has been reached.'
      end select
      write(*,'(a)')
     & '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'

      write(*,*) ''
      write(*,'(a)') 'Last random seeds:'
      write(*,'(2(1x,i0))') seed1,seed2

      write(*,'(a)') 'Elapsed real time (s), excluding init:'
      write(*,'(1x,es12.5)') realtime()

      write(*,'(a)') 'Elapsed CPU time (s), excluding init:'
      write(*,'(1x,es12.5)') nowcpu

      write(*,'(a)') 'Each report update took (in CPU s):'
      write(*,'(1x,es12.5)') cputime()-nowcpu

      write(*,'(a)') 'No. of histories simulated:'
      write(*,'(1x,f18.0)') n

      if (nowcpu.gt.0.0) then
        write(*,'(a)') 'CPU Speed (histories/s):'
        write(*,'(1x,es12.5)') n/nowcpu
      endif

      call endtime  ! Report date and say goodbye
      end


      subroutine jumpx(dsmax,ds)
!*******************************************************************
!*    Determines distance to next interaction. This is a wrap up   *
!*    routine for the various PENELOPE versions depending on       *
!*    type of particle (e.g., protons), interaction forcing status,*
!*    etc.                                                         *
!*                                                                 *
!*    Input:                                                       *
!*      dsmax: max distance (cm), as defined in PENELOPE           *
!*    Output:                                                      *
!*      ds: distance (cm) to next interaction                      *
!*******************************************************************
      use TRACK_mod, only : kpar
      use forcingMod, only : isforcing
      implicit none
      real*8 dsmax,ds

#ifdef PROTONS
      if (kpar.eq.4) then       ! It's a proton
        call jumph(dsmax,ds)    ! Get distance to next interaction (note: forcing not implemented)
      else                      ! Photon-electron transport
#endif
      if (isforcing) then       ! Interaction forcing is active
        call jumpf(dsmax,ds)    ! Get distance DS until next interaction
      else
        call jump(dsmax,ds)     ! Same without interaction forcing
      endif
#ifdef PROTONS
      endif
#endif
      end


      subroutine knockx(de,icol)
!*******************************************************************
!*    Simulates an interaction. This is a wrap up routine for the  *
!*    various PENELOPE versions depending on type of particle      *
!*    (e.g., protons), interaction forcing status, etc.            *
!*                                                                 *
!*    Output:                                                      *
!*      de: energy loss (eV)                                       *
!*      icol: type of interaction, as defined by PENELOPE          *
!*******************************************************************
      use TRACK_mod, only : kpar
      use forcingMod, only : isforcing
      implicit none
      integer*4 icol
      real*8 de

#ifdef PROTONS
      if (kpar.eq.4) then            ! Proton transport
        call knockh(de,icol)         ! Simulate an interaction
      else                           ! Photon-electron transport
#endif
      if (isforcing) then            ! Interaction forcing (IF) is active
        call knockf(de,icol)         ! Simulate an interaction with IF
      else
        call knock(de,icol)          ! Same with no IF
      endif
#ifdef PROTONS
      endif
#endif
      end


      logical function absorb()
!*******************************************************************
!*    Checks whether or not a particle is absorbed given its       *
!*    current energy.                                              *
!*                                                                 *
!*    Input:                                                       *
!*      TRACK_mod                                                  *
!*    Output:                                                      *
!*      .true. if absorbed, .false. else                           *
!*    Comments:                                                    *
!*      -> Tallies deposited energy when necessary; to do so it    *
!*         calls TALLY.                                            *
!*      -> Absorbs particles above 1 GeV; these can be created by  *
!*         SOURCE or be the result of the annihilation of a e+.    *
!*      -> This function is well suited to implement range         *
!*         rejection by making the returned value dependent        *
!*         on the particle's position, material, energy, etc.      *
!*         Notice that in order to preserve the calculation of     *
!*         the fluence unbiased a particle should never be rejected*
!*         *inside* the detector.                                  *
!*******************************************************************
      use PENELOPE_mod, only : eabs
      use TRACK_mod
#ifdef PROTONS
      use PENH_mod, only : eabsh
#endif
      implicit none
      real*8 mc2,twomc2,penemax
      parameter (mc2=5.10998918d5,twomc2=2.0d0*mc2,penemax=1.0d9)

      if (wght.lt.0.0) then     ! Indicates particle must be killed
        call tally(5,0.0d0)
        absorb = .true.
        return
      endif

      if (mat.eq.0) then        ! Particle is gone
        absorb = .true.
        return
      endif

      select case (kpar)
      case (1,2,3)
        if (e.gt.eabs(kpar,mat).and.e.lt.penemax) then
          absorb = .false.
          return
        endif
#ifdef PROTONS
      case (4)
        if (e.gt.eabsh(mat).and.e.lt.penemax) then
          absorb = .false.
          return
        endif
#endif
      case default
        write(*,'(a,i0)')
     &    'absorb:WARNING: unknown particle type: ',kpar
        write(*,'(a)')
     &    '  The particle has been discarded (no E deposition).'
        absorb = .true.
        return
      end select

      absorb = .true.
      call tally(-98,e)         ! Tally remaining kinetic energy
      if (kpar.eq.3.and.e.gt.0.0) then ! e+ absorbed, create 2 gammas; but not if E=0, since KNOCK did it
        call panar(eabs(2,mat)) ! Puts two annihilation photons in the stack if above Eabs
        call tally(-97,twomc2)  ! Tally the extra 2mc^2
      endif

      if (e.ge.penemax) then
        write(*,'(a,es12.5,a)')
     &  'absorb:WARNING: particle with energy out of range: ',e,' eV'
        write(*,'(a)')
     &  '  The particle has been locally absorbed.'
      endif
      end


!*******************************************************************
!*******************************************************************
!*    Source routines start here.                                  *
!*    Source models require:                                       *
!*     i) an initialization routine that must be called by         *
!*        INISOURCE                                                *
!*    ii) a particle generation routine that must be called        *
!*        by SOURCE                                                *
!*******************************************************************
!*******************************************************************

      subroutine inisource(emax,mem)
!*******************************************************************
!*    Init routines for source models.                             *
!*                                                                 *
!*    Output:                                                      *
!*      emax: max source energy (eV).                              *
!*      mem: memory used by source arrays.                         *
!*    Comments:                                                    *
!*      Note that if more than one source is defined and active    *
!*      then EMAX must be the maximum of the value reported by all *
!*      of them.                                                   *
!*******************************************************************
      use sourceMod
      implicit none
      real*8 emax,mem

      real*8 emax1,dmem

      emax = 0.0
      mem  = 0.0

      call BIGSinisrc(BIGSactive,emax1,dmem)
      emax = max(emax,emax1)    ! Max energy
      mem  = mem+dmem           ! Memory used

      ! This illustrates the mechanism to duplicate a source init:
  !!! call BIGS2inisrc(BIGS2active,emax1,dmem)
  !!! emax = max(emax,emax1)
  !!! mem  = mem+dmem

      call PSFSinisrc(PSFSactive,emax1)
      emax = max(emax,emax1)    ! Max energy

      if (.not.BIGSactive.and..not.PSFSactive) then
        write(*,'(a)') 'inisource:ERROR: no sources ON.'
        stop
      endif
      if (BIGSactive.and.PSFSactive) then
        write(*,'(a)') 'inisource:ERROR: PSF source ON is '//
     &    'incompatible with other sources ON.'
        stop
      endif
      end


      subroutine source
!*******************************************************************
!*    Source models.                                               *
!*******************************************************************
      use sourceMod
      implicit none
      if (BIGSactive) call BIGSsource
      ! This illustrates the mechanism to duplicate a source call:
  !!! if (BIGS2active) call BIGS2source

      if (PSFSactive) call PSFSource
      end


!*******************************************************************
!*******************************************************************
!*    Tally routines start here.                                   *
!*    Tallies require:                                             *
!*    i) an initialization routine that must be called by INITALLY *
!*    ii) a tally routine that must be called by TALLY             *
!*    iii) a reporting routine that must be called by TALLYREPORT  *
!*    iv) a dump routine to allow restarting of the simulation.    *
!*                                                                 *
!*    Notice that the ordering of the tally initialization routines*
!*    must coincide with the ordering of the corresponding sections*
!*    in the input file.                                           *
!*******************************************************************
!*******************************************************************

      subroutine initally(emax,mem)
!*******************************************************************
!*    Init tallying routines.                                      *
!*                                                                 *
!*    Input:                                                       *
!*      emax: max source energy (eV).                              *
!*    Output:                                                      *
!*      mem: memory used by arrays in tallies.                     *
!*******************************************************************
      use tallyMod
      implicit none
      real*8 mem,emax,dmem

      mem = 0.0
      call VDDinitally(VDDactive,dmem)
      mem = mem+dmem
      call SDDinitally(SDDactive,dmem)
      mem = mem+dmem
      call CDDinitally(CDDactive,dmem)
      mem = mem+dmem
      call SPDinitally(SPDactive,dmem)
      mem = mem+dmem
      call EDPinitally(EDPactive,dmem)
      mem = mem+dmem
      call PHSinitally(PHSactive,dmem)
      mem = mem+dmem
      call PIDinitally(PIDactive,dmem)
      mem = mem+dmem
      call FTLinitally(FTLactive,dmem)
      mem = mem+dmem
      call PFPinitally(PFPactive,dmem,emax)
      mem = mem+dmem
      call PSFinitally(PSFactive,emax)
      call PCSinitally(PCSactive,dmem)
      mem = mem+dmem
      call PTSinitally(PTSactive)
      end


      subroutine tally(mode,arg)
!*******************************************************************
!*    Tallying routines.                                           *
!*******************************************************************
      use tallyMod
      implicit none
      integer mode
      real*8 arg

      if (VDDactive) call VDDtally(mode,arg)  ! Must be 1st tally to be called
      if (SDDactive) call SDDtally(mode,arg)
      if (CDDactive) call CDDtally(mode,arg)
      if (SPDactive) call SPDtally(mode,arg)
      if (EDPactive) call EDPtally(mode,arg)
      if (PHSactive) call PHStally(mode,arg)
      if (PIDactive) call PIDtally(mode,arg)
      if (FTLactive) call FTLtally(mode,arg)
      if (PFPactive) call PFPtally(mode,arg)
      if (PSFactive) call PSFtally(mode,arg)
      if (PCSactive) call PCStally(mode)
      if (PTSactive) call PTStally(mode,arg)
      end


      subroutine tallyreport(mode,n,cputim,unc)
!*******************************************************************
!*    Calls report routines for all tallies and PSF sources.       *
!*                                                                 *
!*    Input:                                                       *
!*      mode -> -1 if called at end-of-simulation, 0 for dump only,*
!*               1 for tally report only, 2 for both.              *
!*      n -> no. of histories simulated                            *
!*      cputim -> elapsed CPU time                                 *
!*    Output:                                                      *
!*      unc -> 0: uncert not reached, 1: undefined, >1: reached    *
!*    Comments:                                                    *
!*      -> The sequence of calls to the report routines MUST be the*
!*         same used for the corresponding initialization routines *
!*         in INISOURCE and INITALLY (see above). This is to ensure*
!*         proper manipulation of the simulation dump file.        *
!*      -> PSF sources need to report the total num. of particles  *
!*         read to the dump file, which may be needed in case of a *
!*         restart.                                                *
!*******************************************************************
      use sourceMod, only : PSFSactive
      use tallyMod
      use dumpMod
      use RSEEDcommonMod
      implicit none
      integer mode,unc,uncdone
      real*8 n,cputim

      integer finduf,error
      real*8 realtime,cputime

      ! Write dump file if needed:
      if (mode.ne.1.and.dumpfile.ne.-1) then  ! Create sim dump file
        dumpfile = finduf()                   ! Find a valid unit for the file
        open(dumpfile,file=dumpfilen,status='replace',access='stream',
     &       iostat=error)
        if (error.ne.0) then
          write(*,'(a)')
     &      'tallyreport:ERROR: unable to open dump file.'
          dumpfile = -2                       ! Tells others that file coundn't be opened
        else
          write(dumpfile) n,seed1,seed2       ! Write history state to dump file
        endif
      endif

      ! Write source reports:
      if (PSFSactive) call PSFSreport(mode)  ! PSF source dumps the num. of particles read

      ! Write tally reports:
      unc = 1                                ! Unknown uncertainty status at this point
      uncdone = 1
      if (VDDactive) call VDDreport(mode,n,cputim,uncdone)
      unc = unc*uncdone
      if (SDDactive) call SDDreport(mode,n,cputim,uncdone)
      unc = unc*uncdone
      if (CDDactive) call CDDreport(mode,n,cputim,uncdone)
      unc = unc*uncdone
      if (SPDactive) call SPDreport(mode,n,cputim,uncdone)
      unc = unc*uncdone
      if (EDPactive) call EDPreport(mode,n,cputim,uncdone)
      unc = unc*uncdone
      if (PHSactive) call PHSreport(mode,n,cputim,uncdone)
      unc = unc*uncdone
      if (PIDactive) call PIDreport(mode,n,cputim,uncdone)
      unc = unc*uncdone
      if (FTLactive) call FTLreport(mode,n,cputim,uncdone)
      unc = unc*uncdone
      if (PFPactive) call PFPreport(mode,n,cputim,uncdone)
      unc = unc*uncdone
      if (PSFactive) call PSFreport(mode,n,cputim)  ! No uncertainty for this tally
      if (PCSactive) call PCSreport(mode,n,cputim,uncdone)
      unc = unc*uncdone

      if (mode.ne.1.and.dumpfile.ge.0) then
        write(dumpfile) realtime(),cputime() ! Write timings to dump file
        close(dumpfile)
      endif
      end


!>>>> End Of File >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
