!*******************************************************************
!*                         SOURCE                                  *
!*               BOX-ISOTROPIC-GAUSS-SPECTRUM                      *
!*                                                                 *
!* Short description:                                              *
!*   Generation of primary particle states for PENELOPE.           *
!* Rev:                                                            *
!*   2019-09-21                                                    *
!*******************************************************************


!*******************************************************************
!*******************************************************************
!*    Modules section                                              *
!*******************************************************************
!*******************************************************************

      module BIGSmod
!*******************************************************************
!*    Vars for the BIGS source.                                    *
!*******************************************************************
      implicit none
      save
      logical :: isPennucInit=.false.    ! Serves to avoid conflicts in case of source duplication
      logical isPosGauss,isPennuc
      integer xkpar,matsrc,nspc,xipol
      integer, parameter :: nemax=32000
      integer*4 :: metast=0              ! Metastable flag used by PENNUC, must be init to zero
      real*8 espc(nemax),pspc(nemax),despc(nemax),esigma
      real*8 rotv(3,3),rotbox(3,3)
      real*8 cossrc0,cossrc1,dcossrc,phi0,phi1,dphi
      real*8 boxside(3),ctrans(3),xsigma,ysigma,xsp1,xsp2,xsp3
      real*8, parameter :: kfact=6.0d0   ! Maximum num. of sigmas in Gaussian dispersion
      end


!*******************************************************************
!*******************************************************************
!*    Routines section                                             *
!*******************************************************************
!*******************************************************************

      subroutine BIGSsource
!*******************************************************************
!*    Output:                                                      *
!*      The generated particle is pushed to PENELOPE's stack.      *
!*******************************************************************
      use BIGSmod, only : isPennuc,xkpar,
     &                    xipol,xsp1,xsp2,xsp3,matsrc
      use TRACK_mod
      implicit none
      logical BIGSisInfield
      integer*4 ncross
      real*8 dsef
      real*8, parameter :: infty=1.0e30

      if (isPennuc) then              ! Use pennuc radionuclide package
        call BIGSpennuc               ! Generate one cascade and push it to stack
        return                        ! Done
      endif

      kpar = xkpar                    ! Particle type
      wght = 1.0                      ! Init statistical weight
      ipol = xipol                    ! Photon polarization physics
      sp1  = xsp1                     ! Polarization state
      sp2  = xsp2
      sp3  = xsp3
      ilb  = 0                        ! Clear ILB labels
      ilb(1) = 1                      ! Tag as primary source particle
      lage = .false.                  ! Do not track particle age
      page = 0.0                      ! Reset age, just in case

      call BIGSsampleEnergy           ! Sample energy from input spectrum
      call BIGSsampleDirection        ! Sample direction...
      call BIGSsamplePosition         ! ...& position (after direction in case particle on a surface)
      if (matsrc.lt.0) then           ! In-field beam, accept direction & position only if aiming at field
        do                            ! Iterate until a particle aiming at field is found
          if (BIGSisInfield()) exit   ! Particle aiming at field, accept
          call BIGSsampleDirection    ! Else, resample direction & position
          call BIGSsamplePosition     ! This routine keeps track of sampling efficiency, aborts if too low
        enddo
      endif
      if (mat.eq.0) call stepx(infty,dsef,ncross) ! Particle born in vacuum, advance up to object or infty

      call stores(e,x,y,z,u,v,w,wght,kpar,ilb,ipol)  ! Push newborn particle to stack
      call tally(0,e)                 ! Inform tallies that new particle was pushed to stack
      end


      subroutine BIGSpennuc
!*******************************************************************
!*    Output:                                                      *
!*      A decay cascade is generated and pushed to the stack using *
!*      the PENNUC package.                                        *
!*******************************************************************
      use BIGSmod, only :
     &  rotv,cossrc0,cossrc1,phi0,phi1,matsrc,xipol,metast
      use PENNUC_mod, only : nrm,nr,kpnr,itnr,enr,agenr,unr,vnr,wnr
      use SECSTcommonMod, only : nsec
      use ctrsimMod, only : nhist
      use TRACK_mod
      implicit none
      logical BIGSisInfield
      integer*4 i,ncross,bodynuc,matnuc,pennucError
      integer*4, save :: nfails=0       ! Total number of fails since program start
      real*8 dsef,phi,xnuc,ynuc,znuc
      real*8, parameter :: pi=3.1415926535897932d0,twopi=2.0*pi
      real*8, parameter :: infty=1.0e30
      real*8 d(3)

      w = 1.0                                          ! Set arbitrary initial direction for LOCATE to work
      v = 0.0
      u = 0.0
      call BIGSsamplePosition                          ! Sample cascade position
      xnuc = x                                         ! Store position, use later to reset every particle
      ynuc = y
      znuc = z
      bodynuc = ibody
      matnuc = mat

      i = 1                                            ! Used as dummy var for PENNUC, which writes on 1st argument
      call pennuc(i,metast,pennucError)                ! Generate cascade; metast in a module because it must be stored between calls
      if (pennucError.ne.0) then
        write(*,'(a)') 'BIGSpennuc:internal ERROR: '//
     &     'PENNUC could not generate the cascade.'
        stop
      endif

      ! These state vars are common to all particles in the cascade:
      wght = 1.0                                       ! Init statistical weight
      ilb  = 0                                         ! Clear ILB labels
      ilb(1) = 1                                       ! All particles tagged as primaries
      ipol = xipol                                     ! Set user-defined switch for polarization physics
      sp1  = 0.0                                       ! Assume photons are unpolarized
      sp2  = 0.0
      sp3  = 0.0
      lage = .false.                                   ! penEasy does not keep particle ages

      particle: do i=1,nr                              ! For every particle in cascade, check its direction limits
                                                       !   assuming pennuc directions are represented in Sv frame
        if (kpnr(i).gt.3) cycle particle               ! PENNUC produces alpha particles with kpar=4; reject
        d = (/unr(i),vnr(i),wnr(i)/)                   ! Direction vector in vector format
        if (d(3).lt.cossrc1.or.d(3).gt.cossrc0)        ! Out of polar interval
     &    cycle particle                               ! Reject this particle, go to next in stack
        if (max(abs(d(2)),abs(d(1))).gt.0.0) then      ! Avoid atan(0,0); if (0,0) => w=1, accepted
          phi = atan2(vnr(i),unr(i))                   ! Azimuthal angle in [-pi,+pi]
          if (phi.lt.0.0) phi=phi+twopi                ! Same in [0,+2pi)
          if (.not.((phi.ge.phi0.and.phi.le.phi1).or.  ! Out of azimuthal interval
     &        (phi.le.phi1-twopi)))                    ! Allow one turn for phi
     &      cycle particle                             ! Reject this particle, go to next in stack
        endif                                          ! Else, accept particle

        d = matmul(rotv,d)                             ! Rotate direction vector from Sv to Slab
        u = d(1)                                       ! Set particle state direction to rotated vector
        v = d(2)
        w = d(3)
        if (matsrc.lt.0) then                          ! In-field beam, check if direction aims at field
          if (.not.BIGSisInfield()) cycle particle     ! It does not, reject particle
        endif

        kpar   = kpnr(i)                               ! Particle accepted, load state vars into TRACK
        e      = enr(i)
        ilb(4) = itnr(i)                               ! Identifies x-ray, Auger transitions & nuclear transitions (latter with value <0)
        page   = agenr(i)                              ! Keep birth age, just in case
        if (matnuc.eq.0) then                          ! Particles are born in vacuum
          x = xnuc                                     ! Reset position vars in case previous STEPX moved it
          y = ynuc
          z = znuc
          ibody = bodynuc
          mat = 0
          call stepx(infty,dsef,ncross)                ! Advance up to the object or infinity
        endif

        call stores(e,x,y,z,u,v,w,wght,kpar,ilb,ipol)  ! Push accepted newborn particle to stack
        call tally(0,e)                                ! Inform tallies that new particle was pushed to stack
      enddo particle

      ! Check if sampling efficiency is not too low:
      if (nsec.eq.0) nfails = nfails+1                  ! The stack is empty, source failed to produce any valid particle
      if (dble(nfails).gt.max(nhist*0.99d0,1.0d3)) then ! More than 99% failure rate, test after 1,000 histories
        write(*,'(a)')
     &    'BIGSpennuc:ERROR: source emitted particles '//
     &    'inside the user-defined field '//
     &    'in less than 1% of histories.'
        write(*,'(a)')
     &    'Number of histories and sampling efficiency:'
        write(*,'(2x,f18.0,1x,f5.2,a)')
     &    nhist,100.0*(1.0-dble(nfails)/nhist),'%'
        write(*,'(a)')
     &    'Enlarge the source aperture appropriately to '//
     &    'avoid this error.'
        call exit(1)
      endif
      end


      subroutine BIGSsampleDirection
!*******************************************************************
!*    Samples a particle's direction of flight.                    *
!*                                                                 *
!*    Output:                                                      *
!*      The generated position is returned via the TRACK module.   *
!*******************************************************************
      use BIGSmod, only : cossrc0,dcossrc,phi0,dphi,rotv
      use TRACK_mod, only : u,v,w
      implicit none
      real*8, external :: rand
      real*8 d(3),sintheta,phi

      d(3) = cossrc0+dcossrc*rand(0.0d0)
      sintheta = sqrt(1.0-d(3)**2)
      phi = phi0+dphi*rand(1.0d0)
      d(2) = sintheta*sin(phi)
      d(1) = sintheta*cos(phi)  ! d is the flight direction in Sv frame
      d = matmul(rotv,d)        ! Rotate direction vector from Sv to Slab
      u = d(1)                  ! Transfer to PENELOPE vars
      v = d(2)
      w = d(3)
      end


      subroutine BIGSsamplePosition
!*******************************************************************
!*    Samples a valid particle's position keeping track of the     *
!*    number of failed attemps. If ratio of failures is too large  *
!*    a warning message is issued and if it is extreme, stops.     *
!*                                                                 *
!*    Output:                                                      *
!*      The generated position, which defines x,y,z,ibody,mat is   *
!*      returned via the TRACK module.                             *
!*    Comments:                                                    *
!*      - Requires that direction of flight be set previously for  *
!*        SAMPLEBOX to operate reliably.                           *
!*******************************************************************
      use BIGSmod, only : matsrc
      use TRACK_mod, only : mat
      use ctrsimMod, only : nhist
      implicit none

      logical, save :: warned=.false.    ! Warning flag
      integer*8, save :: ntrials=0       ! Total number of trials since program start

      do                                 ! Loop until a valid position is found
        ntrials = ntrials+1              ! Count num. attempts to measure efficiency
        call BIGSsampleBox               ! Sample a particle position inside the box

        ! Test sampling efficiency and issue warnings and errors if needed:
        if (.not.warned.and.dble(ntrials).gt.max(nhist*10.0,1.0d3))
     &    then
          write(*,*) ''
          write(*,'(a)') '***************'
          write(*,'(a)') 'samplePosition:WARNING: '//
     &      'sampling source efficiency is lower than 10%.'
          write(*,'(a)') 'Redefine source parameters appropriately.'
          write(*,'(a)')
     &      'Number of histories and sampling efficiency:'
          write(*,'(2x,f18.0,1x,f5.2,a)')
     &      nhist,nhist/dble(ntrials)*100.0,'%'
          write(*,'(a)') '***************'
          write(*,*) ''
          warned = .true.
        endif
        if (dble(ntrials).gt.max(nhist*1000.0,1.0d5)) then
          write(*,'(a)') 'samplePosition:ERROR: '//
     &      'sampling source efficiency is lower than 0.1%.'
          write(*,'(a)') 'Redefine source parameters appropriately.'
          write(*,'(a)')
     &      'Number of histories and sampling efficiency:'
          write(*,'(2x,f18.0,1x,f5.2,a)')
     &      nhist,nhist/dble(ntrials)*100.0,'%'
          stop
        endif

        if (matsrc.le.0.or.matsrc.eq.mat) return  ! Point accepted, else cycle; note MATSRC<0 for in-beam case
      enddo
      end


      subroutine BIGSsampleBox
!*******************************************************************
!*    Samples a position inside the source box.                    *
!*                                                                 *
!*    Output:                                                      *
!*      The generated position, which defines x,y,z,ibody,mat is   *
!*      returned via the TRACK module.                             *
!*    Comments:                                                    *
!*      - Requires that direction of flight be set previously for  *
!*        LOCATE to operate reliably.                              *
!*******************************************************************
      use BIGSmod, only : isPosGauss,boxside,
     &                    xsigma,ysigma,rotbox,ctrans
      use TRACK_mod, only : x,y,z
      implicit none
      integer i
      real*8 r(3),gausspdf
      real*8, external :: rand

      do i=1,3
        r(i) = boxside(i)*(rand(dble(i))-0.5)  ! Coordinates in Sbox
      enddo
      if (isPosGauss) then                     ! Gaussian x,y displacement in Sbox
        r(1) = r(1)+xsigma*gausspdf()          ! Apply displacement in Sbox frame
        r(2) = r(2)+ysigma*gausspdf()
      endif

      r = matmul(rotbox,r)+ctrans              ! Apply rotation and translation
      x = r(1)                                 ! Transfer to PENELOPE vars
      y = r(2)
      z = r(3)

      call locate                              ! Set quadric body and material
      call locatevox                           ! Set voxel indices and material, if inside a voxel geometry
      end


      logical function BIGSisInfield()
!*******************************************************************
!*    Determines if the particle direction aims at the field       *
!*    determined by -MATSRC.                                       *
!*                                                                 *
!*    Input:                                                       *
!*      Particle state via TRACK module.                           *
!*    Output:                                                      *
!*      TRUE if particle aims at the field, TRUE else.             *
!*******************************************************************
      use BIGSmod, only : matsrc            ! Material defining the field
      use partvoxMod, only : xvox,yvox,zvox,absvox
      use TRACK_mod, only : x,y,z,ibody,mat
      implicit none
      integer*4 ncross,bodytmp,mattmp,xvoxtmp,yvoxtmp,zvoxtmp,voxtmp
      real*8 xtmp,ytmp,ztmp,dsef
      real*8, parameter :: infty=1.0e30

      xtmp    = x                           ! Store position vars
      ytmp    = y
      ztmp    = z
      xvoxtmp = xvox
      yvoxtmp = yvox
      zvoxtmp = zvox
      voxtmp  = absvox
      bodytmp = ibody
      mattmp  = mat

      BIGSisInfield = .false.               ! Default is 'out-of-field'
      do
        if (mat.eq.-matsrc) then            ! Inside field
          BIGSisInfield = .true.
          x = xtmp                          ! Restore position and return
          y = ytmp
          z = ztmp
          xvox   = xvoxtmp
          yvox   = yvoxtmp
          zvox   = zvoxtmp
          absvox = voxtmp
          ibody  = bodytmp
          mat    = mattmp
          return
        endif
        call stepx(infty,dsef,ncross)       ! Move forward to check if the field is on the path
        if (mat.eq.0.or.ncross.eq.0) return ! Gone to infty, out of field
      enddo
      end


      subroutine BIGSsampleEnergy
!*******************************************************************
!*    Samples particle's energy.                                   *
!*                                                                 *
!*    Output:                                                      *
!*      The generated energy is returned via the TRACK module.     *
!*******************************************************************
      use BIGSmod
      use TRACK_mod, only : e
      implicit none
      integer i,seeki
      real*8 randno,sigf,gausspdf
      real*8, external :: rand

      randno = rand(0.0d0)
      i = seeki(pspc,randno,nspc)
      e = espc(i)+(randno-pspc(i))*despc(i)
      if (esigma.ne.0.0) then             ! Gaussian distrib
 10     sigf = gausspdf()
        if (abs(sigf).gt.kfact) goto 10   ! Truncate Gauss at +-kfact
        e = max(e+esigma*sigf,0.0)        ! Prevent negative values
      endif
      end


      subroutine BIGSinisrc(activated,emax,mem)
!*******************************************************************
!*    Initializes the source.                                      *
!*                                                                 *
!*    Output:                                                      *
!*      activated: TRUE if the source is active.                   *
!*      emax: max source energy (eV)                               *
!*      mem: memory used by source arrays.                         *
!*******************************************************************
      use BIGSmod
      use dataTypesMod
#ifdef f2003
      use, intrinsic :: iso_fortran_env, only :
     &  stdin=>input_unit,stdout=>output_unit,stderr=>error_unit
#else
#define stdin  5
#define stdout 6
#define stderr 0
#endif
      implicit none
      logical activated
      real*8 emax,mem

      character*80 buffer,upcase,spcfilen
      character*45 pennucFilename     ! Size matches that of input var in PENNUC init routine
      character*(*) secid,eos
      parameter (secid=
     &  '[SECTION SOURCE BOX ISOTROPIC GAUSS SPECTRUM v.2014-12-21]')
      parameter (eos='[END OF BIGS SECTION]')
      integer j,answer,errorcode,spcfile,finduf
      integer*4 pennucError
      real*8 omega,theta,phi,prob,norm,theta0,theta1
      real*8 boxctr(3),boxtrans(3),dsrc(3)
      real*8 pi,fwhm2sig,deg2rad,mc2,twomc2,zero,oneeps
      parameter (pi=3.1415926535897932d0,deg2rad=pi/180.0)
      parameter (fwhm2sig=1.0/sqrt(8.0*log(2.0d0)))
      parameter (mc2=5.10998918d5,twomc2=2.0*mc2,zero=1.0d-30)
      parameter (oneeps=1.0d0+1.0d-14)

      emax = 0.0  ! Defaults
      mem  = 0.0

      write(*,*) ' '
      write(*,'(a)')
     & '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      call getline(buffer,0)
      if (index(buffer,secid).eq.0) then
        write(*,'(a)') 'BIGSinisrc:ERROR: incorrect section header;'
        write(*,'(a,a)') '  expecting to find: ',secid
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,'(a)') secid

      read(*,'(a80)') buffer
      buffer = adjustl(upcase(buffer))    ! Adjust and capitalize
      buffer = buffer(1:scan(buffer,' ')) ! Clip at 1st blank
      if (buffer.eq.'ON') then
        activated = .true.
      else if (buffer.eq.'OFF') then
        activated = .false.
        write(*, '(a)')
     &    '>>>> Source Box Isotropic Gauss Spectrum is OFF >>>>'
        do
          read(*,'(a80)',iostat=errorcode) buffer
          if (errorcode.ne.0) then
            write(*,'(a,a,a)') 'BIGSinisrc:ERROR: ',
     &       'Unable to find End-Of-Section mark: ',eos
            stop
          endif
          if (index(buffer,eos).ne.0) return
        enddo
      else
        write(*,'(a)')
     &    'BIGSinisrc:ERROR: expecting to find ON or OFF'
        write(*,'(a)') 'found instead:'
        write(*,'(a)') buffer
        stop
      endif

      ! Report memory usage:
      mem = sizeOfReal8*3*real(nemax)
      write(*,'(a)') 'Memory used by static arrays (MB):'
      write(*,'(1x,es12.5)') mem*1.0e-6

      ! Type of particle:
      write(*,'(a)') 'Source particle type:'
      read(*,'(a80)') buffer
      read(buffer,*,iostat=errorcode) xkpar
      if (errorcode.eq.0) then                     ! Standard BIGS source mode, one particle type
        isPennuc = .false.                         ! PENNUC source not used
        write(*,'(1x,i0)') xkpar
#ifdef PROTONS
        if (xkpar.lt.1.or.xkpar.gt.4) then         ! Protons are KPAR=4
#else
        if (xkpar.lt.1.or.xkpar.gt.3) then
#endif
          write(*,'(a)') 'BIGSinisrc:ERROR: invalid particle type'
          stop
        endif
      else                                          ! PENNUC source
        xkpar = 99                                  ! Invalid code signals PENNUC; XKPAR code used by PFPtally
        isPennuc = .true.
        buffer = adjustl(buffer)                    ! Clip leading blanks
        pennucFilename = buffer(1:scan(buffer,' ')) ! Clip at 1st trailing blank
        write(*,'(a,a)') '  radionuclide filename: ',
     &    trim(pennucFilename)
        write(*,'(a)') 'Initializing PENNUC'
        if (isPennucInit) then                      ! To prevent PENNUC duplication
          write(*,'(a)') 'BIGSinisrc:ERROR: '//
     &       'only one instance of PENNUC can be active.'
          stop
        endif
        call pennuc0(pennucFilename,emax,stdout,pennucError)  ! Init PENNUC, returns max E
        if (pennucError.ne.0) then
          write(*,'(a)') 'BIGSinisrc:ERROR: '//
     &       'PENNUC reported an error on initialization.'
          stop
        endif
        write(*,'(a)') 'PENNUC initialization finished'
        isPennucInit = .true.                       ! Flag as already init
      endif

      ! Photon polarization:
      read(*,'(a80)') buffer  ! Read a dummy line
      write(*,'(a)') 'Photon polarization physics:'
      read(*,*) xipol
      if (xipol.eq.0) then
        write(*,'(a)') '  not activated'
      else if (xipol.eq.1) then
        write(*,'(a)') '  activated'
      else
        write(*,'(a)')
     &    'BIGSinisrc:ERROR: invalid entry, should be 0 or 1'
        stop
      endif
      if (xipol.eq.1.and..not.isPennuc) then            ! Polarization is ON, read Stokes
        write(*,'(a)') 'Stokes parameters:'
        read(*,*) xsp1,xsp2,xsp3
        write(*,'(3(1x,es12.5))') xsp1,xsp2,xsp3
        if (xsp1**2+xsp2**2+xsp3**2.gt.oneeps) then
          write(*,'(a)')
     &      'BIGSinisrc:ERROR: vector P={P1,P2,P3} must be P^2 <= 1'
          stop
        endif
      else                                              ! Polarization off or PENNUC active
        write(*,'(a)') 'Stokes parameters not used.'
        read(*,'(a80)') buffer                          ! Read a dummy line with the unused SP's
        xsp1 = 0.0                                      ! Just in case
        xsp2 = 0.0
        xsp3 = 0.0
      endif

      ! Position:
      read(*,'(a80)') buffer  ! Read a dummy line
      write(*,'(a)') 'Coordinates of box center (cm):'
      read(*,*) boxctr
      write(*,'(3(1x,es12.5))') boxctr
      write(*,'(a)') 'Box sides (cm):'
      read(*,*) boxside
      write(*,'(3(1x,es12.5))') boxside
      write(*,'(a)') 'FWHMs of x and y distribs. (cm):'
      read(*,*) xsigma,ysigma
      write(*,'(2(1x,es12.5))') xsigma,ysigma
      isPosGauss=.false.
      if (xsigma.ne.0.0d0.or.ysigma.ne.0.0d0) isPosGauss=.true.
      xsigma = xsigma*fwhm2sig
      ysigma = ysigma*fwhm2sig
      write(*,'(a)') 'Euler angles for box rotation (deg):'
      read(*,*) omega,theta,phi
      write(*,'(3(1x,es12.5))') omega,theta,phi
      call rotation(omega*deg2rad,theta*deg2rad,phi*deg2rad,rotbox)  ! Rotation Sbox -> Slab
      write(*,'(a)') 'Translation of box center (cm):'
      read(*,*) boxtrans
      write(*,'(3(1x,es12.5))') boxtrans
      ! Note: the position X of a source point is determined as follows:
      !   X = R.Xo+D
      ! where R is the rotation defined by the Euler angles and D the translation.
      ! Xo is the position before rotation and translation, given by
      !   Xo = C + (RAND-0.5)*dX + dGauss
      ! with C the center of the box, dX its sides and dGauss the [x,y]-Gaussian dispersion.
      ! The factor RAND is different for each component of dX. So,
      !   X = (R.C+D) + R.[(rand-0.5)*dX + dGauss]
      ! The vector R.C+D is computed and stored here for later use:
      ctrans = matmul(rotbox,boxctr)+boxtrans

      ! Material:
      write(*,'(a)') 'Source material:'
      read(*,*) matsrc
      write(*,'(1x,i0)') matsrc
      if (matsrc.lt.0) write(*,'(a,i0)')
     &  'In-field beam mode; '//
     &  'particles will be aiming at material ',-matsrc

      ! Direction:
      read(*,'(a80)') buffer  ! Read a dummy line
      write(*,'(a)') 'Direction vector (u,v,w):'
      read(*,*) dsrc
      write(*,'(3(1x,es12.5))') dsrc
      norm = sqrt(sum(dsrc**2))
      if (norm.lt.zero) then
        write(*,*) 'BIGSinisrc:ERROR: null direction.'
        stop
      endif
      dsrc = dsrc/norm                                                 ! Direction vector normalized to 1
      if (max(abs(dsrc(1)),abs(dsrc(2))).gt.0.0) then                  ! Avoid atan(0,0)
        call rotation(0.0d0,acos(dsrc(3)),atan2(dsrc(2),dsrc(1)),rotv) ! Rotation Sv -> Sbox
      else                                                             ! Case dsrc = +z or -z
        call rotation(0.0d0,acos(dsrc(3)),0.0d0,rotv)                  ! Phi=0 is irrelevant
      endif
      write(*,'(a)') 'Polar angle interval, [theta0,theta1](deg):'
      read(*,*) theta0,theta1
      write(*,'(2(1x,es12.5))') theta0,theta1
      if (theta1.lt.theta0.or.min(theta0,theta1).lt.0.0.or.
     &    max(theta0,theta1).gt.180.0) then
        write(*,*) 'BIGSinisrc:ERROR: invalid interval.'
        stop
      endif
      cossrc0 = cos(theta0*deg2rad)          ! Converted to rad
      cossrc1 = cos(theta1*deg2rad)
      dcossrc = cossrc1-cossrc0              ! Is negative or zero
      write(*,'(a)') 'Azimuthal angle interval, phi0 and Dphi (deg):'
      read(*,*) phi0,dphi
      write(*,'(2(1x,es12.5))') phi0,dphi
      if (phi0.lt.0.0d0.or.phi0.ge.360.0d0.or.
     &    dphi.lt.0.0d0.or.dphi.gt.360.0d0) then
        write(*,'(a)') 'BIGSinisrc:ERROR: invalid interval.'
        stop
      endif
      phi0 = phi0*deg2rad
      dphi = dphi*deg2rad
      phi1 = phi0+dphi
      write(*,'(a)')
     &  'Apply to direction same rotation used for box position?'
      read(*,*) answer
      select case (answer)
      case (0)
        write(*,'(a)') ' No'
      case (1)
        write(*,'(a)') ' Yes'
        rotv = matmul(rotbox,rotv)  ! Combined rotation Sv -> Sbox -> Slab
      case default
        write(*,'(a)') 'BIGSinisrc:ERROR: '//
     &    'Invalid entry. Must be 0 or 1.'
        stop
      end select

      ! Energy:
      if (.not.isPennuc) then                  ! PENNUC inactive, read spectrum
        read(*,'(a80)') buffer                 ! Read a dummy line
        read(*,'(a80)') buffer                 ! Spectrum filename
        buffer = adjustl(buffer)
        spcfilen = buffer(1:scan(buffer,' '))  ! Clip at 1st blank
        if (spcfilen.eq.'-') then
          write(*,'(a)') 'Reading energy spectrum from config file.'
          read(*,'(a80)') buffer               ! Dummy header line
          write(*,'(a)') '  Energy(eV)  Relat.Probability   Bin#'
        else
          write(*,'(a)') 'Reading energy spectrum from file:'
          write(*,'(1x,a)') spcfilen
          spcfile = finduf()                   ! Find a valid unit for the file
          open(spcfile,file=spcfilen,status='old',iostat=errorcode)
          if (errorcode.ne.0) then
            write(*,'(a)')
     &        'BIGSinisrc:ERROR: unable to open spectrum file.'
            stop
          endif
        endif

        nspc = 0                               ! Init num. E bins
        pspc(1) = 0.0                          ! Init accumulated prob.
        do
          nspc = nspc+1
          if (spcfilen.eq.'-') then            ! Read from stdin
            read(*,*,iostat=errorcode) espc(nspc),prob
          else
            call getline(buffer,spcfile)       ! Read a non-comment line from file
            read(buffer,*,iostat=errorcode) espc(nspc),prob
          endif
          if (errorcode.ne.0) then
            write(*,'(a)') 'BIGSinisrc:ERROR: '//
     &        'invalid entry at data line:'
            write(*,'(1x,i0)') nspc
            stop
          endif
          if (spcfilen.eq.'-')
     &      write(*,'(2(1x,es12.5),1x,i0)') espc(nspc),prob,nspc  ! Write table to stdout
          if (espc(nspc).lt.0.0) then
            write(*,*) 'BIGSinisrc:ERROR: negative energy'
            stop
          else if (espc(nspc).lt.espc(max(nspc-1,1))) then
            write(*,*) 'BIGSinisrc:ERROR: decreasing energy'
            stop
          endif
          if (prob.lt.0.0) exit                ! End of spectrum
          if (nspc.ge.nemax) then
            write(*,*) 'BIGSinisrc:ERROR: too many bins in spectrum;'
            write(*,*) '              enlarge NEMAX'
            stop
          endif
          pspc(nspc+1) = pspc(nspc)+prob       ! Accumulated probability
        enddo
        if (spcfilen.ne.'-') close(spcfile)

        write(*,'(a)') 'No. of bins read:'
        write(*,'(1x,i0)') nspc-1
        if (nspc-1.lt.1) then
          write(*,*)
     &      'BIGSinisrc:ERROR: at least 1 bin must be defined'
          stop
        endif
        write(*,'(a)')
     &    'Sum of relative probabilities before normalization:'
        write(*,'(1x,es12.5)') pspc(nspc)
        if (pspc(nspc).le.0.0) then
          write(*,'(a)')
     &      'BIGSinisrc:ERROR: all probabilities are zero.'
          stop
        endif
        do j=1,nspc    ! Normalize to unity
          pspc(j) = pspc(j)/pspc(nspc)
        enddo
        do j=1,nspc-1    ! Prepare auxiliar array for sampling
          despc(j) = 0.0
          if (pspc(j+1).gt.pspc(j))
     &      despc(j) = (espc(j+1)-espc(j))/(pspc(j+1)-pspc(j))
        enddo

        write(*,'(a)') 'FWHM of Gaussian energy distrib. (eV):'
        read(*,*) esigma
        write(*,'(1x,es12.5)') esigma
        esigma = esigma*fwhm2sig
        emax = espc(nspc)+kfact*esigma  ! Set max energy
        if (xkpar.eq.3) emax = emax+twomc2  ! Allowance for e+ annihilation
      endif ! End of .not.PENNUC

      ! End of section:
      read(*,'(a80)') buffer
      if (index(buffer,eos).eq.0) then
        write(*,*) 'BIGSinisrc:ERROR: End-Of-Section mark not found'
        write(*,'(a,a)') '  expecting to find: ',eos
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,'(a)') '>>>> BIGS source initialization finished >>>>'
      end


!>>>> End Of File >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
