!*******************************************************************
!*                          TALLY                                  *
!*                  PHOTON FLUENCE POINT                           *
!*                                                                 *
!* Short description:                                              *
!*   Tallies photon fluence spectrum at a point.                   *
!* Rev:                                                            *
!*   2019-08-05                                                    *
!* Comments:                                                       *
!*   This tally accesses BIGS source data and, therefore, it could *
!*   deliver biased results if the BIGS source is duplicated.      *
!*******************************************************************


!*******************************************************************
!*******************************************************************
!*    PENELOPE's 'modularized' commons section.                    *
!*    Makes PENELOPE common blocks available via module blocks for *
!*    better maintainability.                                      *
!*******************************************************************
!*******************************************************************

      module CEGRIDcommonMod
!*******************************************************************
!*    Makes available PENELOPE common blocks for better            *
!*    maintainability.                                             *
!*******************************************************************
      use PENELOPE_mod, only: negp
      implicit none
      save
      ! Energy grid and interpolation constants
      integer*4 ke
      real*8 emin,el,eu,et,dlemp,dlemp1,dlfc,xel,xe,xek
      common/cegrid/emin,el,eu,et(negp),dlemp(negp),dlemp1,dlfc,
     &  xel,xe,xek,ke
      end


      module CGIMFPcommonMod
!*******************************************************************
!*    Makes available PENELOPE common blocks for better            *
!*    maintainability.                                             *
!*******************************************************************
      use PENELOPE_mod, only: maxmat,negp
      implicit none
      save
      ! Photon simulation tables
      real*8 sgra,sgco,sgph,sgpp,sgaux
      common/cgimfp/sgra(maxmat,negp),sgco(maxmat,negp),
     &  sgph(maxmat,negp),sgpp(maxmat,negp),sgaux(maxmat,negp)
      end


      module CGRA01commonMod
!*******************************************************************
!*    Makes available PENELOPE common blocks for better            *
!*    maintainability.                                             *
!*******************************************************************
      use PENELOPE_mod, only: maxmat,negp
      implicit none
      save
      ! Rayleigh scattering
      integer*4, parameter :: nq=250,nex=1024
      integer*4 ied,ieu,ne
      real*8 ff,era,xsra
      common/cgra01/ff(maxmat,nq),era(nex),xsra(maxmat,nex),
     &  ied(negp),ieu(negp),ne
      end


      module CGRA02commonMod
!*******************************************************************
!*    Makes available PENELOPE common blocks for better            *
!*    maintainability.                                             *
!*******************************************************************
      use PENELOPE_mod, only: maxmat
      implicit none
      save
      integer*4, parameter :: nq=250
      real*8 qq,ar,br,cr,dr,ff0,qqm
      common/cgra02/qq(nq),ar(maxmat,nq),br(maxmat,nq),cr(maxmat,nq),
     &  dr(maxmat,nq),ff0(maxmat),qqm
      end


      module CGCOcommonMod
!*******************************************************************
!*    Makes available PENELOPE common blocks for better            *
!*    maintainability.                                             *
!*******************************************************************
      use PENELOPE_mod, only: maxmat
      implicit none
      save
      integer*4 noco,kzco,ksco,noscco
      real*8 fco,uico,fj0,ptrsh
      parameter (noco=512)
      common/cgco/fco(maxmat,noco),uico(maxmat,noco),fj0(maxmat,noco),
     &  ptrsh(maxmat,noco),kzco(maxmat,noco),ksco(maxmat,noco),
     &  noscco(maxmat)
      end


      module CGCO00commonMod
!*******************************************************************
!*    Makes available PENELOPE common blocks for better            *
!*    maintainability.                                             *
!*******************************************************************
      implicit none
      save
      integer*4 mm,iosc
      real*8 ee
      common/cgco00/ee,mm,iosc
      end


!*******************************************************************
!*******************************************************************
!*    Modules section                                              *
!*******************************************************************
!*******************************************************************

      module PFPmod
!*******************************************************************
!*    Vars for the PFP tally.                                      *
!*******************************************************************
      use PENELOPE_mod, only: maxmat
      implicit none
      save
      logical :: isWarnVac = .false.                     ! Init warning flag, raises if source in vacuum
      integer nbin,matSphere
      integer, parameter :: nbinmax=32000
      real*8 flutmp(nbinmax),flu(nbinmax),flu2(nbinmax)  ! Fluence counters
      real*8 emin,ebin,iebin,unclimit,rdet(3),egam,dgam(3)
      ! Exclusion sphere:
      real*8, allocatable :: flufact(:,:)
      real*8 rmin2
      end


      module PFPfieldMod
!*******************************************************************
!*    Vars related with BIGS source field of emission.             *
!*******************************************************************
      implicit none
      save
      real*8 phimaxturn,iSolidSrc
      real*8 irotv(3,3)
      end


      module PFPmuMod
!*******************************************************************
!*    Partial att coeffs provided by mu().                         *
!*******************************************************************
      implicit none
      save
      real*8 muRa,muCo,muph,mupp
      end


!*******************************************************************
!*******************************************************************
!*    Routines section                                             *
!*******************************************************************
!*******************************************************************

      subroutine PFPtally(mode,arg)
!*******************************************************************
!*    Input:                                                       *
!*      mode: identifies the state of the calling procedure        *
!*      arg: different meanings for different modes                *
!*******************************************************************
      use PFPmod
      use forcingMod, only: isforcing                 ! .true. if there is interaction forcing
      use sourceMod, only: BIGSactive
      use PFPfieldMod, only: iSolidSrc
      use PENELOPE_mod, only: eabs
      use TRACK_mod
      use PENVARED_mod, only: wforce                  ! Inverse forcing factor
      use COMPOScommonMod, only: vmol
      implicit none
      integer mode
      real*8 arg

      integer i
      integer*4 binindex
      real*8 drel(3),dist2,dist,idist,idist2,eout,costheta
      real*8 samECo,mfpCo,dcsCo,dflu,mfpRa,dcsRa
      real*8, parameter :: pi=3.1415926535897932d0,ifourpi=0.25/pi
      real*8, parameter :: eps=1.0d-13

      ! End-of-history bookkeeping:
      if (mode.eq.6) then
        do i=1,nbin                                   ! For all E bins
          if (flutmp(i).eq.0.0) cycle                 ! Skip empty counters
          flu(i)    = flu(i) +flutmp(i)               ! Transfer temporary counter to mean ...
          flu2(i)   = flu2(i)+flutmp(i)**2            ! ... and variance
          flutmp(i) = 0.0                             ! Clear counter to start a fresh history
        enddo
        return                                        ! Done
      endif

      ! Not a photon:
      if (kpar.ne.2) return                           ! Nothing to do

      ! Photon STEP completed, prepare for a KNOCK:
      if (mode.eq.3) then                             ! Occurs after STEP is done
        egam = e                                      ! Store photon energy
        dgam = (/u,v,w/)                              ! Store photon direction
        return
      endif

      ! Filter modes:
      select case(mode)
      case(-1,-2,-99)                                 ! Rayleigh, Compton or new photon from stack
        drel = rdet-(/x,y,z/)                         ! Detector relative position vector
        dist2 = sum(drel**2)                          ! Distance squared to detector location
        if (dist2.lt.rmin2) return                    ! Photon inside exclusion sphere, do nothing
        dist = sqrt(dist2)                            ! Distance to detector
        idist = 1.0/dist                              ! Normalization factor
        drel = drel*idist                             ! Normalized relative position vector
        idist2 = idist**2                             ! Inverse r2
      case default                                    ! For any other tally mode, do nothing
        return
      end select

      ! Compute emission PDF per solid angle towards detector:
      select case(mode)
      case(-1)    ! Rayleigh scattering
        eout = egam                                   ! Rayleigh is elastic
        binindex = (eout-emin)*iebin+1.0              ! Bin index for scattered photon
        if (binindex.lt.1.or.binindex.gt.nbin) return ! Not in ROI
        costheta = dot_product(dgam,drel)             ! Angle between photon direction and detector line of sight
        dflu = vmol(mat)*mfpRa(egam)*dcsRa(costheta,egam)*wght ! Unattenuated Rayleigh contribution

      case(-2)    ! Compton scattering
        costheta = dot_product(dgam,drel)             ! Angle between photon direction and detector line of sight
        eout = samECo(costheta,egam)                  ! Sample scattered E
        binindex = (eout-emin)*iebin+1.0              ! Bin index for scattered photon
        if (eout.lt.eabs(2,mat).or.                   ! Absorbed
     &      binindex.lt.1.or.binindex.gt.nbin) return ! Not in ROI
        costheta = dot_product(dgam,drel)             ! Angle between photon direction and detector line of sight
        dflu = vmol(mat)*mfpCo(egam)*dcsCo(costheta,egam)*wght  ! Unattenuated Compton contribution
        if (isforcing) dflu = dflu*wforce             ! Virtual photon is from a forced event

      case(-99)   ! New photon recovered from stack
        if (ilb(1).eq.1.and.BIGSactive) then          ! This is a source photon from BIGS model
          if (mat.eq.0) then                          ! User should avoid this, issue warning
            if (.not.isWarnVac) then                  ! Issue only once
              isWarnVac = .true.
              write(*,*) ''
              write(*,'(a)') 'PFPtally WARNING: BIGS source '//
     &          'generated particles in vacuum, which may '
              write(*,'(a)') '  bias PFP tally results. To avoid '//
     &          'this warning embed the source in a material object.'
            endif
            return                                    ! Discard
          endif
          if (isOutField()) return                    ! Detector not in source's field, no contribution
          dflu = iSolidSrc*wght                       ! Unattenuated fluence considers source solid angle
        else if (ilb(4).ne.0.or.ilb(3).eq.6) then     ! Secondary x-ray or annihilation quanta; bremss is neglected
          if (mat.eq.0) return                        ! Photon is vacuum can come from a PSF
          dflu = ifourpi*wght                         ! Assuming isotropic generation
        ! >>>> Comment
        ! The following code block scores bremss photons ignored with the tally algorithm.
        ! A finite detection sphere of cross sectional area DA is considered and each photon aiming at the
        ! sphere contributes 1/DA to the fluence. This code is included for illustrative purposes; other cases
        ! (e.g., PSF source) can be dealt with in a similar manner but they are not developed here in detail.
        !        else if (ilb(3).eq.4) then                    ! Bremss photon (ISI x-rays accounted for in previous if)
        !          rbrem2 = scale2*rmin2                       ! Bremss DetectionSphere radius = scale x ExclusionSphere
        !          if ((dist2+rbrem2)*dot_product((/u,v,w/),drel)**2.lt.dist2)
        !     &      return                                    ! Photon not aiming at detSphere
        !          dflu = wght*dist2/(pi*rbrem2)               ! Bremsss contribution; dist2 cancels 1/dist2 applied later
        ! >>>> End Comment
        else
          return                                      ! Nothing to do
        endif
        eout = e                                      ! Source energy
        binindex = (eout-emin)*iebin+1.0              ! Bin index
        if (eout.lt.eabs(2,mat).or.                   ! Absorbed
     &      binindex.lt.1.or.binindex.gt.nbin) return ! Not in ROI
      end select

      ! Apply 1/dist2 factor:
      dflu = dflu*idist2
      if (dflu.lt.eps*flutmp(binindex)) return        ! Contribution is negligible, discard

      ! Calculate exponential attenuation factor and score result:
      flutmp(binindex) = flutmp(binindex)+dflu*attexp()  ! Att computed at EOUT


      contains
      logical function isOutField()
!*******************************************************************
!*    Returns .true. if vector DREL is outside the BIGS source     *
!*    solid angle of emission.                                     *
!*******************************************************************
      use PFPfieldMod
      use BIGSmod, only: cossrc0,cossrc1,phi0,phi1
      implicit none
      real*8 dv(3),phi
      real*8, parameter :: pi=3.1415926535897932d0,twopi=2.0*pi

      isOutField = .true.                          ! Default
      dv = matmul(irotv,drel)                      ! Rotate direction vector from Slab to Sv
      if (dv(3).lt.cossrc1.or.dv(3).gt.cossrc0) return  ! costheta=dv(3) not in polar angle interval

      if (max(abs(dv(1)),abs(dv(2))).gt.0.0) then  ! Avoid atan(0,0)
        phi = atan2(dv(2),dv(1))                   ! Azimuthal angle in [-pi,+pi]
        if (phi.lt.0.0) phi=phi+twopi              ! Same in [0,+2pi)
        if ((phi.ge.phi0.and.phi.le.phi1).or.      ! In azimuthal angle interval
     &      (phi.le.phimaxturn))                   ! Allow one turn to phi
     &    isOutField = .false.
      else                                         ! Case d = +z or -z, phi irrelevant
        isOutField = .false.
      endif
      end function


      real*8 function attexp()
!*******************************************************************
!*    Photon att factor exp[-sum(mu_i*s_i)]                        *
!*******************************************************************
      use PENELOPE_mod, only: maxmat,eabs
      use PENGEOM_mod, only: dstot
      implicit none
      integer*4 bodyini,matini,ncross
      real*8 rini(3),dini(3),path,muStored(maxmat),muval,mu,dsef
      real*8, parameter :: pathInfty=-log(tiny(dist))  ! Att. paths larger than this produce exp(-path)=0 exactly

      ! Store real photon state:
      rini = (/x,y,z/)
      dini = (/u,v,w/)
      bodyini = ibody
      matini = mat

      ! Set virtual photon direction, pointing at detector:
      u = drel(1)
      v = drel(2)
      w = drel(3)

      ! Flight trough material bodies, accumulate att path:
      path = 0.0                              ! Clear att path
      muStored = -1.0                         ! Init neg value to indicate not stored
      do                                      ! For every visited material
        if (eout.lt.eabs(2,mat)) then         ! Virtual photon absorbed
          path = -1.0                         ! Flag to kill photon
          exit
        endif
        muval = muStored(mat)                 ! Reuse stored value
        if (muval.lt.0.0) then                ! Value was not stored
          muval = mu(eout,mat)                ! Invoke function
          muStored(mat) = muval               ! Store for reuse
        endif
        call step(dist,dsef,ncross)           ! Advance up to next material
        path = path+dsef*muval                ! Accumulate att path
        if (path.gt.pathInfty) then           ! Atten path effectively infinite
          path = -1.0                         ! Flag to kill photon
          exit
        endif
        dist = dist-dstot                     ! Decrease distance by amount traveled, DSTOT includes void regions
        if (ncross.eq.0.or.dist.lt.0.0) exit  ! Detector reached, done; dist<0 catches detector-in-vacuum case
        if (mat.eq.0) then                    ! Protects against a glitch in PENGEOM
          ! write(*,'(a)') 'PFPtally WARNING: PENGEOM internal error, STEP returned MAT=0.'
          call locate                         ! Try to fix the glitch
          if (mat.eq.0) then                  ! Can't fix it, return to avoid crash
            path = -1.0                       ! Flag to kill photon
            exit
          endif
        endif
      enddo

      ! Restore real photon state:
      x = rini(1)
      y = rini(2)
      z = rini(3)
      u = dini(1)
      v = dini(2)
      w = dini(3)
      ibody = bodyini
      mat = matini

      ! Attenuation factor:
      if (path.lt.0.0) then                  ! Kill virtual photon
        attexp = 0.0                         ! No contribution will be scored
      else                                   ! Return value
        attexp = exp(-path)                  ! Exponential attenuation
      endif
      end function
      end


      subroutine PFPreport(mode,n,cputim,uncdone)
!*******************************************************************
!*    Input:                                                       *
!*      mode:   -1 if called at end-of-simulation, 0 for dump only,*
!*               1 for tally report only, 2 for both.              *
!*      n:   no. of histories simulated                            *
!*      cputim:   elapsed CPU time                                 *
!*    Output:                                                      *
!*      uncdone:   2 if uncert reached, 1 if not defined, 0 else   *
!*                 not computed if mode=0                          *
!*******************************************************************
      use PFPmod
      use PENELOPE_mod
      use RSEEDcommonMod
      implicit none
      integer mode,uncdone
      real*8 n,cputim

      logical isWarnQadd
      integer i,nchan,out,finduf,errorcode
      real*8 q,q2,sigma,eff,avesig,maxq,emiddle,elow,uncert,invn
      real*8 qadd,qaddFrac
      real, parameter :: warnQadd=0.03         ! Issue a warning above 3%

      uncdone = 1

      if (mode.ne.1) call PFPdump(1)  ! Sim dump file
      if (mode.eq.0) return           ! Dump only, no report

      ! Prepare output files:
      out = finduf()
      open(out,file='tallyPhotonFluencePoint.dat',
     &  iostat=errorcode)
      if (errorcode.ne.0) then
        write(*,*)
        write(*,'(a)')
     &    '*********************************************'
        write(*,'(a)')
     &    'PFPreport:ERROR: cannot open output data file'
        write(*,'(a)')
     &    '*********************************************'
        close(out)  ! Just in case
        return
      endif

      write(out,'(a)')
     &'#>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      write(out,'(a)')
     & '# [SECTION REPORT PHOTON FLUENCE SPECTRUM AT A POINT]'
      write(out,'(a)') '# Fluence units are 1/(cm^2.eV) per history'
      write(out,'(a)') '#'
      write(out,'(a)') '# Number of energy bins:'
      write(out,'(a,i0)') '#  ',nbin
      write(out,'(a)') '# Energy bin width (eV):'
      write(out,'(a,es12.5)') '#  ',ebin
      write(out,'(a)') '#'
      write(out,'(a)') '# For plotting purposes, '//
     &  'two values per energy bin are given, '//
     &  'namely, the low end and the middle point of each bin.'
      write(out,'(a)') '#'
      write(out,'(a)') '# The approx. contribution from '//
     &  'the exclusion sphere is reported separately.'
      write(out,'(a)') '#'
      write(out,'(a)') '# Elow(eV) : Emiddle(eV) : fluence : '//
     &  '+-2sigma  ::  +fluenceSphere (%total) : totalFluence : nbin'

      avesig = 0.0
      nchan = 0
      invn = 1.0/n
      maxq = 0.5*maxval(flu(1:nbin))  ! 1/2 of the max score
      isWarnQadd = .false.            ! No warning by default

      do i=1,nbin
        q  = flu(i)*invn
        q2 = flu2(i)*invn
        sigma = sqrt(max((q2-q**2)*invn,0.0))
        q = q*iebin
        sigma = sigma*iebin
        elow = emin+ebin*(i-1)                   ! Bin low end
        emiddle = elow+ebin*0.5                  ! Bin middle point

        ! Compute contribution from the exclusion sphere:
        qadd = 0.0
        qaddFrac = 0.0
        if (matSphere.ne.0)                      ! Sphere in vacuum, no contribution
     &    qadd = invn*iebin*dot_product(flufact(i,:),flu(1:nbin)) ! Sphere contribution as normalized matrix product
        if (qadd.gt.0.0) qaddFrac = qadd/(q+qadd) ! Sphere relative contribution
        write(out,'(3(1x,es12.5),1x,es8.1,a,'//
     &    'es12.5,1x,f5.1,a,1x,es12.5,1x,i0,$)')
     &    elow,emiddle,q,2.0*sigma,'  ::  ',
     &    qadd,100.0*qaddFrac,' %',q+qadd,i
        if (qaddFrac.gt.warnQadd) then           ! Too large
          write(out,'(a)') ' (*)'                ! Mark bin with warning sign
          isWarnQadd = .true.                    ! Flag to issue a warning at EOF
        else
          write(out,*) ''                        ! No mark, set EOL
        endif

        ! Evaluate average uncertainty for scores above 1/2 max score:
        if (flu(i).gt.maxq) then
          avesig = avesig+(sigma/q)**2
          nchan = nchan+1
        endif
      enddo

      uncdone = 0
      if (nchan.gt.0) then
        uncert = 200.0*sqrt(avesig/nchan)
        if (uncert.lt.unclimit) uncdone = 2  ! Uncertainty reached
      else
        uncert = 0.0  ! Uncertainty assumed not reached when score is nil
      endif

      ! Generic report:
      if (isWarnQadd) then
        write(out,'(a)') ' '
        write(out,'(a,f3.1,a)') '# WARNING: Exclusion sphere '//
     &    'contribution larger than ',100.0*warnQadd,'% for bins '//
     &    'marked with (*).'
      endif
      write(out,'(a)') ' '
      write(out,'(a)') '# Performance report'
      write(out,'(a)') '#   Random seeds:'
      write(out,'(a,i10)') '#   ',seed1
      write(out,'(a,i10)') '#   ',seed2
      write(out,'(a)') '#   No. of histories simulated [N]:'
      write(out,'(a,f18.0)') '#   ',n
      write(out,'(a)') '#   CPU time [t] (s):'
      write(out,'(a,es12.5)') '#   ',cputim
      if (cputim.gt.0.0) then
        write(out,'(a)') '#   Speed (histories/s):'
        write(out,'(a,es12.5)') '#   ',n/cputim
      endif
      write(out,'(a)')
     & '#   Average uncertainty (above 1/2 max score) in % [uncert]:'
      write(out,'(a,es12.5)') '#   ',uncert
      eff = n*uncert**2
      if (eff.gt.0.0) then
        write(out,'(a)') '#   Intrinsic efficiency [N*uncert^2]^-1:'
        write(out,'(a,es12.5)') '#   ',1.0/eff
      endif
      eff = cputim*uncert**2
      if (eff.gt.0.0) then
        write(out,'(a)') '#   Absolute efficiency [t*uncert^2]^-1:'
        write(out,'(a,es12.5)') '#   ',1.0/eff
      endif
      close(out)
      end


      subroutine PFPinitally(active,mem,pemax)
!*******************************************************************
!*    Initializes. To be called before TALLY.                      *
!*                                                                 *
!*    Output:                                                      *
!*      active: TRUE if the tally is ON.                           *
!*      mem: Memory used by tally arrays                           *
!*      pemax: max source energy (eV).                             *
!*******************************************************************
      use PFPmod
      use PFPfieldMod
      use PFPmuMod
      use sourceMod, only: BIGSactive
      use BIGSmod, only: rotv,dcossrc,phi1,dphi,xkpar
      use dataTypesMod
      use TRACK_mod, only: x,y,z,u,v,w,mat  ! To locate detector MAT
      use PENELOPE_mod, only: maxmat,eabs,nmat
      use geovoxMod, only: bodymask         ! To check if voxels active
      implicit none
      logical active
      real*8 mem,pemax

      character*(*) secid,eos
      parameter (secid=
     &  '[SECTION TALLY PHOTON FLUENCE POINT v.2015-05-31]')
      parameter (eos='[END OF PFP SECTION]')
      character*80 buffer,upcase
      integer errorcode
      integer*4 i,j,binpp
      real*8 emax,gmfp,prange,e,ei,ej,ieqj,memalloc
      real*8 mu,muval,attfactmc2,csKN,dcsKNe,radius,path
      real*8, allocatable :: attfact(:),factRa(:),factCo(:),factpp(:)
      real*8, parameter :: pi=3.1415926535897932d0,twopi=2.0*pi
      real*8, parameter :: mc2=5.10998928d5    ! Electron rest energy (eV)
      real*8, parameter :: eps=1.0d-3,zero=1.0d-30,zeropath=1.0d-10

      write(*,*) ' '
      write(*,'(a)')
     & '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      call getline(buffer,0)
      if (index(buffer,secid).eq.0) then
        write(*,'(a)') 'PFPinitally:ERROR: incorrect section header;'
        write(*,'(a,a)') '  expecting to find: ',secid
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,'(a)') secid

      ! Report memory usage:
      mem = sizeOfReal8*3*real(nbinmax)
      write(*,'(a)') 'Memory used by static arrays (MB):'
      write(*,'(1x,es12.5)') mem*1.0e-6

      read(*,'(a80)') buffer
      buffer = adjustl(upcase(buffer))
      buffer = buffer(1:scan(buffer,' ')) ! Clip at 1st blank
      if (buffer.eq.'ON') then
        active = .true.
      else if (buffer.eq.'OFF') then
        active = .false.
        write(*, '(a)') '>>>> Tally Photon Fluence Point is OFF >>>>'
        do
          read(*,'(a80)',iostat=errorcode) buffer
          if (errorcode.ne.0) then
            write(*,'(a,a,a)') 'PFPinitally:ERROR: ',
     &       'Unable to find End-Of-Section mark: ',eos
            stop
          endif
          if (index(buffer,eos).ne.0) return
        enddo
      else
        write(*,'(a)')
     &    'PFPinitally:ERROR: expecting to find ON or OFF'
        write(*,'(a)') 'found instead:'
        write(*,'(a)') buffer
        stop
      endif

      ! Check if voxels are defined:
      if (bodymask.ne.-1) then
        write(*,'(a)') 'PFPinitally:ERROR: a voxelized geometry '//
     &    'is defined; this implementation of PFP is not '//
     &    'compatible with voxels.'
        stop
      endif

      ! Check if electron transport is active:
      if (any(eabs(1,1:nmat).lt.pemax).or.
     &    any(eabs(3,1:nmat).lt.pemax)) then
        write(*,'(a)')
     &    '**************************************************'
        write(*,'(a)')
     &    'PFPinitally:WARNING: electron transport is active;'
        write(*,'(a)')
     &    'contribution to fluence from bremsstrahlung emission'
        write(*,'(a)')
     &    'is not accounted for when the photon is emitted.'
        write(*,'(a)')
     &    '**************************************************'
      endif

      ! Debug att coeffs:
      read(*,'(a80)') buffer
      buffer = adjustl(upcase(buffer))
      buffer = buffer(1:scan(buffer,' ')) ! Clip at 1st blank
      if (buffer.eq.'ON') then
        call writeDebugFiles(pemax)
      else if (buffer.ne.'OFF') then
        write(*,'(a)')
     &    'PFPinitally:ERROR: expecting to find ON or OFF'
        write(*,'(a)') 'found instead:'
        write(*,'(a)') buffer
        stop
      endif

      ! Detector position:
      write(*,'(a)') 'Detector position coordinates (cm):'
      read(*,*) x,y,z
      write(*,'(3(1x,es12.5))') x,y,z
      rdet = (/x,y,z/)
      write(*,'(a)') 'Detector located in material index:'
      u = 0.0
      v = 0.0
      w = 1.0
      call locate            ! Not usable for voxels
      write(*,'(1x,i0)') mat
      matSphere = mat        ! Assume sphere is homogeneous

      ! Exclusion sphere:
      write(*,'(a)') 'Exclusion sphere radius around detector (cm):'
      read(*,*) rmin2

      if (rmin2.lt.0.0) then
        write(*,'(a)') '  (negative radius was entered; '//
     &    'interpreted as a multiple of the mean free path, MFP)'
        if (mat.eq.0) then
          write(*,'(a)')
     &      'PFPinitally:ERROR: source in vacuum, MFP is infinite.'
          stop
        endif
        gmfp = prange(pemax,2,mat)
        rmin2 = (-rmin2)*gmfp
        write(*,'(1x,es12.5)') rmin2
        write(*,'(a)') 'MFP (cm) for detector material at max '//
     &    'source energy:'
        write(*,'(1x,es12.5)') gmfp
      else
        write(*,'(1x,es12.5)') rmin2
      endif

      if (rmin2.gt.zero) then
        rmin2 = rmin2**2   ! Square of sphere radius
      else
        rmin2 = zero**2
        if (mat.ne.0) then
          write(*,*) ''
          write(*,'(a)') 'PFPinitally:WARNING: the radius of the '//
     &      'exclusion sphere is set to zero;'
          write(*,'(a)') '  due to the divergence at the detector '//
     &      'position, reported statistical uncertainties '//
     &      'are meaningless.'
          write(*,*) ''
        endif
        matSphere = 0  ! Set to avoid computing sphere contribution
      endif

      ! Spectrum ROI:
      write(*,'(a)')
     &  'Spectrum limits [Emin,Emax] (eV) and no. bins:'
      read(*,*) emin,emax,nbin
      write(*,'(2(1x,es12.5),1x,i0)') emin,emax,nbin
      if (emax-emin.lt.eps.or.nbin.lt.1) then
        write(*,'(a)') 'PFPinitally:ERROR: invalid entry.'
        stop
      endif
      ebin = (emax-emin)/nbin
      iebin = 1.0/ebin
      ! Add one bin to store, separately, counts with E=Emax:
      nbin = nbin+1
      if (nbin.gt.nbinmax) then
        write(*,'(a)') 'PFPinitally:ERROR: Too many bins.'
        write(*,'(a,i0)') '  Max no. bins is ',nbinmax-1
        stop
      endif

      ! Uncertainty:
      write(*,'(a)') 'Relative uncertainty (%) requested:'
      read(*,*) unclimit
      write(*,'(1x,es12.5)') unclimit

      ! Clear counters:
      flutmp = 0.0
      flu    = 0.0
      flu2   = 0.0

      ! Set constant vars for BIGS source field of emission:
      if (BIGSactive) then
        irotv = transpose(rotv)       ! Inverse of BIGS rotv, rotates from Slab to Sv
        phimaxturn = phi1-twopi       ! Max azimuthal angle less one turn
        iSolidSrc = abs(dcossrc)*dphi  ! Solid angle subtended by BIGS field of emission
        if (iSolidSrc.lt.zero.and.(xkpar.eq.2.or.xkpar.eq.99)) then  ! 99=PENNUC
          write(*,'(a)')
     &     'PFPinitally:ERROR: BIGS emits a collimated (zero solid '//
     &     ' angle aperture) photon beam.'
          write(*,'(a)') 'This tally cannot determine the '//
     &     'contribution from unscattered photons emitted '//
     &     'by the source.'
          write(*,'(a)') 'A workaround is to define the '//
     &     'direction of emission with a small non-null solid '//
     &     'angle aperture.'
          stop
          ! For an arbitrary shaped volumetric source emitting a collimated beam,
          ! if the beam POINTS at the detector D the contribution to the fluence
          ! per history from newborn photons is
          !   Phi = len(C)/V
          ! where V is the total source volume and len(C) is the length of the
          ! chord C inside V that points at D.
        endif
      else                                    ! No need to define minmax values, conditions won't be tested
        write(*,'(a)') '***************'
        write(*,'(a)')
     &    'PFPinitally:WARNING: BIGS source is not active;'
        write(*,'(a)') '  fluence contributions from newborn '//
     &    'photons are not accounted for with other source models.'
        write(*,'(a)') '***************'
      endif
      iSolidSrc = 1.0/iSolidSrc               ! Inverse solid angle

      ! Compute exclusion sphere matrix:
      if (matSphere.ne.0) then                        ! Else, no contribution from sphere
        ! Allocate arrays:
        memalloc = sizeOfReal8*real(nbin)*real(nbin)  ! Only flufact() is permanent
        write(*,'(a)') 'Memory used to allocate arrays (MB):'
        write(*,'(1x,es12.5)') memalloc*1.0e-6
        mem = mem+memalloc
        allocate(flufact(nbin,nbin),attfact(nbin),factRa(nbin),
     &    factCo(nbin),factpp(nbin),stat=errorcode)
        if (errorcode.ne.0) then
          write(*,'(a)') 'PFPinitally:ERROR: not enough memory.'
          stop
        endif

        ! Clear arrays:
        flufact = 0.0
        attfact = 0.0
        factRa = 0.0
        factCo = 0.0
        factpp = 0.0
        radius = sqrt(rmin2)                            ! Exclusion sphere radius

        ! Compute aux arrays:
        do i=1,nbin                                     ! For all energy bins
          e = emin+ebin*(i-0.5)                         ! Bin middle point
          if (e.gt.pemax) exit                          ! E above source max, avoid mu() out-of-range
          if (e.lt.eabs(2,matSphere)) cycle             ! Below Eabs, avoid mu() out-of-range
          muval = mu(e,matSphere)                       ! Total attCoeff, also sets partial coeffs (muRa,muCo,mupp) via module
          path = muval*radius                           ! Attenuation path
          if (path.gt.zeropath) then                    ! AttPath not negligible
            attfact(i) = (1.0-exp(-path))/muval         ! Attenuation factor
          else                                          ! AttPath << 1
            attfact(i) = radius                         ! Small path approx
          endif
          factRa(i) = muRa
          factCo(i) = muCo*ebin/csKN(e)                 ! csKN(E0) combined with dcsKNe(E;E0) below gives pKN(E;E0)
          factpp(i) = mupp
        enddo

        muval = 0.0                                     ! Set to 0 in case mc2=511k is out of dynamic range
        if (mc2.ge.eabs(2,matSphere).and.mc2.le.pemax) then  ! Check if mc2 in range, avoid mu() out-of-range
          muval = mu(mc2,matSphere)                     ! Att coeff at mc2
        endif

        path = muval*radius                             ! Attenuation path
        if (path.gt.zeropath) then                      ! AttPath not negligible
          attfactmc2 = 2.0*(1.0-exp(-path))/muval       ! Att factor at 511k; 2.0 accounts for 2 annihil photons
        else                                            ! AttPath << 1
          attfactmc2 = radius                           ! Small path approx   CHECK path=0
        endif
        binpp = (mc2-emin)*iebin+1.0                    ! Ebin for 511k; it may not be in [1,nbin]

        ! Fill fluence factor matrix:
        do i=1,nbin
          ei = emin+ebin*(i-0.5)                        ! Scattered photon energy
          flufact(i,i) = attfact(i)*factRa(i)           ! Rayleigh contribution
          ieqj = 0.5                                    ! Bin Ej=Ei contributes only half
          do j=i,nbin                                   ! Only Ej>=Ei contribute
            ej = emin+ebin*(j-0.5)                      ! Primary photon energy
            flufact(i,j) = flufact(i,j)+
     &        ieqj*attfact(i)*factCo(j)*dcsKNe(ei,ej)   ! Compton contribution
            ieqj = 1.0                                  ! Cancel 1/2 factor for j>i
          enddo
          if (i.eq.binpp) flufact(i,:)=flufact(i,:)+
     &      attfactmc2*factpp(:)                        ! PairProd contribution
        enddo

        deallocate(attfact,factRa,factCo,factpp)        ! Free memory
      endif

      ! Finish:
      read(*,'(a80)') buffer
      if (index(buffer,eos).eq.0) then
        write(*,*) 'PFPinitally:ERROR: End-Of-Section mark not found'
        write(*,'(a,a)') '  expecting to find: ',eos
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif

      ! Sim restart file:
      call PFPdump(0)

      write(*,'(a)') '>>>> PFP tally initialization finished >>>>'
      end


      real*8 function mu(e,mat)
!*******************************************************************
!*    Photon att coeff (cm-1).                                     *
!*    This is an adaptation of routine PRANGE in PENELOPE.         *
!*                                                                 *
!*    Input:                                                       *
!*      e: photon energy (eV)                                      *
!*      mat: material index                                        *
!*    Comments:                                                    *
!*    - Partial attCoeffs are saved in module.                     *
!*    - Extended CSs for Rayleigh and photoelectric are also       *
!*      included but commented out. These CSs, which include delta *
!*      interactions, are provided here but not used by the tally. *
!*******************************************************************
      use PFPmuMod
      use COMPOScommonMod
      use CEGRIDcommonMod
      use CGIMFPcommonMod
      use CGRA01commonMod
      implicit none
      integer*4 mat
      real*8 e

      integer*4 ii,iu,it
      real*8, parameter :: twomc2p=1.022e6  ! A little above 2mc2 (eV)

      xel = log(e)
      xe  = 1.0d0+(xel-dlemp1)*dlfc
      ke  = xe                    ! Abscissa's index ...
      if(ke.lt.1.or.ke.ge.negp) then  ! E out of range, bug in calling routine
        write(*,'(a,i0,1x,i0)')
     &    'mu:ERROR: internal error, index out of range: ',ke,negp
        stop
      endif
      xek = xe-ke                 ! ... and fractional 'excess'

      ! Rayleigh:
      ii = ied(ke)
      iu = ieu(ke)
      do                          ! Binary search to find data array index
        it = (ii+iu)/2
        if(xel.gt.era(it)) then
          ii = it
        else
          iu = it
        endif
        if(iu-ii.lt.2) exit       ! Index found
      enddo
      muRa = exp(xsra(mat,ii)+(xsra(mat,ii+1)-xsra(mat,ii))*
     &      (xel-era(ii))/(era(ii+1)-era(ii)))  ! Linear log-log interpol
      ! Rayleigh extended CS including delta interactions (not used):
      ! muRa = sgra(mat,ke)

      ! Compton:
      muCo = exp(sgco(mat,ke)+(sgco(mat,ke+1)-sgco(mat,ke))*xek) ! Compton: linear log-log interpol

      ! Photoelectric:
      call gphat(e,muph,mat)
      muph = muph*vmol(mat)
      ! Photoelectric extended CS including delta interactions (not used):
      ! muph = sgph(mat,ke)

      ! Pair:
      if(e.lt.twomc2p) then     ! No pp below threshold
        mupp = 0.0
      else
        mupp = exp(sgpp(mat,ke)+(sgpp(mat,ke+1)-sgpp(mat,ke))*xek)
      endif

      ! Total att coeff:
      mu = muRa+muCo+muph+mupp
      end


      real*8 function mfpRa(e)
!*******************************************************************
!*    Rayleigh mean free path (cm).                                *
!*    Adapted from routine PRANGE in PENELOPE.                     *
!*                                                                 *
!*    Input:                                                       *
!*      e: photon energy (eV)                                      *
!*      mat: material index taken from TRACK mod                   *
!*    Comments:                                                    *
!*      - Out-of-range protection logic in the original PRANGE has *
!*        has been removed to save CPU time. This routine must not *
!*        be called with E < EABS or E > Emax from source.         *
!*******************************************************************
      use TRACK_mod, only: mat
      use CEGRIDcommonMod
      use CGRA01commonMod
      implicit none
      real*8 e

      integer*4 ii,iu,it

      xel = log(e)
      xe  = 1.0d0+(xel-dlemp1)*dlfc
      ke  = xe                    ! Abscissa's index ...
      xek = xe-ke                 ! ... and fractional 'excess'

      ! Rayleigh:
      ii = ied(ke)
      iu = ieu(ke)
      do                          ! Binary search to find data array index
        it = (ii+iu)/2
        if(xel.gt.era(it)) then
          ii = it
        else
          iu = it
        endif
        if(iu-ii.lt.2) exit       ! Index found
      enddo
      mfpRa = exp(-xsra(mat,ii)-(xsra(mat,ii+1)-xsra(mat,ii))*
     &      (xel-era(ii))/(era(ii+1)-era(ii)))  ! Linear log-log interpol
      end


      real*8 function mfpCo(e)
!*******************************************************************
!*    Compton mean free path (cm).                                 *
!*    Adapted from routine PRANGE in PENELOPE.                     *
!*                                                                 *
!*    Input:                                                       *
!*      e: photon energy (eV)                                      *
!*      mat: material index taken from TRACK mod                   *
!*    Comments:                                                    *
!*      - Out-of-range protection logic in the original PRANGE has *
!*        has been removed to save CPU time. This routine must not *
!*        be called with E < EABS or E > Emax from source.         *
!*******************************************************************
      use TRACK_mod, only: mat
      use CEGRIDcommonMod
      use CGIMFPcommonMod
      implicit none
      real*8 e

      xe  = (log(e)-dlemp1)*dlfc+1.0
      ke  = xe                    ! Abscissa's index ...
      xek = xe-ke                 ! ... and fractional 'excess'
      mfpCo = exp(-sgco(mat,ke)-(sgco(mat,ke+1)-sgco(mat,ke))*xek)
      end


      real*8 function muppf(e)
!*******************************************************************
!*    Pair prod. att coeff (cm-1).                                 *
!*    Adapted from routine PRANGE in PENELOPE.                     *
!*                                                                 *
!*    Input:                                                       *
!*      e: photon energy (eV)                                      *
!*      mat: material index taken from TRACK mod                   *
!*    Comments:                                                    *
!*      - Out-of-range protection logic in the original PRANGE has *
!*        has been removed to save CPU time. This routine must not *
!*        be called with E < EABS or E > Emax from source.         *
!*******************************************************************
      use TRACK_mod, only: mat
      use CEGRIDcommonMod
      use CGIMFPcommonMod
      implicit none
      real*8 e
      real*8, parameter :: twomc2p=1.022e6  ! A little above 2mc2 (eV)

      xe  = (log(e)-dlemp1)*dlfc+1.0
      ke  = xe                    ! Abscissa's index ...
      xek = xe-ke                 ! ... and fractional 'excess'
      if(e.lt.twomc2p) then     ! No pp below threshold
        muppf = 0.0
      else
        muppf = exp(sgpp(mat,ke)+(sgpp(mat,ke+1)-sgpp(mat,ke))*xek)
      endif
      end


      real*8 function muKN(e,mat)
!*******************************************************************
!*    Klein-Nishina att coeff (cm-1).                              *
!*    Used only for debug purposes.                                *
!*                                                                 *
!*    Input:                                                       *
!*      e: photon energy (eV)                                      *
!*      mat: material index                                        *
!*******************************************************************
      use COMPOScommonMod, only: zt,vmol
      implicit none
      integer*4 mat
      real*8 e,csKN

      muKN = vmol(mat)*zt(mat)*csKN(e)
      end


      real*8 function dcsRa(costheta,e)
!*******************************************************************
!*    DCS per unit solid angle for Rayleigh scattering (cm2/sr).   *
!*    This is an adaptation of routine GRAaD in PENELOPE.          *
!*                                                                 *
!*    Input:                                                       *
!*      costheta: cos(theta), scattering angle                     *
!*      e: photon energy (eV)                                      *
!*      mat: material index taken from TRACK mod                   *
!*******************************************************************
      use TRACK_mod, only: mat
      use CGRA02commonMod
      implicit none
      real*8 costheta,e

      integer*4 i
      real*8 ql,q2,formfact
      real*8, parameter :: mc2=5.10998928d5    ! Electron rest energy (eV)
      real*8, parameter :: imc4=2.0/mc2**2
      real*8, parameter :: re=2.8179403267d-13 ! Classical electron radius (cm)
      real*8, parameter :: halfre2 = re**2/2.0

      q2 = imc4*(1.0-costheta)*e**2 ! =(q/mc)^2 eq.2.14v2011; not valid if E>50MeV in original GRAaD
      if(q2.gt.qqm) then            ! Very large momentum transfer
        dcsRa = 0.0
        return
      endif
      if(q2.lt.1.0d-9) then         ! Very small momentum transfer
        formfact = ff0(mat)
      else                          ! Interpolate form factor (routine GRAaF2() in PENELOPE):
        ql = log(q2)                ! Cubic spline log-log interpolation
        call findi(qq,ql,nq,i)
        formfact = exp(ar(mat,i)+
     &                 ql*(br(mat,i)+ql*(cr(mat,i)+ql*dr(mat,i))))
      endif

      dcsRa = halfre2*(1.0+costheta**2)*formfact
      end


      real*8 function dcsCo(costheta,e)
!*******************************************************************
!*    DCS per unit solid angle for Compton scattering (cm2/sr).    *
!*    Based on PENELOPE's GCOaD() routine, which uses the RIA model*
!*    with Doppler broadening and binding.                         *
!*                                                                 *
!*    Input:                                                       *
!*      costheta: cos(theta), scattering angle                     *
!*      e: photon energy (eV)                                      *
!*      mat: material index taken from TRACK mod                   *
!*******************************************************************
      use TRACK_mod, only: mat
      use CGCOcommonMod, only: noscco,fco  ! No. shells and no. e- in each shell
      use CGCO00commonMod
      implicit none
      real*8 costheta,e

      integer i
      real*8 GCOaD
      real*8, parameter :: re=2.8179403267d-13 ! Classical electron radius (cm)
      real*8, parameter :: halfre2 = re**2/2.0

      ee = e                    ! Input energy for GCOaD()
      mm = mat                  ! Input material for GCOaD()

      dcsCo = 0.0               ! Init
      do i=1,noscco(mat)        ! For all shells in material
        iosc = i                ! Input shell for GCOaD()
        dcsCo = dcsCo+fco(mat,i)*GCOaD(costheta)  ! Add shell contribution
      enddo
      dcsCo = halfre2*dcsCo
      end


      real*8 function dcsKN(costheta,e)
!*******************************************************************
!*    Klein-Nishina DCS per e- diff in solid angle (cm2/sr).       *
!*    Used only for debug purposes.                                *
!*                                                                 *
!*    Input:                                                       *
!*      costheta: cos(theta), scattering angle                     *
!*      e: photon energy (eV)                                      *
!*******************************************************************
      implicit none
      real*8 costheta,e

      real*8, parameter :: mc2=5.10998928d5,imc2=1.0/mc2  ! Electron rest energy (eV)
      real*8, parameter :: re=2.8179403267d-13            ! Classical electron radius (cm)
      real*8, parameter :: halfre2 = re**2/2.0
      real*8 EoverEc,EcOverE

      EoverEc = 1.0+e*imc2*(1.0-costheta)
      EcOverE = 1.0/EoverEc                ! Compton line
      dcsKN = halfre2*EcOverE**2*(EcOverE+EoverEc+costheta**2-1.0)
      end


      real*8 function dcsKNe(eout,e)
!*******************************************************************
!*    Klein-Nishina DCS per e- diff in energy (cm2/eV).            *
!*                                                                 *
!*    Input:                                                       *
!*      eout: energy (eV) of the scattered photon                  *
!*      e: energy (eV) of the incoming photon                      *
!*******************************************************************
      implicit none
      real*8 eout,e

      real*8 e0,e1,ie0,ie1,die
      real*8, parameter :: pi=3.1415926535897932d0
      real*8, parameter :: mc2=5.10998928d5,imc2=1.0/mc2  ! Electron rest energy (eV)
      real*8, parameter :: re=2.8179403267d-13            ! Classical electron radius (cm)
      real*8, parameter :: pirmc2=pi*re*re*imc2

      e0 = e*imc2
      e1 = eout*imc2
      ie0 = 1.0/e0
      ie1 = 1.0/e1
      die = ie1-ie0   ! = mc2/E' - mc2/E
      if (die.lt.0.0.or.die.gt.2.0) then  ! Eout out of range costheta=+-1
        dcsKNe = 0.0
        return
      endif
      dcsKNe = pirmc2*ie0**2 *
     &         (e1*ie0 + e0*ie1 + (1.0-die)**2 - 1.0)
      end


      real*8 function csKN(e)
!*******************************************************************
!*    Total Klein-Nishina cross section per e- (cm2).              *
!*                                                                 *
!*    Input:                                                       *
!*      e: photon energy (eV)                                      *
!*******************************************************************
      implicit none
      real*8 e

      real*8, parameter :: mc2=5.10998928d5,imc2=1.0/mc2  ! Electron rest energy (eV)
      real*8, parameter :: pi=3.1415926535897932d0
      real*8, parameter :: re=2.8179403267d-13,twopire2=2.0*pi*re**2 ! Classical electron radius (cm)
      real*8 e0,ie0,ep1,ep2,iep2,lnep2

      e0  = e*imc2       ! E in mc2 units
      ie0 = 1.0/e0       ! 1/E
      ep1 = 1.0+e0       ! 1+E
      ep2 = ep1+e0       ! 1+2E
      iep2 = 1.0/ep2     ! 1/(1+2E)
      lnep2 = log(ep2)   ! log(1+2E)
      csKN = twopire2*(ep1*ie0*ie0*(2.0*ep1*iep2-ie0*lnep2)+
     &       0.5*ie0*lnep2-iep2*iep2*(ep2+e0))
      end


      real*8 function samECo(costheta,e)
!*******************************************************************
!*    Samples a scattered photon energy (eV) for a given Compton   *
!*    scattering angle using the Compton line relation. Doppler    *
!*    and binding effects are not considered.                      *
!*                                                                 *
!*    Input:                                                       *
!*      costheta: cos(theta), scattering angle                     *
!*      e: photon energy (eV)                                      *
!*      mat: material index taken from TRACK mod                   *
!*******************************************************************
      implicit none
      real*8 costheta,e
      real*8, parameter :: mc2=5.10998928d5,imc2=1.0/mc2 ! Electron rest energy (eV)

      samECo = e/(1.0+e*imc2*(1.0-costheta))             ! Compton line
      end


      subroutine writeDebugFiles(emax)
!*******************************************************************
!*    Writes attenuation coefficients and DCSs to files for debug. *
!*                                                                 *
!*    Input:                                                       *
!*      emax: Max source energy (eV), used for tabulations.        *
!*******************************************************************
      use PFPmuMod
      use TRACK_mod, only: e,mat
      use PENELOPE_mod, only: eabs,nmat
      use COMPOScommonMod, only: rho,zt  ! Mass dens & no. e- in molecule
      implicit none
      real*8 emax

      character*80 buffer
      integer i,j,funit,finduf
      integer, parameter :: ne=500,ned=10 ! No. of energies for tables
      real*8 lemin,dle,fact,muval,mu,angle,escat
      real*8 dcs,dcsRa,dcsCo,dcsKN,dcs2,muKN,dcsKNe
      real*8, parameter :: pi=3.1415926535897932d0,deg2rad=pi/180.0
      real*8, parameter :: zero=1.0d-30

      write(*,'(a)') 'DEBUG is ON, writing data files...'

      do mat=1,nmat                    ! For all materials in simulation
        ! Compute tables energies:
        lemin = log(eabs(2,mat))       ! Use log scale
        dle = (log(emax)-lemin)/ne
        if (dle.lt.zero) then
          write(*,'(a,i0,a)') ' Could not write data for material ',
     &      mat,'; EABS>Esource, material is absorbent.'
          cycle                       ! Pick next material
        endif

        ! Att coeffs:
        write(buffer,'(a,i3.3,a)') 'tallyPhotonFluencePoint-'//
     &    'debugAttCoeffs-mat',mat,'.dat'
        write(*,'(a,a)') ' Writing file ',trim(buffer)
        funit = finduf()
        open(unit=funit,file=buffer)
        write(funit,'(a)') '#### Debug file: '//
     &    'Mass attenuation coeffs, mu (cm2/g) ####'
        write(funit,'(a)') '#'
        write(funit,'(a)') '# i : E(eV) : muRayleigh : muCompton : '//
     &    '(muKleinNishina: ) muph : mupp : muTot'
        write(funit,'(a)') '# ------------------------------------'//
     &    '--------------------------------------'
        fact = 1.0/rho(mat)
        do i=1,ne
          e = exp(lemin+(i-0.5)*dle) ! log scale
          muval  =  mu(e,mat)
          write(funit,'(1x,i3,7(1x,es13.5e3))') i,e,muRa*fact,
     &      muCo*fact,muKN(e,mat)*fact,muph*fact,mupp*fact,muval*fact
        enddo
        write(funit,'(a)') '#### EOF ####'
        close(funit)

        ! Recompute energy tables:
        dle = (log(emax)-lemin)/ned

        ! Rayleigh DCS:
        write(buffer,'(a,i3.3,a)') 'tallyPhotonFluencePoint-'//
     &    'debugRayleighDCS-mat',mat,'.dat'
        write(*,'(a,a)') ' Writing file ',trim(buffer)
        funit = finduf()
        open(unit=funit,file=buffer)
        write(funit,'(a)') '#### Debug file: Rayleigh DCS (cm2/sr) '//
     &    '####'
        write(funit,'(a)') '#'
        write(funit,'(a)')
     &    '# Eindex : E(eV) : angIndex : angle(deg) : DCS'
        write(funit,'(a)')
     &    '# --------------------------------------------'
        do i=1,ned
          e = exp(lemin+(i-0.5)*dle)
          do j=1,360             ! Every half degree
            angle = (j-0.5)*0.5
            dcs = dcsRa(cos(angle*deg2rad),e)
            write(funit,'(1x,i3,1x,es13.5e3,1x,i3,2(1x,es13.5e3))')
     &        i,e,j,angle,dcs
          enddo
          write(funit,'(a)') ''
          write(funit,'(a)') ''  ! Two blanks inserted to separate data blocks
        enddo
        write(funit,'(a)') '#### EOF ####'
        close(funit)

        ! Compton DCS per solid angle:
        write(buffer,'(a,i3.3,a)') 'tallyPhotonFluencePoint-'//
     &    'debugComptonDCS-mat',mat,'.dat'
        write(*,'(a,a)') ' Writing file ',trim(buffer)
        funit = finduf()
        open(unit=funit,file=buffer)
        write(funit,'(a)') '#### Debug file: Compton DCS (cm2/sr) '//
     &    '####'
        write(funit,'(a)') '#'
        write(funit,'(a)')
     &    '# Eindex : E(eV) : angIndex : angle(deg) : DCS : '//
     &    'DCSKleinNishina'
        write(funit,'(a)')
     &    '# -----------------------------------------------'//
     &    '---------------'
        do i=1,ned
          e = exp(lemin+(i-0.5)*dle)
          do j=1,360             ! Every half degree
            angle = (j-0.5)*0.5
            dcs = dcsCo(cos(angle*deg2rad),e)
            dcs2 = zt(mat)*dcsKN(cos(angle*deg2rad),e)  ! ZT is no. e- per molecule in mat
            write(funit,'(1x,i3,1x,es13.5e3,1x,i3,3(1x,es13.5e3))')
     &        i,e,j,angle,dcs,dcs2
          enddo
          write(funit,'(a)') ''
          write(funit,'(a)') ''  ! Two blanks inserted to separate data blocks
        enddo
        write(funit,'(a)') '#### EOF ####'
        close(funit)
      enddo

      ! Recompute energy tables:
      lemin = log(100.0)                            ! Emin set to 100 eV
      dle = (log(emax)-lemin)/ned

      ! Klein-Nishina DCS per energy:
      write(buffer,'(a)') 'tallyPhotonFluencePoint-debugKNeDCS.dat'
      write(*,'(a,a)') ' Writing file ',trim(buffer)
      funit = finduf()
      open(unit=funit,file=buffer)
      write(funit,'(a)')
     &  '#### Debug file: Klein-Nishina DCS ####'
      write(funit,'(a)') '#'
      write(funit,'(a)')
     &  '# Column E/E0 is scattered photon energy E '//
     &  'over primary photon energy E0.'
      write(funit,'(a)')
     &  '# Column labeled DCS is dsigma(E,E0)/dE in cm2/eV.'
      write(funit,'(a)')
     &  '#'
      write(funit,'(a)')
     &  '# E0index : E0(eV) : Eindex : E/E0 : DCS'
      write(funit,'(a)')
     &  '# --------------------------------------'
      do i=1,ned
        e = exp(lemin+(i-0.5)*dle)
        do j=1,ne
          escat = (j-0.5)*e/ne
          dcs = dcsKNe(escat,e)
          write(funit,'(1x,i3,1x,es13.5e3,1x,i3,2(1x,es13.5e3))')
     &      i,e,j,escat/e,dcs
        enddo
        write(funit,'(a)') ''
        write(funit,'(a)') ''  ! Two blanks inserted to separate data blocks
      enddo
      write(funit,'(a)') '#### EOF ####'
      close(funit)

      write(*,'(a)') 'Debug done.'
      end


      subroutine PFPdump(mode)
!*******************************************************************
!*    Dumps into or reads data from a dump file.                   *
!*                                                                 *
!*    Input:                                                       *
!*      mode:   1 to write dump file, else to read from it.        *
!*******************************************************************
      use PFPmod, only: flu,flu2,nbin
      use dumpMod
      implicit none
      integer mode

      if (mode.eq.1) then
        if (dumpfile.lt.0) return  ! No dump file open
        write(dumpfile) flu(1:nbin),flu2(1:nbin)
      else
        if (restartfile.lt.0) return  ! No restart file open
        read(restartfile) flu(1:nbin),flu2(1:nbin)
      endif
      end


!>>>> End Of File >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
