!*******************************************************************
!*                          TALLY                                  *
!*               PIXELATED IMAGING DETECTOR                        *
!*                                                                 *
!* Short description:                                              *
!*   Tallies the pixelated image corresponding to the radiation    *
!*   flux arriving at a detector.                                  *
!* Rev:                                                            *
!*   2019-09-13                                                    *
!*******************************************************************


!*******************************************************************
!*******************************************************************
!*    Modules section                                              *
!*******************************************************************
!*******************************************************************

      module PIDmod
!*******************************************************************
!*    Vars for the PID tally.                                      *
!*******************************************************************
      implicit none
      save
      logical isLab
      logical, allocatable :: isScored(:)
      integer detmode,repformat,collided,selectCollision
      integer*4 bodydet,matdet,nx,ny,nebin,ijmax,ijemax,nscore
      integer*4, allocatable :: scorelist(:)
      real*8 lab2det(3,3),r0(3),dx,dy,idx,idy
      real*8 unclimit,emin,emax,ebin,iebin,wghtlast
      real*8, parameter ::
     &       fwhm2sig2=1.0/(8.0*log(2.0d0)),
     &       npixmax=1.024e9    ! Max num pixels=32k squared (do not exceed int*4)
      real*8, allocatable :: image(:),image2(:),imagetmp(:)
      end


      module PIDsignalMod
!*******************************************************************
!*    Vars used to model signal effects (spread and collection eff)*
!*******************************************************************
      implicit none
      save
      logical isSfx
      logical, allocatable :: isPixscored(:)
      integer*4, parameter :: nzmax=10000
      integer*4 dimax(nzmax),djmax(nzmax),nz
      integer*4, allocatable :: pixscorelist(:)
      real*8 idz,zmax,coleff0,coleff1,fwhm0,fwhm1,sn0,sn1
      real*8, allocatable :: spreadmtrx(:,:,:),signaltmp(:)
      end


!*******************************************************************
!*******************************************************************
!*    Routines section                                             *
!*******************************************************************
!*******************************************************************

      subroutine PIDtally(mode,arg)
!*******************************************************************
!*    Input:                                                       *
!*      mode: Identifies the state of the calling routine          *
!*      arg: energy loss (mode<0) or history no. (mode=1)          *
!*******************************************************************
      use PIDmod
      use PIDsignalMod
      use TRACK_mod
      use SECSTcommonMod, only : nsec,ilbs
      implicit none
      integer mode
      real*8 arg

      integer*4 i,j,k,ij,ijk,is,js,ismin,ismax,pscore,n
      real*8 rdet(3),gausspdf,dimage,wghteff

      if (selectCollision.ne.0) then  ! Photon filter is active
      ! selectCollision codes are as follows:
      !  0 No filter
      ! -1 Source photon unscattered
      !  1 Source photon & 1 Rayleigh
      !  2 Source photon & 1 Compton
      !  3 Secondary photons unscattered
      !  9 All the rest: not a photon; source photon & multiply scatt; or secondary photon scattered

        select case (mode)
        case (0)                               ! Source pushed new particle to stack
          if (kpar.eq.2) then                  ! Photon
            if (selectCollision.ne.-1) return  ! Detector must be perfect absorbent, only -1 could score
          else                                 ! Not a photon
            if (selectCollision.ne.9) return   ! Only filter 9 could score
          endif

        case (1)                          ! A new history is about to begin
          ilbs(5,1:nsec) = ilbs(1,1:nsec) ! Tag all particles in the stack as source-created
          return                          ! Nothing to score, we're done

        case (-99)                        ! Particle retrieved from the stack, needs tagging
          if (kpar.eq.2) then             ! Photon
            if (ilb(5).eq.ilb(1)) then    ! It's a source photon
              collided = -1               ! Tag as unscattered source photon
            else
              collided = 3                ! Tag as secondary photon
            endif
          else                            ! Not a photon
            collided = 9                  ! Tag as multicollided event
          endif

        case (-1,-2,-8)                   ! Rayleigh(-1), Compton(-2) or auxiliary(-8) photon interaction
          if (kpar.eq.2) then             ! No additional tagging needed for non-photons
            if (collided.eq.-1) then      ! Not interacted before
              collided = -mode            ! Tag as collided (Rayleigh, Compton or auxiliary)
            else                          ! Has already interacted
              collided = 9                ! Tag as multicollided
            endif
          endif
          return                          ! No interactions occur in detector (perfect absorbent), nothing to score

        case (-98,-97)                    ! Particle has been absorbed
          if (collided.ne.selectCollision) return  ! Score only selected scattering events

        ! Note that photon interaction mechanisms 3 to 7 do not need tagging because
        !   the photon does not survive. Furthermore, they cannot occur inside the detector
        !   because it must be a perfect absorbent.
        end select
      endif

      ! Photon filter finished, proceed with scoring algorithm

      if (mode.le.0) then                       ! Energy deposition ocurred

        if (mat.ne.matdet.or.arg.eq.0.0) return ! No energy deposited in detection material

        rdet = (/x,y,z/)-r0            ! Position vector in detector origin
        if (.not.isLab) rdet = matmul(lab2det,rdet)  ! Slab and Sdet frames bases different, rotation needed
        ! Check if particle is inside pixels grid and depth range:
        i = rdet(1)*idx+1.0            ! Pixel x index
        if (i.lt.1.or.i.gt.nx) return  ! Out of pixels grid
        j = rdet(2)*idy+1.0            ! Pixel y index
        if (j.lt.1.or.j.gt.ny) return  ! Out of pixels grid
        k = rdet(3)*idz+1.0            ! Depth bin index
        if (k.lt.1.or.k.gt.nz) return  ! Not in detector
        ijk = i+(j-1)*nx+(k-1)*ijmax   ! Map pixel indices i,j,k into a global index

        ! Score:
        if (.not.isScored(ijk)) then   ! No previous score in this bin
          isScored(ijk) = .true.       ! Flag as scored
          nscore = nscore+1            ! Update size of Edep scorings list
          scorelist(nscore) = ijk      ! Add current bin index to the list of scorings
        endif
        select case (detmode)
        case (1)                       ! E integration mode
          wghteff = wght               ! Edep contributions in EI mode are weighted...
        case default                   ! Counting modes
          wghteff = 1.0                ! E deposition must not be weighted
          wghtlast = wght              ! Store the last contributing particle's wght, to be scored at bookkeeping
        end select
        imagetmp(ijk) = imagetmp(ijk)+arg*wghteff  ! Score E deposition

      ! End-Of-History bookkeeping:
      else if (mode.eq.6) then          ! Mode for EOH

        if (.not.isSfx) then            ! Signal effects (FX) not simulated
          do n=1,nscore                 ! For each Edep scoring event
            ij = scorelist(n)           ! Bin index where scoring event ocurred
            isScored(ij) = .false.      ! Reset bin flag
            dimage = imagetmp(ij)       ! Signal contribution
            imagetmp(ij) = 0.0          ! Reset Edep temp counter
            ijk = ij                    ! Set scoring bin, modified only if in PS mode
            if (detmode.ne.1) then      ! Counting modes
              k = (dimage-emin)*iebin+1.0      ! Note that k index is used for E bin
              if (k.gt.0.and.k.le.nebin) then  ! Is a valid E bin, otherwise score is 0
                ijk = ij+(k-1)*ijmax    ! Map i,j,m into a single index
                dimage = wghtlast       ! Score one count
              else                      ! Invalid bin
                dimage = 0.0            ! Nothing to score
              endif
            endif                       ! In EI mode score the signal, no further action
            image(ijk)  = image(ijk)+dimage      ! Transfer score to total counter
            image2(ijk) = image2(ijk)+dimage**2  ! ...and variance
          enddo
          nscore = 0                    ! Reset Edep scoring list
          return                        ! Done
        endif

        ! ...else, signal FX are simulated:
        pscore = 0                      ! Reset pixel scoring list
        do n=1,nscore                   ! For each Edep scoring event
          ijk = scorelist(n)            ! Bin index where scoring event ocurred
          isScored(ijk) = .false.       ! Reset bin flag
          call breakindex(ijk,i,j,k)    ! Break down global index into single components

          ! Spread signal among neighbor pixels (note k index is used for depth in detector):
          ismin = max(1,i-dimax(k))                    ! Min x pixel index for k-th depth
          ismax = min(nx,i+dimax(k))                   ! Max x pixel index for k-th depth
          do js=max(1,j-djmax(k)),min(ny,j+djmax(k))   ! For all y pixels in k-th spread matrix
            ij = ismin+(js-1)*nx                       ! Starting x pixel index in the js-th column
            do is=ismin,ismax                          ! For all x pixels in k-th spread matrix
              if (.not.isPixscored(ij)) then           ! No previous score in this pixel
                isPixscored(ij) = .true.               ! Mark as scored
                pscore = pscore+1                      ! Update size of pixel scorings list
                pixscorelist(pscore) = ij              ! Add current pixel index to the list of scorings
              endif
              signaltmp(ij) = signaltmp(ij)+
     &          imagetmp(ijk)*spreadmtrx(is-i,js-j,k)  ! Spread the image around
              ij = ij+1                                ! Next pixel, note that ij=is+(js-1)*nx
            enddo
          enddo
          imagetmp(ijk) = 0.0                          ! Reset temp counter
        enddo
        nscore = 0                                     ! Reset Edep scoring list

        ! Score pixels signal (note k index is used now for energy bin):
        do n=1,pscore                             ! For each scoring event in the list
          ij = pixscorelist(n)                    ! Pixel affected
          isPixscored(ij) = .false.               ! Reset pixel flag
          ijk = ij                                ! Set default scoring bin, modified only if in PS mode
          dimage = signaltmp(ij)                  ! Signal contribution
          signaltmp(ij) = 0.0                     ! Reset temp counter
          if (detmode.ne.1) then                  ! Counting modes, apply signal fluctuation
            dimage = max(0.0,dimage+              ! Cannot be negative after fluctuation
     &        sqrt(sn0+sn1*dimage)*gausspdf())    ! Add a Gaussian fluctuation with variance ~ A+B*signal
            k = (dimage-emin)*iebin+1.0           ! bin where the count will be added
            if (k.gt.0.and.k.le.nebin) then       ! Is a valid bin, otherwise score is 0
              ijk = ij+(k-1)*ijmax                ! Map i,j,k into a single index
              dimage = wghtlast                   ! Score one count
            else                                  ! Invalid bin
              dimage = 0.0                        ! Nothing to score
            endif
          endif                                   ! In EI score is the signal, no further action
          image(ijk)   = image(ijk)+dimage        ! Score image signal in total counter
          image2(ijk)  = image2(ijk)+dimage**2    ! ...and its variance
        enddo
      endif
      end


      subroutine PIDreport(mode,n,cputim,uncdone)
!*******************************************************************
!*    Input:                                                       *
!*      mode: -1 if called at end-of-simulation, 0 for dump only,  *
!*             1 for tally report only, 2 for both                 *
!*      n: no. of histories simulated                              *
!*      cputim: elapsed CPU time                                   *
!*    Output:                                                      *
!*      uncdone: 2 if uncert reached, 1 if not defined, 0 else     *
!*               Not computed if mode=0                            *
!*******************************************************************
      use PIDmod
      use PIDsignalMod, only : isSfx
      use RSEEDcommonMod
      implicit none
      integer mode,uncdone
      real*8 n,cputim

      character*80 buffer
      integer out,finduf,errorcode
      integer*4 nchan,i,j,k,ij,ijk
      real*8 q,sigma,eff,avesig,maxq,uncert,x,y,xmiddle,ymiddle,invn
      real*8 e,emiddle,fact,fact2,idarea

      uncdone = 1

      if (mode.ne.1) call PIDdump(1)  ! Sim dump file
      if (mode.eq.0) return           ! Dump only, no report

      ! Filename:
      select case (detmode)
        case (1)
          buffer = 'tallyPixelImageDetectEI'
        case (2)
          buffer = 'tallyPixelImageDetectPC'
        case (3)
          buffer = 'tallyPixelImageDetectPS'
      end select
      select case (repformat)
        case (1)
          buffer = trim(buffer)//'.dat'
        case (2)
          buffer = trim(buffer)//'-matrix.dat'
        case (3)
          buffer = trim(buffer)//'-binary.dat'
      end select
      out = finduf()
      if (repformat.eq.3) then         ! Open data file in binary format
        open(out,file=buffer,status='replace',access='stream',
     &       iostat=errorcode)
      else
        open(out,file=buffer,iostat=errorcode)
      endif
      if (errorcode.ne.0) then
        write(*,*) ''
        write(*,'(a)')
     &    '*********************************************'
        write(*,'(a)')
     &    'PIDreport:ERROR: cannot open output data file'
        write(*,'(a)')
     &    '*********************************************'
        close(out)  ! Just in case
        return
      endif

      if (repformat.ne.3) then ! Write header only if not in binary mode
        write(out,'(a)')
     &  '#>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
        write(out,'(a)')
     &  '# [SECTION REPORT PIXELATED IMAGING DETECTOR]'

        select case (selectCollision)
        case (-1)
          write(out,'(a)')
     &     '# Reporting ONLY unscattered source photons'
        case (1)
          write(out,'(a)')
     &     '# Reporting ONLY (once) Rayleigh-scattered photons'
        case (2)
          write(out,'(a)')
     &     '# Reporting ONLY (once) Compton-scattered photons'
        case (3)
          write(out,'(a)')
     &     '# Reporting ONLY unscattered secondary photons'
        case (9)
          write(out,'(a)')
     &     '# Reporting ONLY charged particles, scattered secondary '//
     &     'photons and multi-scattered photons'
        end select
        write(out,'(a)') '#'

        if (isSfx) then
          write(out,'(a)') '# Signal collection effects activated'
          write(out,'(a)') '# '
        endif

        select case (detmode)
        case (1)
          write(out,'(a)')
     &     '# Energy integrating mode, units are per unit pixel area:'
          if (isSfx) then
            write(out,'(a)')
     &       '#   arbitrary_signal_units/cm^2 per history.'
          else
            write(out,'(a)')
     &       '#   eV/cm^2 per history.'
          endif
        case (2)
          write(out,'(a)')
     &     '# Photon counting mode, units are counts per history.'
          write(out,'(a)') '#'
          write(out,'(a)') '# Energy interval (eV):'
          write(out,'(a,2(es12.5,1x))') '# ',emin,emax
        case (3)
          write(out,'(a)') '# Pulse height spectrum mode, '//
     &      'units are counts/eV per history.'
          write(out,'(a)') '#'
          write(out,'(a)') '# Energy interval (eV):'
          write(out,'(a,2(es12.5,1x))') '# ',emin,emax
          write(out,'(a)') '# No. of bins and bin width (eV):'
          write(out,'(a,1x,i0,1x,es12.5)') '# ',nebin,ebin
        end select

        write(out,'(a)') '#'
        write(out,'(a)')
     &    '# No. of pixels and pixel size (cm) in x & y directions:'
        write(out,'(a,1x,i0,1x,es12.5)') '# ',nx,dx
        write(out,'(a,1x,i0,1x,es12.5)') '# ',ny,dy
        write(out,'(a)')
     &    '# (x & y with respect to detector reference frame.)'
        write(out,'(a)') '#'
      endif

      select case (repformat)
      case (1)  ! Columnar
        write(out,'(a)')
     &    '# For plotting purposes, two values per '//
     &    'coordinate are given, namely, '//
     &    'the low end and the middle point of each interval.'
        write(out,'(a)') '#'
        write(out,'(a,$)')
     &    '# xPixelIndex : xLow(cm) : xMiddle(cm) : '//
     &      'yPixelIndex : yLow(cm) : yMiddle(cm) : '
        select case (detmode)
        case (1)
          write(out,'(a)') 'image(signal/cm^2/hist) : +-2sigma'
        case (2)
          write(out,'(a)') 'image(counts/hist) : +-2sigma'
        case (3)
          write(out,'(a)') 'Eindex : Elow(eV) : Emiddle(eV) : '//
     &                     'image(counts/eV/hist) : +-2sigma'
        end select
      case (2)  ! Matrix format
        write(out,'(a)') '# Pixel x index:'
        write(out,'(a,$)') '#'
        do i=1,nx
          write(out,'(i12,a,$)') i,' '  ! Extra padding space to match data columns
        enddo
        write(out,*) ''                 ! Terminates non-advancing output
      end select                        ! Nothing to do for case3 binary

      ! Write data:
      invn   = 1.0/n
      nchan  = 0
      avesig = 0.0
      maxq   = 0.5*maxval(image)                ! 1/2 of the max score
      idarea = 1.0/(dx*dy)                      ! Inverse pixel area
      select case (detmode)
      case (1)                                  ! EI
        fact  = invn*idarea                     ! Per unit pixel area
        fact2 = invn*idarea**2
      case (2)                                  ! PC
        fact  = invn                            ! Simply counts
        fact2 = invn
      case (3)                                  ! PS
        fact  = invn*iebin                      ! Per unit E bin width
        fact2 = invn*iebin**2
      end select

      do j=1,ny
        if (repformat.eq.1) then                ! Compute coords
          y = dy*(j-1)
          ymiddle = y+dy*0.5
        endif
        do i=1,nx
          if (repformat.eq.1) then
            x = dx*(i-1)
            xmiddle = x+dx*0.5
          endif
          ij = i+(j-1)*nx                       ! Mapping of i,j into a single index

          if (detmode.eq.3) then                ! PS mode
            do k=1,nebin
              if (repformat.eq.1) then
                e = emin+ebin*(k-1)
                emiddle = e+ebin*0.5
              endif
              ijk = ij+(k-1)*ijmax              ! Map i,j,k into a single index
              q = image(ijk)*fact
              sigma = sqrt(max((image2(ijk)*fact2-q**2)*invn,0.0))
              if (repformat.eq.1) then          ! Columnar
                write(out,'(3(1x,i5,2(1x,es12.5)),1x,es12.5,1x,es7.1)')
     &            i,x,xmiddle,j,y,ymiddle,k,e,emiddle,q,2.0*sigma
              else                              ! Binary; matrix format is disabled in PS mode
                write(out) q,2.0*sigma
              endif
              if (image(ijk).gt.maxq) then      ! Evaluate average uncertainty for scores above 1/2 max score
                avesig = avesig+(sigma/q)**2
                nchan = nchan+1
              endif
            enddo
            if (repformat.ne.3) write(out,*) '' ! Separate data blocks, except in binary mode

          else                                  ! EI and PC modes
            q = image(ij)*fact
            sigma = sqrt(max((image2(ij)*fact2-q**2)*invn,0.0))
            select case (repformat)
            case (1)                            ! Columnar
              write(out,'(2(1x,i5,2(1x,es12.5)),1x,es12.5,1x,es7.1)')
     &          i,x,xmiddle,j,y,ymiddle,q,2.0*sigma
            case (2)                            ! Matrix
              write(out,'(1x,es12.5,$)') q
            case (3)                            ! Binary
              write(out) q,2.0*sigma
            end select
            if (image(ij).gt.maxq) then         ! Evaluate average uncertainty for scores above 1/2 max score
              avesig = avesig+(sigma/q)**2
              nchan = nchan+1
            endif
          endif
        enddo

        if (repformat.ne.3) write(out,*) ''     ! Separate data blocks, except in binary mode
      enddo

      uncdone = 0
      if (nchan.gt.0) then
        uncert = 200.0*sqrt(avesig/nchan)
        if (uncert.lt.unclimit) uncdone = 2  ! Uncertainty reached
      else
        uncert = 0.0                         ! Uncertainty assumed not reached when score is nil
      endif

      if (repformat.ne.3) then               ! Generic report, except in binary mode
        write(out,'(a)') ' '
        write(out,'(a)') '# Performance report'
        write(out,'(a)') '#   Random seeds:'
        write(out,'(a,i0)') '#   ',seed1
        write(out,'(a,i0)') '#   ',seed2
        write(out,'(a)') '#   No. of histories simulated [N]:'
        write(out,'(a,f18.0)') '#   ',n
        write(out,'(a)') '#   CPU time [t] (s):'
        write(out,'(a,es12.5)') '#   ',cputim
        if (cputim.gt.0.0) then
          write(out,'(a)') '#   Speed (histories/s):'
          write(out,'(a,es12.5)') '#   ',n/cputim
        endif
        write(out,'(a)')
     &  '#   Average uncertainty (above 1/2 max score) in % [uncert]:'
        write(out,'(a,es12.5)') '#   ',uncert
        eff = n*uncert**2
        if (eff.gt.0.0) then
          write(out,'(a)') '#   Intrinsic efficiency [N*uncert^2]^-1:'
          write(out,'(a,es12.5)') '#   ',1.0/eff
        endif
        eff = cputim*uncert**2
        if (eff.gt.0.0) then
          write(out,'(a)') '#   Absolute efficiency [t*uncert^2]^-1:'
          write(out,'(a,es12.5)') '#   ',1.0/eff
        endif
      endif

      close(out)
      end


      subroutine PIDinitally(active,mem)
!*******************************************************************
!*    Initializes. To be called before TALLY                       *
!*                                                                 *
!*    Output:                                                      *
!*      active: TRUE if the tally is ON.                           *
!*      mem: Memory used by tally arrays.                          *
!*******************************************************************
      use PIDmod
      use PIDsignalMod
      use dataTypesMod
      use PENELOPE_mod, only : eabs,maxmat
#ifdef PROTONS
      use PENH_mod, only : eabsh
#endif
      implicit none
      logical active
      real*8 mem

      character*(*) secid,eos
      parameter (secid=
     &'[SECTION TALLY PIXELATED IMAGING DETECTOR v.2016-07-04]')
      parameter (eos='[END OF PID SECTION]')
      character*80 buffer,upcase
      integer errorcode,prtpisf
      integer*4 ijkmax
      real*8 pixlen,dmem
      real*8, parameter :: einfty=1.1e9  ! Above max E in PENELOPE

      write(*,*) ' '
      write(*,'(a)')
     & '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      call getline(buffer,0)
      if (index(buffer,secid).eq.0) then
        write(*,'(a)') 'SDDinitally:ERROR: incorrect section header;'
        write(*,'(a,a)') '  expecting to find: ',secid
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,'(a)') secid

      ! Dynamic memory usage, zero if tally not active, set later if active:
      mem = 0.0

      read(*,'(a80)') buffer
      buffer = adjustl(upcase(buffer))      ! Adjust and capitalize
      buffer = buffer(1:scan(buffer,' '))   ! Clip at 1st blank
      if (buffer.eq.'ON') then
        active = .true.
      else if (buffer.eq.'OFF') then
        active = .false.
        write(*, '(a)')
     &    '>>>> Tally Pixelated Imaging Detector is OFF >>>>'
        do
          read(*,'(a80)',iostat=errorcode) buffer
          if (errorcode.ne.0) then
            write(*,'(a,a,a)') 'PIDinitally:ERROR: ',
     &       'Unable to find End-Of-Section mark: ',eos
            stop
          endif
          if (index(buffer,eos).ne.0) return
        enddo
      else
        write(*,'(a)')
     &    'PIDinitally:ERROR: expecting to find ON or OFF'
        write(*,'(a)') 'found instead:'
        write(*,'(a)') buffer
        stop
      endif

      write(*,'(a)') 'Detection material set to:'
      read(*,*) matdet
      write(*,'(1x,i0)') matdet
      if (matdet.le.0.or.matdet.gt.maxmat) then
        write(*,'(a)') 'PIDinitally:ERROR: invalid entry.'
        stop
      endif

      write(*,'(a)') 'Filter photon collisions option:'
      read(*,*) selectCollision
      write(*,'(1x,i0)') selectCollision
      select case (selectCollision)
      case (0)           ! Valid entry, do nothing
      case (-1,1,2,3,9)  ! Valid entries, check if perfect absorbent
#ifdef PROTONS
        if (eabsh(matdet).lt.einfty) then
          write(*,'(a,es12.5)')
     &      'PIDinitally:ERROR: detection material must be a '//
     &      'perfect absorbent; increase absorption energies above ',
     &      einfty
          stop
        endif
#endif
        if (minval(eabs(:,matdet)).lt.einfty) then
          write(*,'(a,es12.5)')
     &      'PIDinitally:ERROR: detection material must be a '//
     &      'perfect absorbent; increase absorption energies above ',
     &      einfty
          stop
        endif
      case default
        write(*,'(a)') 'PIDinitally:ERROR: invalid value.'
        stop
      end select

      write(*,'(a)') 'Pixel size (cm) & No. pixels in x direction:'
      read(*,*) dx,nx
      write(*,'(1x,es12.5,1x,i0)') dx,nx
      if (dx.lt.0.0.or.nx.lt.0) then
        write(*,'(a)') 'PIDinitally:ERROR: invalid value.'
        stop
      endif
      if (dx.eq.0.0.and.nx.eq.0) then
        write(*,'(a)') 'PIDinitally:ERROR: '//
     &    'pixel size and no. pixels are both zero.'
        stop
      endif
      pixlen = dx*nx  ! Length of the pixels grid along x
      if (pixlen.ne.0.0) then
        write(*,'(a)') 'Pixels grid x-size is (cm):'
        write(*,'(1x,es12.5)') pixlen
      endif

      write(*,'(a)') 'Pixel size (cm) & No. pixels in y direction:'
      read(*,*) dy,ny
      write(*,'(1x,es12.5,1x,i0)') dy,ny
      if (dy.lt.0.0.or.ny.lt.0) then
        write(*,'(a)') 'PIDinitally:ERROR: invalid value.'
        stop
      endif
      if (dy.eq.0.0.and.ny.eq.0) then
        write(*,'(a)') 'PIDinitally:ERROR: '//
     &    'pixel size and no. pixels are both zero.'
        stop
      endif
      pixlen = dy*ny  ! Length of the pixels grid along y
      if (pixlen.ne.0.0) then
        write(*,'(a)') 'Pixels grid y-size is (cm):'
        write(*,'(1x,es12.5)') pixlen
      endif

      ! Set detector reference frame (Sdet), pixels size&num and depth bins:
      call setDetectFrame
      write(*,'(a)') 'Number of megapixels total:'
      write(*,'(1x,es12.5)') real(nx)*real(ny)*1.0e-6
      if (real(nx)*real(ny).gt.npixmax) then  ! Set a sensible max to avoid int*4 overflow
        write(*,'(a,es12.5,a)') 'PIDinitally:ERROR: ',
     &    npixmax*1.0e-6,' max no. megapixels exceeded.'
        stop
      endif

      ! Detector mode:
      write(*,'(a)') 'Detector mode:'
      read(*,*) detmode
      select case (detmode)
      case (1)
        write(*,'(a)') ' Energy integrating'
        write(*,'(a)') 'E bins and no. bins ignored, '//
     &    'not used for this mode.'
        read(*,'(a80)') buffer  ! Read dummy line
        nebin = 1
      case (2)
        write(*,'(a)') ' Photon counting'
        write(*,'(a)')
     &    'Energy threshold and maximum energy [Emin,Emax] (eV):'
        read(*,*) emin,emax     ! Ignore last datum in input line, i.e., num E bins
        write(*,'(2(1x,es12.5),1x,i0)') emin,emax
        if (emin.lt.0.0.or.emin.ge.emax) then
          write(*,'(a)') 'PIDinitally:ERROR: invalid entry.'
          stop
        endif
        write(*,'(a)') 'No. of E bins set to 1 (photon counting)'
        nebin = 1
        ebin  = emax-emin
        iebin = 1.0/ebin
      case (3)
        write(*,'(a)') ' Pulse height spectra (E discriminating)'
        write(*,'(a)')
     &    'Spectrum limits [Emin,Emax] (eV) and no. bins:'
        read(*,*) emin,emax,nebin
        write(*,'(2(1x,es12.5),1x,i0)') emin,emax,nebin
        if (emin.lt.0.0.or.emin.ge.emax.or.nebin.lt.2) then
          write(*,'(a)') 'PIDinitally:ERROR: invalid entry.'
          stop
        endif
        ebin = (emax-emin)/nebin
        iebin = 1.0/ebin
        nebin = nebin+1  ! Add one bin to store separately counts with E=Emax
      case default
        write(*,'(1x,i0)') detmode
        write(*,'(a)') 'PIDinitally:ERROR: invalid mode.'
        stop
      end select
      if (real(nx*ny)*nebin.gt.npixmax) then  ! Avoid overflow
        write(*,'(a,es12.5,a)') 'PIDinitally:ERROR: ',
     &    npixmax*1.0e-6,' max no. (Megapixels x E bins) exceeded.'
        stop
      endif

      ! Format
      write(*,'(a)') 'Report format:'
      read(*,*) repformat
      select case (repformat)
      case (1)
        write(*,'(a)') ' Columnar'
      case (2)
        write(*,'(a)') ' Matrix'
        if (detmode.eq.3) then
          write(*,'(a)') 'PIDinitally:ERROR: '//
     &      'Matrix format is incompatible with '//
     &      'pulse height spectra mode.'
          stop
        endif
      case (3)
        write(*,'(a)') ' Binary'
      case default
        write(*,'(1x,i0)') repformat
        write(*,'(a)') 'PIDinitally:ERROR: invalid format.'
        stop
      end select

      write(*,'(a)') 'Relative uncertainty (%) requested:'
      read(*,*) unclimit
      write(*,'(1x,es12.5)') unclimit

      ! Read and validate signal collection effect params:
      read(*,'(a80)') buffer                ! Skip subsection header
      read(*,'(a80)') buffer
      buffer = adjustl(upcase(buffer))      ! Adjust and capitalize
      buffer = buffer(1:scan(buffer,' '))   ! Clip at 1st blank
      if (buffer.eq.'ON') then
        isSfx = .true.
      else if (buffer.eq.'OFF') then
        isSfx = .false.
      else
        write(*,'(a)')
     &    'PIDinitally:ERROR: expecting to find ON or OFF'
        write(*,'(a)') 'found instead:'
        write(*,'(a)') buffer
        stop
      endif

      if (isSfx) then
        write(*,'(a)') 'Signal collection effects are ON'
        read(*,*) prtpisf
        write(*,'(a)') 'Collection efficiency coefficients:'
        read(*,*) coleff0,coleff1
        write(*,'(2(1x,es12.5))') coleff0,coleff1
        write(*,'(a)') 'FWHM coefficients:'
        read(*,*) fwhm0,fwhm1
        write(*,'(2(1x,es12.5))') fwhm0,fwhm1

        call setSignalSpread(prtpisf,dmem)    ! Setup spread matrix
        mem = mem+dmem                        ! Memory used by matrix

        write(*,'(a)') 'Signal noise coefficients:'
        read(*,*) sn0,sn1
        write(*,'(2(1x,es12.5))') sn0,sn1
        if (min(sn0,sn1).lt.0.0) then
          write(*,*) 'PIDinitally:ERROR: negative values are invalid.'
          stop
        endif
        sn0 = sn0*fwhm2sig2  ! This is in units of sigma^2
        sn1 = sn1*fwhm2sig2

      else
        write(*, '(a)') 'Signal collection efects are OFF'
        read(*,'(a80)') buffer  ! Skip rest of data
        read(*,'(a80)') buffer
        read(*,'(a80)') buffer
        read(*,'(a80)') buffer
        nz = 1
        idz = 1.0/zmax
      endif

      ! EOF:
      read(*,'(a80)') buffer
      if (index(buffer,eos).eq.0) then
        write(*,'(a)')
     &    'PIDinitally:ERROR: End-Of-Section mark not found'
        write(*,'(a,a)') '  expecting to find: ',eos
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif

      ! Allocate scoring arrays:
      ijmax = nx*ny                           ! Size of 2D arrays
      ijemax = ijmax                          ! Size of 3D arrays for EI and PC modes
      if (detmode.eq.3) ijemax = ijmax*nebin  ! Size 3D for PS mode
      ijkmax = ijmax*nz                       ! Number of pixels X depthBins
      mem = mem+
     &      sizeofLogic*(real(ijkmax)+real(ijmax))+
     &      sizeofInt4*(real(ijkmax)+real(ijmax))+
     &      sizeofReal8*(real(ijkmax)+3*real(ijemax))
      write(*,'(a)') 'Memory used to allocate arrays (MB):'
      write(*,'(1x,es12.5)') mem*1.0e-6
      allocate(imagetmp(ijkmax),isScored(ijkmax),scorelist(ijkmax),
     &         image(ijemax),image2(ijemax),signaltmp(ijemax),
     &         isPixscored(ijmax),pixscorelist(ijmax),stat=errorcode)
      if (errorcode.ne.0) then
        write(*,'(a)') 'PIDinitally:ERROR: not enough memory.'
        stop
      endif

      ! Init vars:
      isScored    = .false.
      nscore      = 0
      image       = 0.0
      image2      = 0.0
      imagetmp    = 0.0
      signaltmp   = 0.0
      isPixscored = .false.

      ! Sim restart file:
      call PIDdump(0)

      write(*,'(a)') '>>>> PID tally initialization finished >>>>'
      end


      subroutine setDetectFrame
!*******************************************************************
!*    Identifies detector reference frame and prepares rot matrix. *
!*******************************************************************
      use PIDmod
      use PIDsignalMod, only : zmax
      use PENGEOMcommonsMod
      use PENGEOM_mod, only : nxg,mater,nbody
      use dataTypesMod
      implicit none
      integer*4 i,isurf,sidepoint,surf,surfmax
      real*8 normx(3),scalarx,normy(3),scalary,normz(3),scalarz
      real*8 normtop(3),scalartop,imodu,pixgrid,a(3,3),detbox,det3x3
      real*8, parameter :: eps=1.0d-13,oneplus=1.0d0+eps

      ! Search last body containing the detector material
      bodydet = 0                     ! Init void
      do i=nbody,1,-1                 ! Reverse search, start from top
        if (mater(i).eq.matdet) then  ! This body is detector
          write(*,'(a)')
     &      'Detection material found in body/module no. '//
     &      '(PENGEOM internal numbering): '
          write(*,'(1x,i0)') i
          bodydet = i
          exit                        ! Only 1st ocurrence matters
        endif
      enddo
      if (bodydet.eq.0) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'detection material not found.'
        stop
      endif
      surfmax = ksurf(bodydet,nxg)    ! Num. surfaces limiting this body

      ! Search for detector box faces:
      write(*,'(a)') 'Searching for detector box faces...'
      isurf = 0                     ! Init
      do i=1,surfmax                ! For all surfaces limiting BODYDET
        sidepoint = 2*kflag(bodydet,i)-3  ! Side pointer -1,+1,+3,etc
        if (sidepoint.gt.+1) cycle  ! Skip surfaces limiting inserted bodies
        isurf = i                   ! Found x-face
        exit
      enddo
      if (isurf.eq.0) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'detector box faces not defined.'
        stop
      endif
      surf = ksurf(bodydet,isurf)   ! Surface number
      write(*,'(a)') 'Found x-normal face; it is surface no. '//
     &  '(PENGEOM internal numbering):'
      write(*,'(1x,i0)') surf
      if (kplane(surf).ne.1) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &      'x-normal face is not a plane.'
        stop
      endif
      normx = (/ax(surf),ay(surf),az(surf)/)  ! Surface normal
      imodu = sidepoint/sqrt(sum(normx**2))   ! Signed inverse modulus
      normx = normx*imodu                     ! Inwards-oriented unit normal
      scalarx = a0(surf)*imodu                ! Scalar term, normalized

      ! Seach for x-normal top face:
      isurf = isurf+1                         ! Next surface ID should be x-normal top face
      if (isurf.gt.surfmax) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'x-normal top face not defined.'
        stop
      endif
      sidepoint = 2*kflag(bodydet,isurf)-3
      if (sidepoint.gt.+1) then               ! This surface is not a valid face, it limits an inserted body
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'x-normal top face not defined.'
        stop
      endif
      surf = ksurf(bodydet,isurf)             ! Surface number
      write(*,'(a)') 'Found x-normal top face; it is surface no. '//
     &  '(PENGEOM internal numbering):'
      write(*,'(1x,i0)') surf
      if (kplane(surf).ne.1) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &      'x-normal top face is not a plane.'
        stop
      endif
      normtop = (/ax(surf),ay(surf),az(surf)/)
      imodu = -sidepoint/sqrt(sum(normtop**2))
      normtop = normtop*imodu                 ! Outwards-oriented unit normal, should coincide with NORMX
      scalartop = a0(surf)*imodu              ! Scalar term, normalized

      ! Check if parallel and correctly oriented:
      if (abs(dot_product(normx,normtop)-1.0d0).gt.eps) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'x-normal and x-normal top faces are not parallel or '//
     &    'their side pointers are incorrectly defined.'
        stop
      endif

      ! Set grid size along x:
      detbox = abs(scalarx-scalartop)        ! Detector box size = Distance between planes
      write(*,'(a)') 'Detector box x-size is (cm):'
      write(*,'(1x,es12.5)') detbox
      if (detbox.lt.eps) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'detector box size is too small.'
        stop
      endif
      pixgrid = nx*dx                         ! Pixels grid size, as defined by user
      if (pixgrid.ne.0.0) then                ! Pixels grid size explicitly given by user
        if (abs(pixgrid-detbox).gt.eps*detbox) then
          write(*,'(a)') '(pixels grid and detector box sizes '//
     &      'do not coincide; detector box x-size is ignored.)'
        endif
      else if (nx.eq.0) then                  ! Infer num. pixels from geometry
        write(*,'(a)') 'Number of pixels along x set to:'
        if (detbox*oneplus/dx.gt.npixmax) then
          write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'overflow in no. pixels; redefine detector box parameters.'
          stop
        endif
        nx = detbox*oneplus/dx
        write(*,'(1x,i0)') nx
      else                                    ! user entered dx=0 => infer pixel size
        dx = detbox/nx
        write(*,'(a)') 'Pixel size along x set to (cm):'
        write(*,'(1x,es12.5)') dx
      endif
      idx = 1.0/dx

      ! Seach for y-normal face:
      isurf = isurf+1                         ! Next surface ID should be y-normal face
      if (isurf.gt.surfmax) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'y-normal face not defined.'
        stop
      endif
      sidepoint = 2*kflag(bodydet,isurf)-3
      if (sidepoint.gt.+1) then               ! This surface is not a valid face, it limits an inserted body
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'y-normal face not defined.'
        stop
      endif
      surf = ksurf(bodydet,isurf)             ! Surface number
      write(*,'(a)') 'Found y-normal face; it is surface no. '//
     &  '(PENGEOM internal numbering):'
      write(*,'(1x,i0)') surf
      if (kplane(surf).ne.1) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &      'y-normal face is not a plane.'
        stop
      endif
      normy = (/ax(surf),ay(surf),az(surf)/)
      imodu = sidepoint/sqrt(sum(normy**2))
      normy = normy*imodu                     ! Inwards-oriented unit normal
      scalary = a0(surf)*imodu                ! Scalar term, normalized

      ! Seach for y-normal top face:
      isurf = isurf+1                         ! Next surface ID should be y-normal top face
      if (isurf.gt.surfmax) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'y-normal top face not defined.'
        stop
      endif
      sidepoint = 2*kflag(bodydet,isurf)-3
      if (sidepoint.gt.+1) then               ! This surface is not a valid face, it limits an inserted body
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'y-normal top face not defined.'
        stop
      endif
      surf = ksurf(bodydet,isurf)             ! Surface number
      write(*,'(a)') 'Found y-normal top face; it is surface no. '//
     &  '(PENGEOM internal numbering):'
      write(*,'(1x,i0)') surf
      if (kplane(surf).ne.1) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &      'y-normal top face is not a plane.'
        stop
      endif
      normtop = (/ax(surf),ay(surf),az(surf)/)
      imodu = -sidepoint/sqrt(sum(normtop**2))
      normtop = normtop*imodu                 ! Outwards-oriented unit normal, should coincide with NORMY
      scalartop = a0(surf)*imodu              ! Scalar term, normalized

      ! Check if parallel and correctly oriented:
      if (abs(dot_product(normy,normtop)-1.0d0).gt.eps) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'y-normal and y-normal top faces are not parallel or '//
     &    'their side pointers are incorrectly defined.'
        stop
      endif

      ! Set grid size along y:
      detbox = abs(scalary-scalartop)         ! Detector box size = Distance between planes
      write(*,'(a)') 'Detector box y-size is (cm):'
      write(*,'(1x,es12.5)') detbox
      if (detbox.lt.eps) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'detector box size is too small.'
        stop
      endif
      pixgrid = ny*dy
      if (pixgrid.ne.0.0) then                ! Pixels grid size given by user
        if (abs(pixgrid-detbox).gt.eps*detbox) then
          write(*,'(a)') '(pixels grid and detector box sizes '//
     &      'do not coincide; detector box y-size is ignored.)'
        endif
      else if (ny.eq.0) then                  ! Infer num. pixels
        write(*,'(a)') 'Number of pixels along y set to:'
        if (detbox*oneplus/dy.gt.npixmax) then
          write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'overflow in no. pixels; redefine detector box parameters.'
          stop
        endif
        ny = detbox*oneplus/dy
        write(*,'(1x,i0)') ny
      else                                    ! Infer pixel size
        dy = detbox/ny
        write(*,'(a)') 'Pixels size along y set to (cm):'
        write(*,'(1x,es12.5)') dy
      endif
      idy = 1.0/dy

      ! Check perpendicularity of x&y faces:
      if (abs(dot_product(normx,normy)).gt.eps) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'x & y faces are not perpendicular.'
        stop
      endif

      ! Seach for z-normal face:
      isurf = isurf+1                         ! Next surface ID should be z-normal face
      if (isurf.gt.surfmax) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'z-normal face not defined.'
        stop
      endif
      sidepoint = 2*kflag(bodydet,isurf)-3
      if (sidepoint.gt.+1) then               ! This surface is not a valid face, it limits an inserted body
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'z-normal face not defined.'
        stop
      endif
      surf = ksurf(bodydet,isurf)             ! Surface number
      write(*,'(a)') 'Found z-normal face; it is surface no. '//
     &  '(PENGEOM internal numbering):'
      write(*,'(1x,i0)') surf
      if (kplane(surf).ne.1) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &      'z-normal face is not a plane.'
        stop
      endif
      normz = (/ax(surf),ay(surf),az(surf)/)
      imodu = sidepoint/sqrt(sum(normz**2))
      normz = normz*imodu                     ! Inwards-oriented unit normal
      scalarz = a0(surf)*imodu                ! Scalar term, normalized

      ! Seach for z-normal top face:
      isurf = isurf+1                         ! Next surface ID should be z-normal top face
      if (isurf.gt.surfmax) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'z-normal top face not defined.'
        stop
      endif
      sidepoint = 2*kflag(bodydet,isurf)-3
      if (sidepoint.gt.+1) then               ! This surface is not a valid face, it limits an inserted body
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'z-normal top face not defined.'
        stop
      endif
      surf = ksurf(bodydet,isurf)             ! Surface number
      write(*,'(a)') 'Found z-normal top face; it is surface no. '//
     &  '(PENGEOM internal numbering):'
      write(*,'(1x,i0)') surf
      if (kplane(surf).ne.1) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &      'z-normal top face is not a plane.'
        stop
      endif
      normtop = (/ax(surf),ay(surf),az(surf)/)
      imodu = -sidepoint/sqrt(sum(normtop**2))
      normtop = normtop*imodu                 ! Outwards-oriented unit normal, should coincide with NORMZ
      scalartop = a0(surf)*imodu              ! Scalar term, normalized

      ! Check if parallel and correctly oriented:
      if (abs(dot_product(normz,normtop)-1.0d0).gt.eps) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'z-normal and z-normal top faces are not parallel or '//
     &    'their side pointers are incorrectly defined.'
        stop
      endif

      ! Set detector depth:
      zmax = abs(scalarz-scalartop)         ! Depth = Distance between planes
      write(*,'(a)') 'Detector depth is (cm):'
      write(*,'(1x,es12.5)') zmax
      if (zmax.lt.eps) then
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'detector depth is too small.'
        stop
      endif

      ! Check perpendicularity of z with x&y faces:
      a(1,:) = normx
      a(2,:) = normy
      a(3,:) = normz
      if (abs(abs(det3x3(a))-1.0d0).gt.eps) then  ! Notice that the system could be left handed with det=-1
        write(*,'(a)') 'setDetectFrame:ERROR: '//
     &    'z faces are not perpendicular to x & y faces.'
        stop
      endif

      ! Lab->Pixels change of base matrix:
      lab2det(1,:) = normx
      lab2det(2,:) = normy
      lab2det(3,:) = normz

      isLab = .false.
      if (abs(normx(1)-1.0d0).lt.eps.and.
     &    abs(normy(2)-1.0d0).lt.eps.and.
     &    abs(normz(3)-1.0d0).lt.eps) then
        isLab = .true.
        write(*,'(a)') 'Detector reference frame coincides with lab '//
     &    'frame; no rotation required.'
      endif

      ! Origin of the detector system:
      ! The origin r0 is the solution of the linear eqs
      !      normx.r0 + scalarx  = 0  (x-face, or yz, plane)
      !      normy.r0 + scalary  = 0  (y-face, or xz, plane)
      !      normz.r0 + scalarz  = 0  (z-face, or xy, plane)
      !
      ! or, in matrix form
      !      {lab2det} . r0 + scalar = 0
      !
      ! where 'r0' and 'scalar' are column vectors and '.' represents the matrix product.
      ! The solution is
      !      r0  = - {lab2det}^ . scalar
      !
      ! where ^ represents matrix transposition. Equivalently, using row vectors,
      !      r0^ = - scalar^ . {lab2det}
      !
      r0 = -matmul((/scalarx,scalary,scalarz/),lab2det)
      end


      subroutine setSignalSpread(prtpisf,mem)
!*******************************************************************
!*    Prepares data for signal spread effects.                     *
!*                                                                 *
!*    Input:                                                       *
!*      prtpisf: 0 to skip writing a PISFile, else write it.       *
!*    Output:                                                      *
!*      mem: Memory used by tally arrays.                          *
!*    Comments:                                                    *
!*      The following parameter values can be adjusted:            *
!*      nfwhm:                                                     *
!*        The radial spread is truncated at nfwhm*FWHM(z). For a   *
!*        Lorentzian this is unavoidable because 2D Lorentzians are*
!*        not normalizable. It also limits the size of the spread  *
!*        matrix.                                                  *
!*      dfwhm:                                                     *
!*        Allowed fractional change of FWHM(z) when increasing one *
!*        depth bin dz; used to set an upper limit on dz.          *
!*      dcoleff:                                                   *
!*        Allowed fractional change of COLEFF(z) when increasing   *
!*        one depth bin dz; used to set an upper limit on dz.      *
!*******************************************************************
      use PIDmod
      use PIDsignalMod
      use dataTypesMod
      implicit none
      integer prtpisf
      real*8 mem

      character*80, parameter :: pisfname='tallyPixelImageDetectPISF-'
      character*80 fnametail
      integer errorcode,finduf,out,m
      integer*4 i,j,k,dimx,dimy
      real*8 radius,x,y,z,prob,dz
      real*8, parameter :: pi=3.1415926535897932d0
      real*8, parameter :: nfwhm=5.0,dfwhm=0.001,dcoleff=0.001  ! See comments in header

      write(*,'(a)')
     &  'Lorentzian model for the optical point spread function'
      write(*,'(a,es8.1,a)') 'Neglecting contributions farther than ',
     &  nfwhm,' times the FWHM;'
      write(*,'(a,f3.1,a)') '  this represents ',
     &  100.0*pisf(nfwhm**2,1.0d0)/pisf(0.0d0,1.0d0),
     &  '% of the signal at the center'
      write(*,'(a)') 'FWHM (cm) at zDet=0 & zDet=zmax:'
      write(*,'(2(1x,es12.5))') fwhm(0.0d0),fwhm(zmax)
      write(*,'(a)') 'Parameter max allowed fractional FWHM change:'
      write(*,'(1x,es8.1)') dfwhm
      write(*,'(a)') 'Collection efficiency at zDet=0 & zDet=zmax:'
      write(*,'(2(1x,es12.5))') coleff(0.0d0),coleff(zmax)
      write(*,'(a)')
     &  'Parameter max allowed fractional colection eff change:'
      write(*,'(1x,es8.1)') dcoleff

      ! Set no. of depth bins so that colection eff and FWHM change by less than specified:
      nz = max(1,ceiling(zmax*abs(coleff1)/dcoleff),
     &           ceiling(zmax*abs(fwhm1)/dfwhm))
      write(*,'(a)') 'No. of depth bins:'
      write(*,'(1x,i0)') nz
      if (nz.gt.nzmax) then
        write(*,'(a)')
     &    'setSignalSpread:WARNING: Max no. of depth bins exceeded.'
        write(*,'(a,i0)') '  Resetting to max no.: ',nzmax
        write(*,'(a,i0)')
     &    '  Accuracy of FWHM or collection eff might be compromised.'
        nz = nzmax
      endif
      dz = zmax/nz
      idz = 1.0/dz
      do k=1,nz
        radius = nfwhm*fwhm((k-0.5)*dz)   ! Max distance considered for spreading signal
        dimax(k) = int(radius/dx)         ! Number of spread pixels in x direction
        djmax(k) = int(radius/dy)         ! Number of spread pixels in y direction
      enddo

      ! Allocate spread matrix:
      dimx = maxval(dimax)   ! Max dimension of spread matrix in x direction
      dimy = maxval(djmax)   ! Max dimension of spread matrix in y direction
      write(*,'(a,i0,a,i0,a,i0,a)') 'Allocating spread matrix(',
     &  2*dimx+1,',',2*dimy+1,',',nz,')'
      mem = sizeofReal8*real(2*dimx+1)*real(2*dimy+1)*real(nz)
      write(*,'(a)') 'Memory used by matrix (MB):'
      write(*,'(1x,es12.5)') mem*1.0e-6
      allocate(spreadmtrx(-dimx:+dimx,-dimy:+dimy,nz),stat=errorcode)
      if (errorcode.ne.0) then
        write(*,'(a)') 'setSignalSpread:ERROR: '//
     &     'unable to allocate spread matrix.'
        stop
      endif
      spreadmtrx = 0.0  ! Clear

      ! Set spread matrix to product (colection eff) x (Image Point Spread Function) and normalize it to 1:
      do k=1,nz
        z = (k-0.5)*dz
        do j=0,djmax(k)
          y = j*dy
          do i=0,dimax(k)
            x = i*dx
            prob = pisf(x**2+y**2,fwhm(z)**2)
            spreadmtrx(i,j,k)   = prob
            spreadmtrx(-i,j,k)  = prob  ! Take advantage of circular symmetry
            spreadmtrx(i,-j,k)  = prob
            spreadmtrx(-i,-j,k) = prob
          enddo
        enddo

        spreadmtrx(0,0,k) = volpisf(dx*dy/pi,fwhm(z)**2)/(dx*dy)      ! Correct central (circular approx) pixel overestimation
        spreadmtrx(:,:,k) =  spreadmtrx(:,:,k)/sum(spreadmtrx(:,:,k)) ! Renormalize matrix to unit volume
     &                       *coleff(z)                               ! ...and introduce collection efficiency
      enddo

      ! If requested, write spread matrix (at zdet=0, zmid & zmax):
      if (prtpisf.eq.0) return  ! Do not write
      write(*,'(a)') 'Writing point spread funtion to files '//
     &  trim(pisfname)//'.dat'
      out = finduf()

      do m=1,3                    ! For z=0, zmid and zmax
        select case (m)
        case (1)
          fnametail = 'z=0'       ! Used to complete the file name
          k = 1
        case (2)
          if (nz.lt.2) return     ! Do not write an identical matrix twice
          fnametail = 'z=zmid'
          k = ceiling(0.5*(nz+1)) ! Midpoint index; notice equals 2 if nz=2
        case (3)
          if (nz.lt.3) return     ! Do not write an identical matrix twice
          fnametail = 'z=zmax'
          k = nz
        end select

        open(out,file=trim(pisfname)//trim(fnametail)//'.dat',
     &    iostat=errorcode)
        if (errorcode.ne.0) then
          write(*,*) ''
          write(*,'(a)') 'setSignalSpread:ERROR: '//
     &      'cannot open point spread matrix data file.'
          stop
        endif
        write(out,'(a)') '# Image point spread matrix at '//
     &    trim(fnametail)
        write(out,'(a,i0,a,i0,a)') '# Matrix dimensions are (',
     &    (2*dimx+1),'x',(2*dimy+1),')'
        write(out,'(a)')
     &    '# x_det (cm) runs vertical in 1st column, '//
     &    'y_det (cm) runs horizontal in 1st row'
        write(out,'(a,$)') '   0      '           ! Write dummy coordinate 0 and align columns
        do j=-dimy,dimy
          write(out,'(1x,es9.2,$)') j*dy
        enddo
        write(out,*) ''
        do i=-dimx,dimx
          write(out,'(1x,es9.2,$)') i*dx
          do j=-dimy,dimy
            write(out,'(1x,es9.2,$)') spreadmtrx(i,j,k)
          enddo
          write(out,*) ''                   ! EOL
        enddo
        close(out)
      enddo

      contains
        real*8 function pisf(r2,fwhm2)
        !*******************************************************************
        !*    Image point spread function for optical effects.             *
        !*    2D unnormalized Lorentz distribution in Cartesian coords     *
        !*    truncated at NFHWM (see routine parameters).                 *
        !*                                                                 *
        !*    Input:                                                       *
        !*      r2: x^2+y^2, the radius squared (cm2)                      *
        !*      fwhm2: FWHM^2 (cm2); if 0 a delta function is assumed.     *
        !*******************************************************************
        implicit none
        real*8 r2,fwhm2
        real*8, parameter :: zero=1.0d-60,nfwhm2=nfwhm**2

        if (fwhm2.lt.zero) then  ! The distribution is a delta function
          if (r2.lt.zero) then   ! Not null only at the central pixel
            pisf = 1.0           ! Arbitrary value
          else
            pisf = 0.0           ! Null elsewhere
          endif
          return
        endif

        if (r2.lt.nfwhm2*fwhm2) then
          pisf = 1.0/(r2+0.25*fwhm2)
        else                     ! Truncate
          pisf = 0.0
        endif
        end function

        real*8 function volpisf(r2,fwhm2)
        !*******************************************************************
        !*    Volume of the PISF distribution in (x,y) up to r=sqrt(r2)    *
        !*    Lorentzian PISF assumed; the integ over all (x,y) is infinite*
        !*                                                                 *
        !*    Input:                                                       *
        !*      r2: radius squared (cm^2)                                  *
        !*      fwhm2: FWHM^2 (cm2); if 0 a delta function is assumed.     *
        !*******************************************************************
        implicit none
        real*8 r2,fwhm2
        real*8, parameter :: zero=1.0d-60

        if (fwhm2.lt.zero) then  ! The distribution is a delta function
          volpisf = 1.0
          return
        endif
        volpisf = pi*log(1.0+4.0*r2/fwhm2)
        end function

        real*8 function fwhm(z)
        !*******************************************************************
        !*    FWHM (cm) of signal spread as a function of depth z (cm),    *
        !*    linear approx                                                *
        !*                                                                 *
        !*    Input:                                                       *
        !*      z: depth in detector frame (cm)                            *
        !*******************************************************************
        implicit none
        real*8 z

        fwhm = max(0.0,fwhm0+fwhm1*z)  ! Cannot be negative
        end function

        real*8 function coleff(z)
        !*******************************************************************
        !*    Signal collection efficiency (0<CE<1) as a function of       *
        !*    depth, linear approx.                                        *
        !*                                                                 *
        !*    Input:                                                       *
        !*      z: depth in detector frame (cm), from z=0                  *
        !*******************************************************************
        implicit none
        real*8 z

        coleff = min(1.0,max(0.0,coleff0+coleff1*z))  ! Must be in [0,1]
        end function
      end


      subroutine breakindex(n,i,j,k)
!*******************************************************************
!*    Finds indices i,j,k corresponding to a global index n        *
!*    defined as n = i+(j-1)*imax+(k-1)*imax*jmax.                 *
!*                                                                 *
!*    Input:                                                       *
!*      n: global index                                            *
!*      i: first index                                             *
!*      j: second index                                            *
!*      k: third index                                             *
!*    Comments:                                                    *
!*      imax and imax*jmax are input via module.                   *
!*******************************************************************
      use PIDmod, only : ijmax,imax=>nx  ! IMAX is alias for NX
      implicit none
      integer*4 n,i,j,k

      integer*4 m,kc,jc

      ! Index names ending in 'c' represent C style indexing, e.g., kc == k-1
      kc = int((n-1)/ijmax)
      m = n-kc*ijmax
      k = kc+1

      jc = int((m-1)/imax)
      i = m-jc*imax

      j = jc+1
      end


      subroutine PIDdump(mode)
!*******************************************************************
!*    Dumps into or reads data from a dump file.                   *
!*                                                                 *
!*    Input:                                                       *
!*      mode: 1 to write dump file, else to read from it.          *
!*******************************************************************
      use PIDmod, only : image,image2,ijemax
      use dumpMod
      implicit none
      integer mode

      if (mode.eq.1) then
        if (dumpfile.lt.0) return  ! No dump file open
        write(dumpfile) image(1:ijemax),image2(1:ijemax)
      else
        if (restartfile.lt.0) return  ! No restart file open
        read(restartfile) image(1:ijemax),image2(1:ijemax)
      endif
      end


!>>>> End Of File >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
