!*******************************************************************
!* This package has been adapted for its use in conjunction with   *
!* penEasy. Trivial modifications have been introduced, which are  *
!* marked with '>>>> penEasy' for easy identification.             *
!*                                                                 *
!* Rev:                                                            *
!*   2019-09-20 by JS                                              *
!*******************************************************************


CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                      C
C            PPPPP   EEEEEE  N    N  N    N  U    U   CCCC             C
C            P    P  E       NN   N  NN   N  U    U  C    C            C
C            P    P  E       N N  N  N N  N  U    U  C                 C
C            PPPPP   EEEE    N  N N  N  N N  U    U  C                 C
C            P       E       N   NN  N   NN  U    U  C    C            C
C            P       EEEEEE  N    N  N    N   UUUU    CCCC             C
C                                                                      C
C                                                   (version 2018).    C
C                                                                      C
C  Subroutine package for Monte Carlo simulation of decay paths of     C
C  radioactive nuclei.                                                 C
C                                                                      C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                      C
C  PENNUC (version 2018)                                               C
C  Copyright 2016-2018                                                 C
C  CIEMAT, Laboratoire National Henri Becquerel, and                   C
C  Universitat de Barcelona                                            C
C                                                                      C
C  Permission to use, copy, modify, distribute and sell this software  C
C  and its documentation for any purpose is hereby granted without     C
C  fee, provided that the above copyright notice appears in all        C
C  copies and that both that copyright notice and this permission      C
C  notice appear in all supporting documentation. The Ciemat and the   C
C  Universitat de Barcelona make no representations about the suita-   C
C  bility of this software for any purpose. It is provided "as is"     C
C  without express or implied warranty.                                C
C                                                                      C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                      C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C  This subroutine package simulates the decay of radionuclides by
C  using information from the NUCLEIDE database. The decay parameters
C  of the source nuclides are read from a text file, in the dedicated
C  PenNuc format, which can be downloaded from the NUCLEIDE web site
C     http://www.nucleide.org/DDEP_WG/DDEPdata.htm
C
C  A radioactive nucleus may reach a metastable level (i.e., a level
C  with a life time longer than the detector resolution time) in the
C  course of its de-excitation. Such levels are treated as effective
C  halts of the decay process. The radiations emitted between successive
C  halts constitute a 'cascade' and are considered to define a single
C  emission event.
C
C  The decay simulation of each radioisotope is initialised by calling
C  subroutine PENNUC0, which reads the NUCLEIDE datafile of that isotope
C  and extracts atomic relaxation data from the PENELOPE database files.
C  PENNUC0 prepares tables of relevant quantities and defines the value
C  of the output parameter EPMAX (the highest energy of the emitted
C  radiations). The PENNUC subroutine package can handle up to NISM
C  radionuclides, which are identified by a label IS defined by the
C  order in which they are loaded. That is, IS=1, 2, ...  denotes the
C  first, second, ... loaded nuclides.
C
C  At initialisation, the present subroutines read information from
C  the files 'pdatconf.p14' and 'pdrelax.p11' of the PENELOPE database.
C  When running the program, these two files must be present in the
C  PENNUC data directory, whose relative path (RPATH) is defined in the
C  module PENNUC_mod.
C
C  NOTE: The initialisation subroutines open and close the output units
C  numbers NRD=51 and MRD=52. Make sure these units are not open when
C  calling subroutine PENNUC0.
C


C  >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      MODULE PENNUC_mod
C
C  ****  Simulation of nuclear decay. Global quantities and state
C        variables of emitted particles. All energies in eV.
C
      SAVE  ! Saves all items in the module.
C  ----  Resolution time of the detector.
      DOUBLE PRECISION, PARAMETER :: DRTIME=5.0D-6
C  ----  Cutoff energy for the de-excitation of inner subshells.
      DOUBLE PRECISION, PARAMETER :: ECNUC=200.0D0
C  ----  Relative path to the PENNUC data directory.
!>>>> penEasy: Define path to PENNUC database:
!     CHARACTER*2 :: RPATH='./'  ! The working directory.
      CHARACTER*11 :: RPATH='./nuclides/'
!>>>> penEasy ends
C  ----  Physical dimensions of arrays:
      INTEGER*4, PARAMETER :: NISM=5   ! Max. no. of nuclides.
      INTEGER*4, PARAMETER :: NDM=5    ! Max. no. of daughters.
      INTEGER*4, PARAMETER :: NBM=200  ! Max. no. of branches.
      INTEGER*4, PARAMETER :: NLM=200  ! Max. no. of levels.
      INTEGER*4, PARAMETER :: NTM=200  ! Max. no. of transitions.
      INTEGER*4 :: NIR=0  ! Number of loaded radionuclides.
C
C  ----  Daughter nucleus (IZD,IAD) and current level (LEVD).
      INTEGER*4 :: IZD,IAD,LEVD
C  ----  Particles emitted in the last decay cascade.
      INTEGER*4, PARAMETER :: NRM=500  ! Max. no. of particles.
      INTEGER*4 :: NR  ! No. of particles in the cascade (<= NRM).
C  ----  Particle types (1=electron, 2=photon, 3=positron, 4=alpha).
      INTEGER*4, DIMENSION (NRM) :: KPNR
C  ----  Transitions that originated the particles.
      INTEGER*4, DIMENSION (NRM) :: ITNR
C  ----  Particle energies, ages, and direction cosines.
      DOUBLE PRECISION, DIMENSION (NRM) :: ENR,AGENR,UNR,VNR,WNR
C
      END MODULE PENNUC_mod
C  <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


C  *********************************************************************
C                       SUBROUTINE PENNUC
C  *********************************************************************
      SUBROUTINE PENNUC(IS,METAST,IER)
C
C  This subroutine generates random decay pathways of the radioactive
C  nuclide IS using information from the NUCLEIDE database.
C
C  Metastable levels (i.e., levels with life times longer than the
C  detector resolution time) are treated as effective halts of the decay
C  process. When a metastable level is reached, the output value of
C  METAST is set to 1 and control is returned to the main program. At
C  the next call to subroutine PENNUC, the simulation of the decay
C  continues down to the following metastable level or to the ground
C  level. The output value METAST=0 indicates that the nucleus has
C  reached its ground level. This operation scheme allows simulating the
C  response of real detectors, which may give several output counts as
C  the result of the decay of a single nucleus through a metastable
C  state.
C
C  Each call to PENNUC returns the initial states of the set of
C  particles released in a cascade. The initial state of each particle
C  is defined by its type (KPNR), energy (ENR), age (AGENR), and
C  direction cosines (UNR,VNR,WNR). The quantity ITNR identifies the
C  transition that originated the particle. The number NR of emitted
C  particles and their initial-state variables are returned through the
C  module PENNUC_mod.
C
C  The transition identifier, ITNR, is defined as follows:
C  -- For atomic transitions IS0-IS1 or IS0-IS1-IS2 of an ion of atomic
C     number IZ,
C        ITNR=IZ*1000000+IS0*10000+IS1*100+IS2
C     where ISi is the label code of a subshell (see the next table).
C     For radiative transitions, IS2=0.
C
C  ---------------------------------------------------------------------
C  Label code IS for electron subshells:
C      1 = K  (1s1/2),     11 = N2 (4p1/2),     21 = O5 (5d5/2),
C      2 = L1 (2s1/2),     12 = N3 (4p3/2),     22 = O6 (5f5/2),
C      3 = L2 (2p1/2),     13 = N4 (4d3/2),     23 = O7 (5f7/2),
C      4 = L3 (2p3/2),     14 = N5 (4d5/2),     24 = P1 (6s1/2),
C      5 = M1 (3s1/2),     15 = N6 (4f5/2),     25 = P2 (6p1/2),
C      6 = M2 (3p1/2),     16 = N7 (4f7/2),     26 = P3 (6p3/2),
C      7 = M3 (3p3/2),     17 = O1 (5s1/2),     27 = P4 (6d3/2),
C      8 = M4 (3d3/2),     18 = O2 (5p1/2),     28 = P5 (6d5/2),
C      9 = M5 (3d5/2),     19 = O3 (5p3/2),     29 = Q1 (7s1/2),
C     10 = N1 (4s1/2),     20 = O4 (5d3/2),     30 = outer shells.
C  ---------------------------------------------------------------------
C
C  -- For nuclear disintegrations and de-excitations,
C        ITNR=-(IDAUGH*1000000+IBRANCH*1000+ITR)
C     where IDAUGH and IBRANCH are the indices of the daughter and the
C     branch, and ITR is the type of nuclear transition that released
C     the particle.
C
C        ITR= 0, alpha disintegration,
C             1, beta- disintegration,
C             2, beta+ disintegration,
C             3, gamma emission,
C             4, internal conversion in the K shell,
C             5, internal conversion in the L1 subshell,
C             6, internal conversion in the L2 subshell,
C             7, internal conversion in the L3 subshell,
C             8, internal conversion in the M, N and outer subshells.
C
C
C  Input argument:
C    IS ........... active radioisotope.
C
C  Output arguments:
C    METAST ....... output flag for metastable levels,
C                   = 0, the de-excitation cascade has been completed,
C                   = 1, a metastable state has been reached.
C    IER  ......... error code. When IER.NE.0 the isotope IS is not
C                   defined.
C
      USE PENNUC_mod
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      CHARACTER NUCLIDE*7,DNUCLIDE*7
C  ----  Minimum lifetime for metastable levels.
      PARAMETER (TIMELIM=DRTIME*1.0D-2)
C  ****  Nuclear dacay data.
      COMMON/CPN1/NUCLIDE(NISM),DNUCLIDE(NISM,NDM)
      COMMON/CPN2/KTYPE(NISM,NDM,NBM),NBRANCH(NISM,NDM),
     1  NLEVEL(NISM,NDM),IZNUC(NISM),IANUC(NISM),NDAUGH(NISM)
      COMMON/CPN3/Q(NISM,NDM),UQ(NISM,NDM),PDAUGH(NISM,NDM),
     1  UPDAUGH(NISM,NDM),ELEVEL(NISM,NDM,0:NTM),
     2  UELEVEL(NISM,NDM,0:NTM),DTIME(NISM,NDM,NLM),
     3  UDTIME(NISM,NDM,NLM)
      COMMON/CPN4/BR(NISM,NDM,NBM),UBR(NISM,NDM,NBM),
     1  EBRANCH(NISM,NDM,NBM),UEBRANCH(NISM,NDM,NBM),IFLB(NISM,NDM,NBM),
     2  IPRBETA(NISM,NDM,NBM),INDBETA(NISM,NDM,NBM),IDT
      COMMON/CPN5/PTRANS(NISM,NDM,NLM,NTM),UPTRANS(NISM,NDM,NLM,NTM),
     1  ETRANS(NISM,NDM,NLM,NTM),UETRANS(NISM,NDM,NLM,NTM),
     2  NTRANS(NISM,NDM,NLM),MODE(NISM,NDM,NLM,NTM),
     3  IFLT(NISM,NDM,NLM,NTM)
C  ****  Output of ATREL.
      PARAMETER (NEMP=500)
      DIMENSION ES(NEMP),PAGES(NEMP),KPARS(NEMP),ITR(NEMP)
C
      SAVE IDAUGH,IBRANCH,ISA
C
      EXTERNAL RAND
C
      NR=0
      IF(IS.LT.1.OR.IS.GT.NIR) THEN
        WRITE(6,*) 'NIR =',NIR
        WRITE(6,2001) IS
 2001 FORMAT(1X,'The intended isotope, IS =',I3,' is not loaded.')
        IER=1
        LEVD=9999
        RETURN
      ELSE
        IER=0
      ENDIF
      DPAGE=0.0D0
C
C  ****  Sample the daughter nuclide.
C
      IF(METAST.NE.1) THEN
        ISA=IS
        METAST=0  ! Just for protection.
C
        NDHS=NDAUGH(IS)
        IF(NDHS.EQ.1) THEN
          IDAUGH=1
        ELSE  ! Binary search.
          RD=RAND(1.0D0)
          IF(RD.LT.PDAUGH(IS,1)) THEN
            IDAUGH=1
          ELSE
            IDAUGH=1
            I1=NDHS
 100        IT=(IDAUGH+I1)/2
            IF(RD.GT.PDAUGH(IS,IT)) THEN
              IDAUGH=IT
            ELSE
              I1=IT
            ENDIF
            IF(I1-IDAUGH.GT.1) GO TO 100
            IDAUGH=IDAUGH+1
          ENDIF
        ENDIF
C
C  ****  Select the disintegration branch to the daughter nuclide.
C
        NBRS=NBRANCH(IS,IDAUGH)
        IF(NBRS.EQ.1)THEN
          IBRANCH=1
        ELSE
          RD=RAND(2.0D0)
          IF(RD.LT.BR(IS,IDAUGH,1)) THEN
            IBRANCH=1
          ELSE  ! Binary search.
            IBRANCH=1
            I1=NBRS
 200        IT=(IBRANCH+I1)/2
            IF(RD.GT.BR(IS,IDAUGH,IT)) THEN
              IBRANCH=IT
            ELSE
              I1=IT
            ENDIF
            IF(I1-IBRANCH.GT.1) GO TO 200
            IBRANCH=IBRANCH+1
          ENDIF
        ENDIF
C
C  ****  Simulate the disintegration branch to the daughter nuclide.
C
        MODEBR=KTYPE(IS,IDAUGH,IBRANCH)
        IAD=IANUC(IS)
C
C  >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        SELECT CASE (MODEBR)
C  ----  Alpha emission.
        CASE(1)
C         WRITE(12,*) ' ALPHA emitted'
          IZD=IZNUC(IS)-2
          IAD=IAD-4
          NR=NR+1  ! Released particle.
          ENR(NR)=EBRANCH(IS,IDAUGH,IBRANCH)
          CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
          AGENR(NR)=DPAGE
          KPNR(NR)=4
          ITNR(NR)=-(IDAUGH*1000000+IBRANCH*1000+0)
C  ----  Beta- emission, energy sampled from the beta spectrum.
        CASE(2)
C         WRITE(12,*) ' BETA- emitted'
          IZD=IZNUC(IS)+1
          IDT=INDBETA(IS,IDAUGH,IBRANCH)
          EBETAS=SBETAS(IDT)
          NR=NR+1  ! Released particle.
          ENR(NR)=EBETAS
          CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
          AGENR(NR)=DPAGE
          KPNR(NR)=1
          ITNR(NR)=-(IDAUGH*1000000+IBRANCH*1000+1)
C  ----  Beta+ emission, energy sampled from the beta spectrum.
        CASE(3)
C         WRITE(12,*) ' BETA+ emitted'
          IZD=IZNUC(IS)-1
          IDT=INDBETA(IS,IDAUGH,IBRANCH)
          EBETAS=SBETAS(IDT)
          NR=NR+1  ! Released particle.
          ENR(NR)=EBETAS
          CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
          AGENR(NR)=DPAGE
          KPNR(NR)=3
          ITNR(NR)=-(IDAUGH*1000000+IBRANCH*1000+2)
C  ----  Electron capture K.
        CASE(4)
C         WRITE(12,*) ' Electron capture K'
          IZD=IZNUC(IS)-1
          CALL ATREL(IZD,1,ECNUC,ES,PAGES,KPARS,ITR,NS)
          IF(NS.GT.0) THEN
            PAGES1=1.0D35
            DO I=1,NS
              PAGES1=MIN(PAGES1,PAGES(I))
            ENDDO
            DO I=1,NS
              NR=NR+1  ! Released particle.
              ENR(NR)=ES(I)
              CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
              AGENR(NR)=DPAGE+PAGES(I)-PAGES1
              KPNR(NR)=KPARS(I)
              ITNR(NR)=ITR(I)
            ENDDO
          ENDIF
C  ----  Electron capture L1.
        CASE(5)
C         WRITE(12,*) ' Electron capture L1'
          IZD=IZNUC(IS)-1
          CALL ATREL(IZD,2,ECNUC,ES,PAGES,KPARS,ITR,NS)
          IF(NS.GT.0) THEN
            PAGES1=1.0D35
            DO I=1,NS
              PAGES1=MIN(PAGES1,PAGES(I))
            ENDDO
            DO I=1,NS
              NR=NR+1  ! Released particle.
              ENR(NR)=ES(I)
              CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
              AGENR(NR)=DPAGE+PAGES(I)-PAGES1
              KPNR(NR)=KPARS(I)
              ITNR(NR)=ITR(I)
            ENDDO
          ENDIF
C  ----  Electron capture L2.
        CASE(6)
C         WRITE(12,*) ' Electron capture L2'
          IZD=IZNUC(IS)-1
          CALL ATREL(IZD,3,ECNUC,ES,PAGES,KPARS,ITR,NS)
          IF(NS.GT.0) THEN
            PAGES1=1.0D35
            DO I=1,NS
              PAGES1=MIN(PAGES1,PAGES(I))
            ENDDO
            DO I=1,NS
              NR=NR+1  ! Released particle.
              ENR(NR)=ES(I)
              CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
              AGENR(NR)=DPAGE+PAGES(I)-PAGES1
              KPNR(NR)=KPARS(I)
              ITNR(NR)=ITR(I)
            ENDDO
          ENDIF
C  ----  Electron capture L3.
        CASE(7)
C         WRITE(12,*) ' Electron capture L3'
          IZD=IZNUC(IS)-1
          CALL ATREL(IZD,4,ECNUC,ES,PAGES,KPARS,ITR,NS)
          IF(NS.GT.0) THEN
            PAGES1=1.0D35
            DO I=1,NS
              PAGES1=MIN(PAGES1,PAGES(I))
            ENDDO
            DO I=1,NS
              NR=NR+1  ! Released particle.
              ENR(NR)=ES(I)
              CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
              AGENR(NR)=DPAGE+PAGES(I)-PAGES1
              KPNR(NR)=KPARS(I)
              ITNR(NR)=ITR(I)
            ENDDO
          ENDIF
C  ----  Electron capture M.
        CASE(8)
C         WRITE(12,*) ' Electron capture M'
          IZD=IZNUC(IS)-1
          CALL ATREL(IZD,5,ECNUC,ES,PAGES,KPARS,ITR,NS)
          IF(NS.GT.0) THEN
            PAGES1=1.0D35
            DO I=1,NS
              PAGES1=MIN(PAGES1,PAGES(I))
            ENDDO
            DO I=1,NS
              NR=NR+1  ! Released particle.
              ENR(NR)=ES(I)
              CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
              AGENR(NR)=DPAGE+PAGES(I)-PAGES1
              KPNR(NR)=KPARS(I)
              ITNR(NR)=ITR(I)
            ENDDO
          ENDIF
C  ----  Metastable level of parent nuclide.
        CASE(9)
C         WRITE(12,*) ' Parent nuclide in a metastable state'
          IZD=IZNUC(IS)
          IAD=IANUC(IS)
          METAST=1
        END SELECT
C  <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
C
C  ****  Level fed by the disintegration.
        LEVD=IFLB(IS,IDAUGH,IBRANCH)
        IF(LEVD.EQ.0) RETURN
      ELSE
        IS=ISA
      ENDIF

C
C  ****  Continue the de-excitation cascade of the daughter nucleus
C        until the ground level, or until the next metastable level.
C
      DO 400 WHILE (LEVD.GT.0)
C
        IF(METAST.EQ.1) THEN
          METAST=0  ! This is the start of a new cascade.
        ELSE
C  ----  Sample the de-excitation time of the level, except for levels
C        with extremely short life times.
          IF(DTIME(IS,IDAUGH,LEVD).GT.TIMELIM) THEN
            TAU=-LOG(RAND(3.0D0))*1.4426950408889634D0
     1         *DTIME(IS,IDAUGH,LEVD)
            IF(TAU.GT.DRTIME) THEN  ! The level is metastable.
              METAST=1
              RETURN
            ELSE
              DPAGE=DPAGE+TAU
              METAST=0
            ENDIF
          ELSE
            METAST=0
          ENDIF
        ENDIF
C
C  ****  Now, select the transition that leaves the level.
C
        NTRS=NTRANS(IS,IDAUGH,LEVD)
        IF(NTRS.EQ.1)THEN
          ITRANS=1
        ELSE  ! Binary search.
          RD=RAND(4.0D0)
          IF(RD.LT.PTRANS(IS,IDAUGH,LEVD,1)) THEN
            ITRANS=1
          ELSE
            ITRANS=1
            I1=NTRS
 300        IT=(ITRANS+I1)/2
            IF(RD.GT.PTRANS(IS,IDAUGH,LEVD,IT)) THEN
              ITRANS=IT
            ELSE
              I1=IT
            ENDIF
            IF(I1-ITRANS.GT.1) GO TO 300
            ITRANS=ITRANS+1
          ENDIF
        ENDIF
C  ----  A transition has been determined.
        ENERTR=ETRANS(IS,IDAUGH,LEVD,ITRANS)
        MODETR=MODE(IS,IDAUGH,LEVD,ITRANS)
        LEVD=IFLT(IS,IDAUGH,LEVD,ITRANS)
C
C  >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        SELECT CASE (MODETR)
C  ----  Gamma emission.
        CASE(1)
C         WRITE(12,*) ' + gamma emitted'
          NR=NR+1  ! Released particle.
          ENR(NR)=ENERTR
          CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
          AGENR(NR)=DPAGE
          KPNR(NR)=2
          ITNR(NR)=-(IDAUGH*1000000+IBRANCH*1000+3)
C  ----  K conversion electron.
        CASE(2)
C         WRITE(12,*) ' + K ic electron'
          NR=NR+1  ! Released particle.
          ENR(NR)=ENERTR
          CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
          AGENR(NR)=DPAGE
          KPNR(NR)=1
          ITNR(NR)=-(IDAUGH*1000000+IBRANCH*1000+4)
          CALL ATREL(IZD,1,ECNUC,ES,PAGES,KPARS,ITR,NS)
          IF(NS.GT.0) THEN
            DO I=1,NS
              NR=NR+1  ! Released particle.
              ENR(NR)=ES(I)
              CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
              AGENR(NR)=DPAGE+PAGES(I)
              KPNR(NR)=KPARS(I)
              ITNR(NR)=ITR(I)
            ENDDO
          ENDIF
C  ----  L1 conversion electron.
        CASE(3)
C         WRITE(12,*) ' + L1 ic electron'
          NR=NR+1  ! Released particle.
          ENR(NR)=ENERTR
          CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
          AGENR(NR)=DPAGE
          KPNR(NR)=1
          ITNR(NR)=-(IDAUGH*1000000+IBRANCH*1000+5)
          CALL ATREL(IZD,2,ECNUC,ES,PAGES,KPARS,ITR,NS)
          IF(NS.GT.0) THEN
            DO I=1,NS
              NR=NR+1  ! Released particle.
              ENR(NR)=ES(I)
              CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
              AGENR(NR)=DPAGE+PAGES(I)
              KPNR(NR)=KPARS(I)
              ITNR(NR)=ITR(I)
            ENDDO
          ENDIF
C  ----  L2 conversion electron.
        CASE(4)
C         WRITE(12,*) ' + L2 ic electron'
          NR=NR+1  ! Released particle.
          ENR(NR)=ENERTR
          CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
          AGENR(NR)=DPAGE
          KPNR(NR)=1
          ITNR(NR)=-(IDAUGH*1000000+IBRANCH*1000+6)
          CALL ATREL(IZD,3,ECNUC,ES,PAGES,KPARS,ITR,NS)
          IF(NS.GT.0) THEN
            DO I=1,NS
              NR=NR+1  ! Released particle.
              ENR(NR)=ES(I)
              CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
              AGENR(NR)=DPAGE+PAGES(I)
              KPNR(NR)=KPARS(I)
              ITNR(NR)=ITR(I)
            ENDDO
          ENDIF
C  ----  L3 conversion electron.
        CASE(5)
C         WRITE(12,*) ' + L3 ic electron'
          NR=NR+1  ! Released particle.
          ENR(NR)=ENERTR
          CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
          AGENR(NR)=DPAGE
          KPNR(NR)=1
          ITNR(NR)=-(IDAUGH*1000000+IBRANCH*1000+7)
          CALL ATREL(IZD,4,ECNUC,ES,PAGES,KPARS,ITR,NS)
          IF(NS.GT.0) THEN
            DO I=1,NS
              NR=NR+1  ! Released particle.
              ENR(NR)=ES(I)
              CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
              AGENR(NR)=DPAGE+PAGES(I)
              KPNR(NR)=KPARS(I)
              ITNR(NR)=ITR(I)
            ENDDO
          ENDIF
C  ----  M conversion electron.
        CASE(6)
C         WRITE(12,*) ' + MN ic electron'
          NR=NR+1  ! Released particle.
          ENR(NR)=ENERTR
          CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
          AGENR(NR)=DPAGE
          KPNR(NR)=1
          ITNR(NR)=-(IDAUGH*1000000+IBRANCH*1000+8)
          CALL ATREL(IZD,5,ECNUC,ES,PAGES,KPARS,ITR,NS)
          IF(NS.GT.0) THEN
            DO I=1,NS
              NR=NR+1  ! Released particle.
              ENR(NR)=ES(I)
              CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
              AGENR(NR)=DPAGE+PAGES(I)
              KPNR(NR)=KPARS(I)
              ITNR(NR)=ITR(I)
            ENDDO
          ENDIF
C  ----  Undefined transition. Assumed gamma to ground state.
        CASE(7)
C         WRITE(12,*) ' + Undefined transition'
          NR=NR+1  ! Released particle.
          ENR(NR)=ENERTR
          CALL ISOTROP(UNR(NR),VNR(NR),WNR(NR))
          AGENR(NR)=DPAGE
          KPNR(NR)=2
          ITNR(NR)=0
        END SELECT
C  <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
C
  400 CONTINUE
C
      RETURN
      END
C  *********************************************************************
C                       SUBROUTINE ISOTROP
C  *********************************************************************
      SUBROUTINE ISOTROP(U,V,W)
C  Gives the direction cosines of a random isotropic direction.
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      PARAMETER (PI=3.1415926535897932D0, TWOPI=2.0D0*PI)
      EXTERNAL RAND
      W=2.0D0*RAND(1.0D0)-1.0D0
      UV=SQRT(1.0D0-W*W)
      PHI=RAND(2.0D0)*TWOPI
      U=UV*COS(PHI)
      V=UV*SIN(PHI)
      RETURN
      END


C  *********************************************************************
C                       SUBROUTINE PENNUC0
C  *********************************************************************
      SUBROUTINE PENNUC0(NUCFNAME,EPMAX,IWR,IER)
C
C  Initialisation of decay simulation for the radionuclide defined by
C  the NUCLEIDE data file NUCFNAME.
C
C  PENNUC can handle up to NISM different nuclides simultaneously. To
C  simulate decays of several nuclides, call subroutine PENNUC0 for each
C  of them. Each radionuclide is assigned a label IS defined by the
C  order in which they are initialised. That is, IS=1, 2, ... for the
C  first, second, ... loaded nuclides. The sampling of the active
C  nuclide (according to the characteristics of the radioactive source)
C  must be performed in the calling program.
C
C  Input arguments:
C    NUCFNAME ... filename of the NUCLEIDE data file of the considered
C                 radionuclide.
C    IWR ........ input decay data are summarised in an output file,
C                 which must be open in the calling program as UNIT=IWR
C                 before the first call to PENNUC0.
C  Output values:
C    EPMAX ...... highest energy of photons and electrons released in
C                 the decay of the radionuclide (and of any other that
C                 has been previously loaded).
C    IER  ....... error code. When IER.NE.0 the initialisation has not
C                 succeeded; the cause of the problem is written in
C                 the standard output unit (6).
C
C  ---------------------------------------------------------------------
C  Quantities stored in memory:
C     NUCLIDE(I) ... name of the nuclide.
C     NDAUGH(I) ... number of daughters.
C
C  For each daughter ID,
C     PDAUGH(I,ID) ... relative disintegration probability to ID.
C     DNUCLIDE(I,ID) ... name of the daughter nuclide.
C     Q(I,ID) ... total available energy of the disintegration.
C     NLEVEL(I,ID) ... number of (excited) levels of the daughter
C        nuclide.
C     ELEVEL(I,ID,LEV) ... energies of the daughter nuclide levels,
C        sorted in increasing order of energies. The index LEV of the
C        ground level is equal to 0.
C     DTIME(I,ID,LEV) ... half-life of the level LEV.
C
C     NBRANCH(I,ID) ... number of disintegration branches.
C     KTYPE(I,ID,IBR) ... disintegration type (1-9)
C        1=ALPHA, 2=BETA-, 3=BETA+, 4=ECK, 5=ECL1, 6=ECL2, 7=ECL3,
C        8=ECLM, 9=META.
C     IFLB(I,ID,IBR) ... level fed of the daughter nuclide.
C     IPRBETA(I,ID,IBr) ... prohibition parameter defining the nature
C        of BETA branches (=0 for ALPHA and EC branches).
C           IPRBETA=0: Allowed and first forbidden transitions.
C           IPRBETA=1: Unique-first forbidden and second forbidden.
C           IPRBETA=2: Unique-second forbidden and third forbidden.
C           IPRBETA=3: Unique-third forbidden and fourth forbidden.
C     EBRANCH(I,ID,IBR) ... energy of the disintegration branch
C        (kinetic energy for alpha branches, maximum kinetic energy of
C        the electron/positron in beta branches, none for EC and META).
C     BR(I,ID,IBR) ... branching ratios, sum normalised to unity.
C
C     NTRANS(I,ID,LEV) ... number of transitions depopulating the level
C        LEV, including electron conversion,
C     PTRANS(I,ID,LEV,IT) ... probability of the transition IT that
C        depopulates the level LEV.
C     MODE(I,ID,LEV,IT) ... kind of transition (1-6):
C        1=GAMMA, 2=ECK, 3=ECL1, 4=ECL2, 5=ECL3, 6=ECM.
C     ETRANS(I,ID,LEV,IT) ... energy of the transition ITR.
C     IFLT(I,ID,LEV,IT) ... level fed by the transition IT.
C
C     The uncertainty of a quantity QQQQ is named UQQQQ. Uncertainties
C     are read from file NUCFNAME, but not used in the simulation.
C  ---------------------------------------------------------------------
C
      USE PENNUC_mod
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      CHARACTER NUCLIDE*7,DNUCLIDE*7,TEMPBUF*100,WCODE*3
      CHARACTER NUCFNAME*45
C
      CHARACTER WTYPE(9)*5
      DATA WTYPE/'ALPHA','BETA-','BETA+','ECK  ','ECL1 ','ECL2 ',
     1  'ECL3 ','ECM  ','METAS'/  ! EC = electron capture.
      CHARACTER WMODE(7)*5  ! IC = internal conversion.
      DATA WMODE/'GAMMA','ICK  ','ICL1 ','ICL2 ','ICL3 ','ICMN ',
     1  'NONE '/
      COMMON/CPN1/NUCLIDE(NISM),DNUCLIDE(NISM,NDM)
      COMMON/CPN2/KTYPE(NISM,NDM,NBM),NBRANCH(NISM,NDM),
     1  NLEVEL(NISM,NDM),IZNUC(NISM),IANUC(NISM),NDAUGH(NISM)
      COMMON/CPN3/Q(NISM,NDM),UQ(NISM,NDM),PDAUGH(NISM,NDM),
     1  UPDAUGH(NISM,NDM),ELEVEL(NISM,NDM,0:NTM),
     2  UELEVEL(NISM,NDM,0:NTM),DTIME(NISM,NDM,NLM),
     3  UDTIME(NISM,NDM,NLM)
      COMMON/CPN4/BR(NISM,NDM,NBM),UBR(NISM,NDM,NBM),
     1  EBRANCH(NISM,NDM,NBM),UEBRANCH(NISM,NDM,NBM),IFLB(NISM,NDM,NBM),
     2  IPRBETA(NISM,NDM,NBM),INDBETA(NISM,NDM,NBM),IDT
      COMMON/CPN5/PTRANS(NISM,NDM,NLM,NTM),UPTRANS(NISM,NDM,NLM,NTM),
     1  ETRANS(NISM,NDM,NLM,NTM),UETRANS(NISM,NDM,NLM,NTM),
     2  NTRANS(NISM,NDM,NLM),MODE(NISM,NDM,NLM,NTM),
     3  IFLT(NISM,NDM,NLM,NTM)
C
      IER=0
      NIR=NIR+1
      IF(NIR.GT.NISM) THEN
        WRITE(6,2001) NIR
        WRITE(IWR,2001) NIR
 2001 FORMAT(/1X,'PENNUC0: Too many radionuclides. NIR =',I3,
     1  /1X,'Please, increase the value of the parameter NISM.')
        IER=11
        RETURN
      ENDIF
      DO ID=1,NDM
        NBRANCH(NIR,ID)=0
        NLEVEL(NIR,ID)=0
        PDAUGH(NIR,ID)=0.0D0
        DO IT=0,NTM
          ELEVEL(NIR,ID,IT)=0.0D0
        ENDDO
        DO IB=1,NBM
          KTYPE(NIR,ID,IB)=0
          BR(NIR,ID,IB)=0.0D0
          EBRANCH(NIR,ID,IB)=0.0D0
          IFLB(NIR,ID,IB)=0
          IPRBETA(NIR,ID,IB)=0
          INDBETA(NIR,ID,IB)=0
        ENDDO
        DO IL=1,NLM
          NTRANS(NIR,ID,IL)=0
          DTIME(NIR,ID,IL)=0.0D0
          DO IT=1,NTM
            PTRANS(NIR,ID,IL,IT)=0.0D0
            ETRANS(NIR,ID,IL,IT)=0.0D0
            MODE(NIR,ID,IL,IT)=0
            IFLT(NIR,ID,IL,IT)=0
          ENDDO
        ENDDO
      ENDDO
C
      IF(NIR.EQ.1) THEN
        CALL ATREL0  ! Initialise the atomic relaxation routines.
        IDT=0
      ELSE
        WRITE(IWR,'(//,1X,76(''=''))')
      ENDIF
      EPMAX=0.0D0
C
      NRD=51
      OPEN(NRD,FILE=RPATH//NUCFNAME,STATUS='OLD',IOSTAT=KODE)
      IF(KODE.NE.0) THEN
        WRITE(IWR,2002) RPATH//NUCFNAME
        WRITE(6,2002) RPATH//NUCFNAME
 2002 FORMAT(/1X,'PENNUC0: The program could not open the file ',A,
     1  /1X,'Please, copy that file into the PENNUC data directory.')
        IER=12
        RETURN
      ENDIF
C
      CALL READSKP(NRD,WCODE,TEMPBUF)
      READ(TEMPBUF,*) NUCLIDE(NIR)
      CALL READSKP(NRD,WCODE,TEMPBUF)
      READ(TEMPBUF,*) IANUC(NIR),IZNUC(NIR)
      CALL READSKP(NRD,WCODE,TEMPBUF)
      READ(TEMPBUF,*) NDAUGH(NIR)
C
      IF(NIR.GT.1) THEN
        IZZ=IZNUC(NIR)
        IAA=IANUC(NIR)
        DO IS=1,NIR-1
          IF(IZZ.EQ.IZNUC(IS).AND.IAA.EQ.IANUC(IS)) THEN
            WRITE(IWR,2003) RPATH//NUCFNAME
            WRITE(6,2003) RPATH//NUCFNAME
 2003 FORMAT(/1X,'PENNUC0: The radioisotope decay file ',A,
     1  /10X,'has already been loaded.')
            IER=13
            RETURN
          ENDIF
        ENDDO
      ENDIF
C
      IZACT=IZNUC(NIR)
      CALL ATRELI(IZACT,0,IER)  ! Parent nucleus.
      IF(IER.NE.0) RETURN
C
      IF(IZNUC(NIR).GT.2) THEN
        IZACT=IZNUC(NIR)-2
        CALL ATRELI(IZACT,0,IER)  ! Daughter nucleus (alpha).
        IF(IER.NE.0) RETURN
      ENDIF
C
      IF(IZNUC(NIR).GT.1) THEN
        IZACT=IZNUC(NIR)-1
        CALL ATRELI(IZACT,0,IER)  ! Daughter (beta+ and el. capture).
        IF(IER.NE.0) RETURN
      ENDIF
C
      IZACT=IZNUC(NIR)+1
      CALL ATRELI(IZACT,0,IER)  ! Daughter nucleus (beta-).
      IF(IER.NE.0) RETURN
C
      DO 300 IDAUGH=1,NDAUGH(NIR)
        CALL READSKP(NRD,WCODE,TEMPBUF)
        READ(TEMPBUF,*) DNUCLIDE(NIR,IDAUGH)
        CALL READSKP(NRD,WCODE,TEMPBUF)
        READ(TEMPBUF,*) PDAUGH(NIR,IDAUGH),UPDAUGH(NIR,IDAUGH),
     1    NLEVEL(NIR,IDAUGH),NBRANCH(NIR,IDAUGH)
        CALL READSKP(NRD,WCODE,TEMPBUF)
        READ(TEMPBUF,*) Q(NIR,IDAUGH),UQ(NIR,IDAUGH)
        Q(NIR,IDAUGH)=Q(NIR,IDAUGH)*1.0D3  ! keV to eV.
        UQ(NIR,IDAUGH)=UQ(NIR,IDAUGH)*1.0D3
C
        DO 100 IB=1,NBRANCH(NIR,IDAUGH)
          CALL READSKP(NRD,WCODE,TEMPBUF)
          READ(TEMPBUF,*) BR(NIR,IDAUGH,IB),UBR(NIR,IDAUGH,IB),
     1      IFLB(NIR,IDAUGH,IB),EBRANCH(NIR,IDAUGH,IB),
     2      UEBRANCH(NIR,IDAUGH,IB),ITPRO
C  The sign of ITPRO distinguishes non-unique beta transitions.
          IPRBETA(NIR,IDAUGH,IB)=ABS(ITPRO)
C
          IF(WCODE.EQ.'ALP') KTYPE(NIR,IDAUGH,IB)=1
          IF(WCODE.EQ.'BEM') KTYPE(NIR,IDAUGH,IB)=2
          IF(WCODE.EQ.'BEP') KTYPE(NIR,IDAUGH,IB)=3
          IF(WCODE.EQ.'CK ') KTYPE(NIR,IDAUGH,IB)=4
          IF(WCODE.EQ.'CL1') KTYPE(NIR,IDAUGH,IB)=5
          IF(WCODE.EQ.'CL2') KTYPE(NIR,IDAUGH,IB)=6
          IF(WCODE.EQ.'CL3') KTYPE(NIR,IDAUGH,IB)=7
          IF(WCODE.EQ.'CM ') KTYPE(NIR,IDAUGH,IB)=8
          IF(WCODE.EQ.'MS ') KTYPE(NIR,IDAUGH,IB)=9  ! Metastable level.
          IF(WCODE.EQ.'CL ') KTYPE(NIR,IDAUGH,IB)=5  ! EL ~ EL1.
          IF(WCODE.EQ.'CN ') KTYPE(NIR,IDAUGH,IB)=8  ! EN ~ EM.
C
          EBRANCH(NIR,IDAUGH,IB)=EBRANCH(NIR,IDAUGH,IB)*1.0D3
          UEBRANCH(NIR,IDAUGH,IB)=UEBRANCH(NIR,IDAUGH,IB)*1.0D3

          IZD=IZNUC(NIR)
          IF(KTYPE(NIR,IDAUGH,IB).EQ.2) THEN
            IDT=IDT+1
            INDBETA(NIR,IDAUGH,IB)=IDT
            IZD=IZNUC(NIR)+1
            IMASS=IANUC(NIR)
            CALL SBETA0(IMASS,IZD,EBRANCH(NIR,IDAUGH,IB),
     1        IPRBETA(NIR,IDAUGH,IB),IDT,IER)
            IF(IER.NE.0) RETURN
          ENDIF
C
          IF(KTYPE(NIR,IDAUGH,IB).EQ.3) THEN
            IDT=IDT+1
            INDBETA(NIR,IDAUGH,IB)=IDT
            IZD=-(IZNUC(NIR)-1)
            IMASS=IANUC(NIR)
            CALL SBETA0(IMASS,IZD,EBRANCH(NIR,IDAUGH,IB),
     1        IPRBETA(NIR,IDAUGH,IB),IDT,IER)
            IF(IER.NE.0) RETURN
          ENDIF
 100    CONTINUE
C
        DO 200 IL=NLEVEL(NIR,IDAUGH),1,-1
          CALL READSKP(NRD,WCODE,TEMPBUF)
          READ(TEMPBUF,*) ELEVEL(NIR,IDAUGH,IL),UELEVEL(NIR,IDAUGH,IL),
     1      NTRANS(NIR,IDAUGH,IL),DTIME(NIR,IDAUGH,IL),
     2      UDTIME(NIR,IDAUGH,IL)
C
          ELEVEL(NIR,IDAUGH,IL)=ELEVEL(NIR,IDAUGH,IL)*1.0D3
          UELEVEL(NIR,IDAUGH,IL)=UELEVEL(NIR,IDAUGH,IL)*1.0D3
C  ****  Levels with no defined transitions.
          IF(NTRANS(NIR,IDAUGH,IL).LT.1) THEN
            PTRANS(NIR,IDAUGH,IL,1)=1.0d0
            UPTRANS(NIR,IDAUGH,IL,1)=0.0D0
            ETRANS(NIR,IDAUGH,IL,1)=ELEVEL(NIR,IDAUGH,IL)
            UETRANS(NIR,IDAUGH,IL,1)=UELEVEL(NIR,IDAUGH,IL)
            NTRANS(NIR,IDAUGH,IL)=1
            MODE(NIR,IDAUGH,IL,1)=7
            IFLT(NIR,IDAUGH,IL,1)=0
            GO TO 200
          ENDIF
C
          DO IT=NTRANS(NIR,IDAUGH,IL),1,-1
            CALL READSKP(NRD,WCODE,TEMPBUF)
            READ(TEMPBUF,*) PTRANS(NIR,IDAUGH,IL,IT),
     1        UPTRANS(NIR,IDAUGH,IL,IT),ETRANS(NIR,IDAUGH,IL,IT),
     2        UETRANS(NIR,IDAUGH,IL,IT),IFLT(NIR,IDAUGH,IL,IT)
C
            IF(WCODE.EQ.'GA ') MODE(NIR,IDAUGH,IL,IT)=1
            IF(WCODE.EQ.'EK ') MODE(NIR,IDAUGH,IL,IT)=2
            IF(WCODE.EQ.'EL1') MODE(NIR,IDAUGH,IL,IT)=3
            IF(WCODE.EQ.'EL2') MODE(NIR,IDAUGH,IL,IT)=4
            IF(WCODE.EQ.'EL3') MODE(NIR,IDAUGH,IL,IT)=5
            IF(WCODE.EQ.'EM ') MODE(NIR,IDAUGH,IL,IT)=6
            IF(WCODE.EQ.'EL ') MODE(NIR,IDAUGH,IL,IT)=3  ! EL ~ EL1.
            IF(WCODE.EQ.'EN ') MODE(NIR,IDAUGH,IL,IT)=6  ! EN ~ EM.
C
            ETRANS(NIR,IDAUGH,IL,IT)=ETRANS(NIR,IDAUGH,IL,IT)*1.D3
            UETRANS(NIR,IDAUGH,IL,IT)=UETRANS(NIR,IDAUGH,IL,IT)*1.D3
          ENDDO
 200    CONTINUE
 300  CONTINUE
      CLOSE(NRD)
C
C  ************  Write nuclear data.
C
      WRITE(IWR,1001) NIR,NUCLIDE(NIR),IZNUC(NIR),IANUC(NIR)
 1001 FORMAT(' Radioisotope number',I3,
     1 /,' Parent nuclide:  ',A,' (Z =',I3,', A = ',I3,')')
      WRITE(IWR,'(A,I3)') ' Number of daughters =',NDAUGH(NIR)
      WRITE(IWR,'(/,A)')
     1  ' EC = electron capture,  IC = internal conversion.'
      WRITE(IWR,'(A)') ' BR = branching ratio,   PR = probability.'
      DO ID=1,NDAUGH(NIR)
        IF(KTYPE(NIR,ID,1).EQ.1) THEN  ! ALPHA
          IZD=IZNUC(NIR)-2
          IAD=IANUC(NIR)-2
        ELSE IF(KTYPE(NIR,ID,1).EQ.2) THEN  ! BETA-
          IZD=IZNUC(NIR)+1
          IAD=IANUC(NIR)
        ELSE IF(KTYPE(NIR,ID,1).EQ.9) THEN  ! Metastable
          IZD=IZNUC(NIR)
          IAD=IANUC(NIR)
        ELSE  ! BETA+ and EC
          IZD=IZNUC(NIR)-1
          IAD=IANUC(NIR)
        ENDIF
        WRITE(IWR,1002) ID,DNUCLIDE(NIR,ID),IZD,IAD,Q(NIR,ID),
     1    PDAUGH(NIR,ID)
 1002   FORMAT(1X,76('-'),/1X,'Daughter nuclide number',I4,':  ',A7,
     1    ' (Z =',I3,', A = ',I3,')',/3X,'Q =',1P,E13.6,
     2    ' eV,  BR =',E13.6)
C
        IF(NLEVEL(NIR,ID).GT.0) THEN
          WRITE(IWR,'(/,'' Excited levels:'')')
          DO IL=NLEVEL(NIR,ID),1,-1
            WRITE(IWR,1003) IL,ELEVEL(NIR,ID,IL),DTIME(NIR,ID,IL)
 1003       FORMAT(1X,I3,',  E =',1P,E13.6,
     1        ' eV,  half-life =',E13.6,' s')
          ENDDO
        ENDIF
C
        WRITE(IWR,'(/,'' Disintegration branches:'')')
        DO IB=1,NBRANCH(NIR,ID)
          IFLB(NIR,ID,IB)=MIN(IFLB(NIR,ID,IB),NLEVEL(NIR,ID)) ! Caution
          IF(KTYPE(NIR,ID,IB).EQ.1) THEN  ! Alpha decay.
            WRITE(IWR,1004) IB,WTYPE(KTYPE(NIR,ID,IB)),IFLB(NIR,ID,IB),
     1        BR(NIR,ID,IB),EBRANCH(NIR,ID,IB)
 1004       FORMAT(1X,I3,',  ',A5,' to level ',I3,',  BR =',1P,E13.6,
     1        ',  E =',E13.6,' eV')
          ELSE IF(KTYPE(NIR,ID,IB).EQ.2) THEN  ! Beta- decay.
            WRITE(IWR,1005) IB,WTYPE(KTYPE(NIR,ID,IB)),IFLB(NIR,ID,IB),
     1        BR(NIR,ID,IB),EBRANCH(NIR,ID,IB),IPRBETA(NIR,ID,IB)
 1005       FORMAT(1X,I3,',  ',A5,' to level ',I3,',  BR =',1P,E13.6,
     1        ',  E =',E13.6,' eV, Ipro =',I2)
            EPMAX=MAX(EPMAX,EBRANCH(NIR,ID,IB))
          ELSE IF(KTYPE(NIR,ID,IB).EQ.3) THEN  ! Beta+ decay.
            WRITE(IWR,1005) IB,WTYPE(KTYPE(NIR,ID,IB)),IFLB(NIR,ID,IB),
     1        BR(NIR,ID,IB),EBRANCH(NIR,ID,IB),IPRBETA(NIR,ID,IB)
C  ----  Positrons eventually give annihilation gamma rays. The maximum
C        energy of annihilation photons is .lt. 1.21*(E0+me*c^2).
            EPMAX=MAX(EPMAX,1.21D0*(EBRANCH(NIR,ID,IB)+5.12D5))
          ELSE  ! Electron capture and metastable nuclides.
            WRITE(IWR,1006) IB,WTYPE(KTYPE(NIR,ID,IB)),IFLB(NIR,ID,IB),
     1        BR(NIR,ID,IB)
 1006       FORMAT(1X,I3,',  ',A5,' to level ',I3,',  BR =',1P,E13.6)
          ENDIF
        ENDDO
C
        IF(NLEVEL(NIR,ID).GT.0) THEN
          DO IL=NLEVEL(NIR,ID),1,-1
            WRITE(IWR,1007) IL
 1007       FORMAT(/1X,'Transitions from level ',I3)
            DO IT=1,NTRANS(NIR,ID,IL)
              WRITE(IWR,1008) IL,IFLT(NIR,ID,IL,IT),
     1          WMODE(MODE(NIR,ID,IL,IT)),
     2          ETRANS(NIR,ID,IL,IT),PTRANS(NIR,ID,IL,IT)
 1008         FORMAT(1X,I3,' --> ',I3,1X,A5,',  E =',
     1          1P,E13.6,' eV,  PR =',E13.6)
              EPMAX=MAX(EPMAX,ETRANS(NIR,ID,IL,IT))
            ENDDO
          ENDDO
        ENDIF
      ENDDO
C
C  ****  Accumulated daugther probabilities.
C
      PTOT=0.0D0
      DO ID=1,NDAUGH(NIR)
        PTOT=PTOT+PDAUGH(NIR,ID)
      ENDDO
      DO ID=1,NDAUGH(NIR)
        PDAUGH(NIR,ID)=PDAUGH(NIR,ID)/PTOT
        UPDAUGH(NIR,ID)=UPDAUGH(NIR,ID)/PTOT
      ENDDO
      WRITE(IWR,'(1X,76(''-''))')
      WRITE(IWR,'(A)') ' Branching ratios to daughters (normalised):'
      WRITE(IWR,'(1P,5E13.6)') (PDAUGH(NIR,ID),ID=1,NDAUGH(NIR))
      WRITE(IWR,1009) EPMAX
 1009 FORMAT(/1X,'Maximum energy of released particles = ',1P,E13.6,
     1  ' eV')
      WRITE(IWR,'(/,A)') ' ---  END  ---'
C
      IF(NDAUGH(NIR).GT.1) THEN
        DO ID=2,NDAUGH(NIR)
          PDAUGH(NIR,ID)=PDAUGH(NIR,ID)+PDAUGH(NIR,ID-1)
        ENDDO
        PDAUGH(NIR,NDAUGH(NIR))=1.0D0
      ENDIF
C
C  ****  Accumulated branch probabilities.
C
      DO ID=1,NDAUGH(NIR)
        BRTOT=0.0D0
        DO IB=1,NBRANCH(NIR,ID)
          BRTOT=BRTOT+BR(NIR,ID,IB)
        ENDDO
        DO IB=1,NBRANCH(NIR,ID)
          BR(NIR,ID,IB)=BR(NIR,ID,IB)/BRTOT
          UBR(NIR,ID,IB)=UBR(NIR,ID,IB)/BRTOT
        ENDDO
C
        IF(NBRANCH(NIR,ID).GT.1)THEN
          DO IB=2,NBRANCH(NIR,ID)
            BR(NIR,ID,IB)=BR(NIR,ID,IB)+BR(NIR,ID,IB-1)
          ENDDO
        ENDIF
        BR(NIR,ID,NBRANCH(NIR,ID))=1.0D0
      ENDDO
C
C  ****  Accumulated level depopulation probabilities.
C
      DO ID=1,NDAUGH(NIR)
        DO IL=1,NLEVEL(NIR,ID)
          PTOT=0.0D0
          DO IT=1,NTRANS(NIR,ID,IL)
            PTOT=PTOT+PTRANS(NIR,ID,IL,IT)
          ENDDO
          DO IT=1,NTRANS(NIR,ID,IL)
            PTRANS(NIR,ID,IL,IT)=PTRANS(NIR,ID,IL,IT)/PTOT
            UPTRANS(NIR,ID,IL,IT)=UPTRANS(NIR,ID,IL,IT)/PTOT
          ENDDO
C
          IF(NTRANS(NIR,ID,IL).GT.1)THEN
            DO IT=2,NTRANS(NIR,ID,IL)
              PTRANS(NIR,ID,IL,IT)=PTRANS(NIR,ID,IL,IT)
     1          +PTRANS(NIR,ID,IL,IT-1)
            ENDDO
          ENDIF
          PTRANS(NIR,ID,IL,NTRANS(NIR,ID,IL))=1.0d0
        ENDDO
      ENDDO
C
      IZD=0
      IAD=0
      LEVD=9999
      RETURN
      END
C  *********************************************************************
C                       SUBROUTINE READSKP
C  *********************************************************************
      SUBROUTINE READSKP(NRD,WCODE,TEMPBUF)
C
C  Reads and reformats an input record. Skips comment records, and in
C  data records replaces any ';' by ','.
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      CHARACTER TEMPBUF*100,WCODE*3
C
      WCODE='COM'
      DO WHILE(WCODE.EQ.'COM')
        READ(NRD,'(A3,1X,A100)') WCODE,TEMPBUF
      ENDDO
C
      IPOS=1
      DO WHILE (IPOS.GT.0)
        IPOS=SCAN(TEMPBUF,';')
        IF(IPOS.GT.0) TEMPBUF(IPOS:IPOS)=','
      ENDDO
C
      RETURN
      END


C  *********************************************************************
C                       SUBROUTINE ATREL
C  *********************************************************************
      SUBROUTINE ATREL(IZ,IS,ECNUC,ES,PAGES,KPARS,ITR,NS)
C
C  This subroutine simulates the relaxation of a singly ionised atom of
C  the element IZ with a vacancy in the IS subshell (the K shell or an
C  L, M or N subshell). This initial vacancy is filled by electrons from
C  outer subshells through radiative and non-radiative transitions,
C  which may produce additional vacancies.
C
C  Input arguments:
C     IZ ......... atomic number of the ionised element.
C     IS ......... electron subshell containing the initial vacancy.
C     ECNUC ...... cutoff energy for atomic transitions.
C
C  Output arguments:
C     ES(500) .... energies of emitted particles (in eV).
C     KPARS(500) ... kind of emitted particle (1=electron; 2=photon).
C     PAGES(500) ... emission times (seconds), time elapsed since the
C                  generation of the initial vacancy.
C     ITR(500) ... transition (coded as in PENELOPE's ILB(4) flag).
C     NS ......... number of particles in the relaxation cascade.
C
C  We use the following notation to designate the possible transitions:
C  *  Radiative: IS0-IS1 (an electron from the IS1 subshell fills the
C     vacancy in the IS0 subshell, leaving a hole in the IS1 subshell).
C  *  Non-radiative: IS0-IS1-IS2 (an electron from the IS1 subshell
C     fills the vacancy in the IS0 subshell, and the released energy is
C     taken away by an electron in the IS2 subshell; this process leaves
C     holes, in the IS1 and IS2 subshells).
C  The de-excitation cascade (i.e. the set of transitions that occur for
C  a given initial vacancy) is sampled from the transition probabilities
C  extracted from the Livermore Evaluated Atomic Data Library (EADL).
C  The energies of the particles emitted in each transition are read
C  from the PENELOPE database.
C
C  The simulation of the de-excitation cascade is discontinued either
C  when the K to N subshells have been filled up or when there is not
C  enough energy to produce 'active' radiation (with energy larger than
C  ECNUC).
C
C  De-excitation data for the loaded elements are stored in the common
C  block /CATREL/, in a form designed to minimise the amount of memory
C  and to facilitate the random sampling. The quantities in the common
C  block are the following:
C  IFIRST(99,16) ... de-excitation data for a vacancy in the subshell IS
C     of the element IZ start at the position K=IFIRST(IZ,IS) in the
C     storage arrays. The allowed values for IS are 1 to 16 (K shell
C     and L, M and N subshells).
C  ILAST(99,16) ... the de-excitation data for a vacancy in the subshell
C     IS of the element IZ end at the position K=ILAST(IZ,IS) in the
C     storage arrays.
C  IS1(K), IS2(K) ... subshells that are active in the transition (see
C     the subshell label code below). For radiative transitions,
C     IS2(K)=0.
C  P(K) ... relative probability for the transition IS-IS1(K)-IS2(K).
C  ET(K) ... energy of the secondary particle emitted in the transition.
C  F(K), IAL(K) ... cutoff and alias values (Walker's sampling method).
C
C  ---------------------------------------------------------------------
C  Label code IS for electron subshells:
C      1 = K  (1s1/2),     11 = N2 (4p1/2),     21 = O5 (5d5/2),
C      2 = L1 (2s1/2),     12 = N3 (4p3/2),     22 = O6 (5f5/2),
C      3 = L2 (2p1/2),     13 = N4 (4d3/2),     23 = O7 (5f7/2),
C      4 = L3 (2p3/2),     14 = N5 (4d5/2),     24 = P1 (6s1/2),
C      5 = M1 (3s1/2),     15 = N6 (4f5/2),     25 = P2 (6p1/2),
C      6 = M2 (3p1/2),     16 = N7 (4f7/2),     26 = P3 (6p3/2),
C      7 = M3 (3p3/2),     17 = O1 (5s1/2),     27 = P4 (6d3/2),
C      8 = M4 (3d3/2),     18 = O2 (5p1/2),     28 = P5 (6d5/2),
C      9 = M5 (3d5/2),     19 = O3 (5p3/2),     29 = Q1 (7s1/2),
C     10 = N1 (4s1/2),     20 = O4 (5d3/2),     30 = outer shells.
C  ---------------------------------------------------------------------
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
C  ****  Atomic relaxation data.
      PARAMETER (NRX=60000,NSHP1=31)
      COMMON/CATREL/EBR(99,NSHP1),P(NRX),ET(NRX),F(NRX),IAL(NRX),
     1       IS0(NRX),IS1(NRX),IS2(NRX),IFIRST(99,16),ILAST(99,16),
     2       NCUR
      COMMON/CATRE2/ALW(99,30)
C  ****  Subhell occupancies, and auxiliary arrays.
      DIMENSION OCUP(30),ISV(30),PTIME(16,256)
      DATA OCUP/2.0D0,2.0D0,2.0D0,4.0D0,2.0D0,2.0D0,4.0D0,4.0D0,6.0D0,
     1    2.0D0,2.0D0,4.0D0,4.0D0,6.0D0,6.0D0,8.0D0,2.0D0,2.0D0,4.0D0,
     2    4.0D0,6.0D0,6.0D0,8.0D0,2.0D0,2.0D0,4.0D0,4.0D0,6.0D0,2.0D0,
     3    1.0D9/
C
      PARAMETER (NEMP=500)
      DIMENSION ES(NEMP),PAGES(NEMP),KPARS(NEMP),ITR(NEMP)
C
      EXTERNAL RAND
C
C  ****  Initialisation.
C
      NS=0
      IF(IZ.LT.3.OR.IS.GT.16) RETURN
C  ****  If the subshell ionisation energy is less than ECNUC, the
C        cascade is not followed.
      IF(EBR(IZ,IS).LT.ECNUC) RETURN
      PAGE0=0.0D0
C
      NV=0
      DO I=1,30
        ISV(I)=0
      ENDDO
      ISA=IS
      PAGE=PAGE0
C
C  ****  Next transition.
C
 1    CONTINUE
      KF=IFIRST(IZ,ISA)
      KL=ILAST(IZ,ISA)
      IF(KL.LE.KF) RETURN
C  ****  Walker's sampling algorithm.
 2    CONTINUE
      RN=RAND(1.0D0)*(KL-KF+1)
      K1=INT(RN)
      TST=RN-K1
      IF(TST.GT.F(KF+K1)) THEN
        KS=IAL(KF+K1)
      ELSE
        KS=KF+K1
      ENDIF
C
      JS1=IS1(KS)
      JS2=IS2(KS)
C  ****  Vacancies in the intervening subshells.
      IF(ISV(JS1).GT.0) THEN
        FREJ=(OCUP(JS1)-ISV(JS1))/OCUP(JS1)
      ELSE
        FREJ=1.0D0
      ENDIF
      IF(JS2.GT.0) THEN
        FREJ=FREJ*((OCUP(JS2)-ISV(JS2))/OCUP(JS2))
      ENDIF
      IF(FREJ.LT.1.0D0) THEN
        IF(RAND(2.0D0).GT.FREJ) GO TO 2
      ENDIF
C  ****  The particle age is evaluated.
      IF(ISV(ISA).GT.1) THEN
        PAGE=PAGE-(ALW(IZ,ISA)/DBLE(ISV(ISA)))*LOG(RAND(3.0D0))
      ELSE
        PAGE=PAGE-ALW(IZ,ISA)*LOG(RAND(3.0D0))
      ENDIF
C
C  ****  Fluorescence radiation.
C
      IF(JS2.EQ.0) THEN
        IF(JS1.LT.17) THEN
          IF(EBR(IZ,JS1).GT.ECNUC) THEN
            NV=NV+1
            ISV(JS1)=ISV(JS1)+1
            PTIME(JS1,ISV(JS1))=PAGE
          ENDIF
        ENDIF
        IF(ET(KS).GT.ECNUC) THEN
          NS=NS+1
          IF(NS.GT.NEMP) THEN
            WRITE(6,*) 'NS =',NS
            WRITE(6,'(A)')
     1        'PENNUC warning: Dimension NEMP exceeded (1).'
            RETURN
          ENDIF
          ES(NS)=ET(KS)
          KPARS(NS)=2
          ITR(NS)=IZ*1000000+ISA*10000+JS1*100
          PAGES(NS)=PAGE
        ENDIF
      ELSE
        IF(JS1.LT.17) THEN
          IF(EBR(IZ,JS1).GT.ECNUC) THEN
            NV=NV+1
            ISV(JS1)=ISV(JS1)+1
            PTIME(JS1,ISV(JS1))=PAGE
          ENDIF
        ENDIF
        IF(JS2.LT.17) THEN
          IF(EBR(IZ,JS2).GT.ECNUC) THEN
            NV=NV+1
            ISV(JS2)=ISV(JS2)+1
            PTIME(JS2,ISV(JS2))=PAGE
          ENDIF
        ENDIF
        IF(ET(KS).GT.ECNUC) THEN
          NS=NS+1
          IF(NS.GT.NEMP) THEN
            WRITE(6,*) 'NS =',NS
            WRITE(6,'(A)')
     1        'PENNUC warning: Dimension NEMP exceeded (2).'
            RETURN
          ENDIF
          ES(NS)=ET(KS)
          KPARS(NS)=1
          ITR(NS)=IZ*1000000+ISA*10000+JS1*100+JS2
          PAGES(NS)=PAGE
        ENDIF
      ENDIF
C
C  ****  Are there any K-, L-, M- or N-vacancies unfilled?
C
      IF(NV.GT.0) THEN
 3      CONTINUE
        IF(EBR(IZ,ISA).LT.ECNUC) RETURN
        IF(ISV(ISA).GT.0) THEN  ! Vacancies in the current subshell.
          PAGE=PTIME(ISA,ISV(ISA))
          ISV(ISA)=ISV(ISA)-1
          NV=NV-1
          GO TO 1
        ELSE
          DO IST=ISA+1,16  ! Outer subshells.
            IF(ISV(IST).GT.0) THEN
              ISA=IST
              GO TO 3
            ENDIF
          ENDDO
        ENDIF
      ENDIF
C
      RETURN
      END
C  *********************************************************************
C                       SUBROUTINE ATREL0
C  *********************************************************************
      SUBROUTINE ATREL0
C
C  This subroutine sets all static values in common /CATREL/ to zero.
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
C  ****  Atomic relaxation data.
      PARAMETER (NRX=60000,NSHP1=31)
      COMMON/CATREL/EBR(99,NSHP1),P(NRX),ET(NRX),F(NRX),IAL(NRX),
     1       IS0(NRX),IS1(NRX),IS2(NRX),IFIRST(99,16),ILAST(99,16),
     2       NCUR
C
      DO I=1,99
        DO J=1,16
          IFIRST(I,J)=0
          ILAST(I,J)=0
        ENDDO
        DO J=1,NSHP1
          EBR(I,J)=0.0D0
        ENDDO
      ENDDO
C
      DO I=1,NRX
        P(I)=0.0D0
        ET(I)=0.0D0
        F(I)=0.0D0
        IAL(I)=0
        IS0(I)=0
        IS1(I)=0
        IS2(I)=0
      ENDDO
      NCUR=0
      RETURN
      END
C  *********************************************************************
C                       SUBROUTINE ATRELI
C  *********************************************************************
      SUBROUTINE ATRELI(IZ,IWR,IER)
C
C  This subroutine produces a table of atomic relaxation data for the
C  element IZ.
C
C  Input arguments:
C    IZ ...... atomic number.
C    IWR ..... relaxation data are written in a file (UNIT=IWR) named
C              'atreli.dat' if IWR is greater than zero. It is assumed
C              that unit IWR is not open in the calling program.
C  Output value:
C    IER  .... error code. When IER.NE.0 the initialisation has not
C              succeeded; the cause of the problem is written in
C              the standard output unit (6).
C
C  Data are read from file 'pdrelax.p11', which contains data pertaining
C  to singly ionised atoms with the initial vacancy in one of the K, L,
C  M and N shells. This file was prepared from the Livermore Evaluated
C  Atomic Data Library (EADL). The energies of x-ray lines were replaced
C  by more accurate experimental and theoretical values given by
C  Deslattes et al. (2003) -K and L shells- and by Burr (1967) -M and N
C  shells.
C
      USE PENNUC_mod
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      PARAMETER (HBAR=6.58211928D-16)  ! Planck's constant (eV*s)
      CHARACTER CH5*5,CH2*2,FNAME*45
      CHARACTER*5 CSH5(30)
      DIMENSION EB(30),IFI(30),IKS(30)
C
      DATA CSH5/'1s1/2','2s1/2','2p1/2','2p3/2','3s1/2','3p1/2',
     1          '3p3/2','3d3/2','3d5/2','4s1/2','4p1/2','4p3/2',
     2          '4d3/2','4d5/2','4f5/2','4f7/2','5s1/2','5p1/2',
     3          '5p3/2','5d3/2','5d5/2','5f5/2','5f7/2','6s1/2',
     4          '6p1/2','6p3/2','6d3/2','6d5/2','7s1/2',' free'/
C  ****  Atomic relaxation data.
      PARAMETER (NRX=60000,NSHP1=31)
      COMMON/CATREL/EBR(99,NSHP1),P(NRX),ET(NRX),F(NRX),IAL(NRX),
     1       IS0(NRX),IS1(NRX),IS2(NRX),IFIRST(99,16),ILAST(99,16),
     2       NCUR
      COMMON/CATRE2/ALW(99,30)
C
      PARAMETER (NTRAN=2500)
      DIMENSION JS0(NTRAN),JS1(NTRAN),JS2(NTRAN),PR(NTRAN),ER(NTRAN),
     1       WW(NTRAN),FF(NTRAN),KK(NTRAN),IORD(NTRAN),ISR(NTRAN),EE(30)
C
      CHARACTER*2 LSHELL(0:30)
      DATA LSHELL/'  ','K ','L1','L2','L3','M1','M2','M3','M4','M5',
     1   'N1','N2','N3','N4','N5','N6','N7','O1','O2','O3','O4',
     1   'O5','O6','O7','P1','P2','P3','P4','P5','Q1','X '/
C
C  ****  Check if this element's data have already been loaded.
C
      IER=0
      IF(IFIRST(IZ,1).NE.0) then
        WRITE(6,*) 'ATRELI: The element has already been loaded.'
        RETURN
      ENDIF
C
      IF(IWR.GT.0) OPEN(IWR,FILE='atreli.dat')
C
C  ****  Loads atomic configuration and subshell binding energies.
C
      FNAME='pdatconf.p14'
      MRD=52
      OPEN(MRD,FILE=RPATH//FNAME,STATUS='OLD',IOSTAT=KODE)
      IF(KODE.NE.0) THEN
        IF(IWR.GT.0) WRITE(IWR,2001) RPATH//FNAME
        WRITE(6,2001) RPATH//FNAME
 2001 FORMAT(/1X,'ATRELI: The program could not open the file ',A,
     1  /1X,'Please, copy that file into the PENNUC data directory.')
        IER=21
        RETURN
      ENDIF
C
      DO J=1,22
        READ(MRD,'(A5)') CH5
      ENDDO
      DO IS=1,30
        IKS(IS)=0
      ENDDO
      NS=0
      IZT=0
      DO J=1,150000
        READ(MRD,2002,END=100) IIZ,IS,CH2,CH5,IIF,EIE,CCP,GA1,GA2
 2002   FORMAT(I3,I4,1X,A2,1X,A5,I3,4E13.5)
        IF(IIZ.EQ.IZ) THEN
          NS=NS+1
          IF(NS.GT.30) THEN
            WRITE(6,'(/1X,''NS ='',I4)') NS
            WRITE(6,'(A)') 'ATRELI: Too many shells.'
            IER=22
            RETURN
          ENDIF
          IF(IS.LT.1.OR.IS.GT.30) THEN
            WRITE(6,'(/1X,''IS ='',I4)') IS
            WRITE(6,'(A)') 'ATRELI: Wrong shell number.'
            IER=23
            RETURN
          ENDIF
          IZT=IZT+ABS(IIF)
          EB(IS)=EIE
          IF(GA2.GT.0.0D0) THEN
            ALW(IZ,IS)=HBAR/GA2
          ELSE IF(GA1.GT.0.0D0) THEN
            ALW(IZ,IS)=HBAR/GA1
          ELSE
            ALW(IZ,IS)=0.0D0
          ENDIF
          IFI(IS)=ABS(IIF)
          IKS(NS)=IS
        ENDIF
      ENDDO
 100  CONTINUE
      NSHT=NS
      IF(IZ.NE.IZT) THEN
        WRITE(6,*) 'IZ, IZT=',IZ,IZT
        WRITE(6,'(A)') 'ATRELI: Unbalanced charges (element).'
        IER=24
        RETURN
      ENDIF
      CLOSE(MRD)
C
      FNAME='pdrelax.p11'
      OPEN(MRD,FILE=RPATH//FNAME,STATUS='OLD',IOSTAT=KODE)
      IF(KODE.NE.0) THEN
        IF(IWR.GT.0) WRITE(IWR,2003) RPATH//FNAME
        WRITE(6,2003) RPATH//FNAME
 2003 FORMAT(/1X,'ATRELI: The program could not open the file ',A,
     1  /1X,'Please, copy that file into the PENNUC data directory.')
        IER=25
        RETURN
      ENDIF
C
      NT=0
      READ(MRD,*,END=200) IZR,IS0R
      DO I=1,150000
        READ(MRD,*,END=200) IZR,IS0R,IS1R,IS2R,PPR,EIN
        IF(IZR.EQ.IZ) THEN
          NT=NT+1
          JS0(NT)=IS0R
          JS1(NT)=IS1R
          JS2(NT)=IS2R
          PR(NT)=PPR
          ER(NT)=EIN
        ENDIF
      ENDDO
 200  CONTINUE
      CLOSE(MRD)
C
      IF(IWR.GT.0) WRITE(IWR,2004) IZ,NSHT,NT
 2004 FORMAT(1X,'*** ATREL:  Z =',I3,',  No. of shells =',I3,
     1          ',  no. of transitions =',I4)
      NSHR=NSHT
      IF(NSHR.GT.NSHP1-1) THEN
        IF(IWR.GT.0) WRITE(IWR,*)
     1    'Insufficient memory storage in ATRELI.'
        IF(IWR.GT.0) WRITE(IWR,*)
     1    'Increase the value of the parameter NSHP1 to ',NSHR+1
        WRITE(6,*) 'Insufficient memory storage in ATRELI.'
        WRITE(6,*) 'Increase the value of the parameter NSHP1 to ',
     1    NSHR+1
        IER=26
        WRITE(6,'(A)') 'ATRELI: Insufficient memory storage.'
        RETURN
      ENDIF
C
      IS=0
      DO J=1,30
        KS=IKS(J)
        IF(KS.GT.0) THEN
          IF(IFI(KS).NE.0) THEN
            IF(IWR.GT.0) WRITE(IWR,2005) KS,CSH5(KS),IFI(KS),EB(KS)
 2005       FORMAT(1X,I3,1X,A5,1X,I1,1P,E16.8)
            IS=IS+1
            ISR(IS)=KS
            EE(IS)=EB(KS)
          ENDIF
        ENDIF
      ENDDO
C
      IF(NT.GT.0) THEN
        DO I=1,NT
          IF(JS2(I).EQ.0) THEN
            IF(ER(I).LT.1.0D0) THEN
              ETT=EE(JS0(I))-EE(JS1(I))
            ELSE
              ETT=ER(I)
            ENDIF
          ELSE
            IF(ER(I).LT.1.0D0) THEN
              ETT=EE(JS0(I))-EE(JS1(I))-EE(JS2(I))
            ELSE
              ETT=ER(I)
            ENDIF
          ENDIF
          ETT=MAX(ETT,1.0D0)
          IF(IWR.GT.0) WRITE(IWR,2006) JS0(I),JS1(I),JS2(I),PR(I),ETT
          ER(I)=ETT
        ENDDO
 2006   FORMAT(1X,3I3,1P,2E16.8)
      ENDIF
C
C  ****  Check if this element's data have already been loaded.
C
      IF(IFIRST(IZ,1).NE.0) RETURN
C
      EBR(IZ,31)=DBLE(NSHR+1.0D-12)
      DO IS=1,NSHR
        EBR(IZ,ISR(IS))=EE(IS)
      ENDDO
      IF(NT.EQ.0) THEN
        DO IS=1,16
          NCUR=NCUR+1
          IFIRST(IZ,IS)=NCUR
          ILAST(IZ,IS)=NCUR
          P(NCUR)=1.0D0
          ET(NCUR)=0.0D0
          F(NCUR)=1.0D0
          IAL(NCUR)=NCUR+1
          IS0(NCUR)=1
          IS1(NCUR)=1
          IS2(NCUR)=1
        ENDDO
        RETURN
      ENDIF
C
C  ****  Walker's aliasing.
C
      DO IS=1,16
        N=0
        DO J=1,NT
          IF(JS0(J).EQ.IS) THEN
            N=N+1
            IORD(N)=J
            WW(N)=PR(J)
          ENDIF
        ENDDO
        IF(N.GT.1) THEN
          CALL IRND0(WW,FF,KK,N)
          IFIRST(IZ,IS)=NCUR+1
          ILAST(IZ,IS)=NCUR+N
          DO L=1,N
            P(NCUR+L)=WW(L)
            ET(NCUR+L)=ER(IORD(L))
            F(NCUR+L)=FF(L)
            IAL(NCUR+L)=IFIRST(IZ,IS)+KK(L)-1
            IS0(NCUR+L)=JS0(IORD(L))
            IS1(NCUR+L)=JS1(IORD(L))
            IS2(NCUR+L)=JS2(IORD(L))
          ENDDO
          NCUR=NCUR+N
        ELSE
          NCUR=NCUR+1
          IFIRST(IZ,IS)=NCUR
          ILAST(IZ,IS)=NCUR
          P(NCUR)=1.0D0
          ET(NCUR)=ER(1)
          F(NCUR)=1.0D0
          IAL(NCUR)=NCUR
          IS0(NCUR)=JS0(1)
          IS1(NCUR)=JS1(1)
          IS2(NCUR)=JS2(1)
        ENDIF
      ENDDO
C
C  ****  Verify that transition probabilities are correctly reproduced.
C
      TST=0.0D0
      DO IS=1,16
        I0=IFIRST(IZ,IS)
        IN=ILAST(IZ,IS)
        PT=0.0D0
        DO I=I0,IN
          PT=PT+P(I)
        ENDDO
        DO I=I0,IN
          PPI=0.0D0
          DO J=I0,IN
            IF(IAL(J).EQ.I) PPI=PPI+(1.0D0-F(J))
          ENDDO
          PPI=(PPI+F(I))/DBLE(IN-I0+1)
          TST=MAX(TST,ABS(1.0D0-PPI*PT/P(I)))
        ENDDO
      ENDDO
      IF(TST.GT.1.0D-12) THEN
        WRITE(6,*) 'TST =',TST
        WRITE(6,'(A)') 'ATRELI. Rounding error is too large.'
        IER=27
        RETURN
      ENDIF
C
      IF(IWR.GT.0) CLOSE(IWR)
      RETURN
      END


C  *********************************************************************
C                        FUNCTION SBETAS
C  *********************************************************************
      FUNCTION SBETAS(IDT)
C
C  Random sampling of the initial beta ray energy for the IDT beta
C  transition.
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
C
      PARAMETER (NN=64,ID=200)
      COMMON/CBET1/X(NN,ID),A(NN,ID),B(NN,ID),F(NN,ID),IA(NN,ID),
     1  NPM1(ID)
C
      EXTERNAL RAND
C  ****  Selection of the interval (Walker's aliasing).
      RN=RAND(1.0D0)*NPM1(IDT)+1.0D0
      K=INT(RN)
      TST=RN-K
      IF(TST.LT.F(K,IDT)) THEN
        I=K
        RR=TST
        D=F(K,IDT)
      ELSE
        I=IA(K,IDT)
        RR=TST-F(K,IDT)
        D=1.0D0-F(K,IDT)
      ENDIF
C  ****  Sampling from the rational inverse cumulative distribution.
      IF(RR.GT.1.0D-12) THEN
        SBETAS=X(I,IDT)+((1.0D0+A(I,IDT)+B(I,IDT))*D*RR
     1    /(D*D+(A(I,IDT)*D+B(I,IDT)*RR)*RR))*(X(I+1,IDT)-X(I,IDT))
      ELSE
        SBETAS=X(I,IDT)+RAND(2.0D0)*(X(I+1,IDT)-X(I,IDT))
      ENDIF
C
      RETURN
      END
C  *********************************************************************
C                       SUBROUTINE SBETA0
C  *********************************************************************
      SUBROUTINE SBETA0(IMASS,IZ,EMAX,IPRO,IDT,IER)
C
C    Initialisation of the algorithm for random sampling of the initial
C  kinetic energy of a beta ray according to the theoretical Fermi spec-
C  trum.
C
C  Input parameters:
C    IMASS ... nuclear mass number of the daughter nucleus.
C    IZ ...... nuclear charge number (atomic number) of the daughter
C              nucleus, with negative sign for positron emitters.
C    EMAX .... maximum kinetic energy (in eV) of beta rays.
C    IPRO .... prohibition parameter
C              IPRO=0: Allowed and first forbidden transitions.
C              IPRO=1: Unique-first forbidden and second forbidden.
C              IPRO=2: Unique-second forbidden and third forbidden.
C              IPRO=3: Unique-third forbidden and fourth forbidden.
C    IDT ..... transition identification number (.LE. 100)
C
C  Output value:
C    IER  .... error code. When IER.NE.0 the initialisation has not
C              succeeded; the cause of the problem is written in
C              the standard output unit (6).
C
C  After invoking subroutine SBETA0 in the form
C        CALL SBETA0(IMASS,ZI,EMAX,IPRO,IDT)
C  the statement
C        ERND=SBETAS(IDT)
C  produces a random value ERND of the initial energy of the beta ray.
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      COMMON/BETAS/EMAXC,IMASSC,IZC,IPROC
C
      PARAMETER (NN=64,ID=200)
      COMMON/CBET1/X(NN,ID),A(NN,ID),B(NN,ID),F(NN,ID),IA(NN,ID),
     1  NPM1(ID)
C
      PARAMETER (NM=512)
      COMMON/CRITAA/XA(NM),AA(NM),BA(NM),FA(NM),JA(NM),NPM1A
C
      EXTERNAL FERMI
C
      IF(IDT.GT.ID) THEN
        WRITE(6,*) 'SBETA0. The input value of IDT is too large.'
        WRITE(6,*) 'Increase the value of the paramete ID to ',IDT
        IER=31
        RETURN
      ENDIF
C
      IER=0
      IMASSC=IMASS
      IZC=IZ
      EMAXC=EMAX
      IPROC=IPRO
C
      XL=0.0D0
      XU=EMAX
      CALL RITA0(FERMI,XL,XU,NN,NN/4,ERRM,0,IER)
      IF(IER.NE.0) RETURN
      NPM1(IDT)=NPM1A
      DO I=1,NPM1A+1
        X(I,IDT)=XA(I)
        A(I,IDT)=AA(I)
        B(I,IDT)=BA(I)
        F(I,IDT)=FA(I)
        IA(I,IDT)=JA(I)
      ENDDO
      RETURN
      END
C  *********************************************************************
C                       FUNCTION FERMI
C  *********************************************************************
      FUNCTION FERMI(E0)
C
C  ****  Computation of the (unnormalised) Fermi beta spectrum.
C        Written by E. Garcia-Torano. December 1990.
C
C  ****  Beta emitter data (input through common /BETAS/):
C  IMASS ... mass number of the daughter nucleus.
C  IZ ...... atomic number of the daughter nucleus, with negative sign
C            for positron emitters.
C  EMAX .... maximum kinetic energy (in eV) of beta rays.
C  IPRO .... prohibition parameter.
C            IPRO=0: allowed and first forbidden transitions.
C            IPRO=1: unique-first forbidden and second forbidden.
C            IPRO=2: unique-second forbidden and third forbidden.
C            IPRO=3: unique-third forbidden and fourth forbidden.
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      COMMON/BETAS/EMAX,IMASS,IZ,IPRO
      PARAMETER (PI=3.1415926535897932D0)
C
      ALPHA=1.0D0/137.036D0
      AZ=ALPHA*IZ
      GAM=SQRT(1.0D0-AZ**2)
      A13=DBLE(IMASS)**0.333333333333333D0
      RAD=(1.123D0*A13-0.941D0/A13)/386.144D0
      EXPN=2.0D0*(GAM-1.0D0)
      W0=EMAX/511.0D3+1.0D0
      CTE=(2.0D0*RAD)**EXPN
C
      W=E0/511.0D3+1.0D0
      P=SQRT(W**2-1.0D0)
      EXPN1=2.0D0*GAM-1.0D0
      RL0=1.0D0-7.0D0*AZ**2/20.0D0-28.0D0*(W*RAD*AZ)/15.0D0
     1  -8.0D0*(RAD*AZ)/(15.0D0*W)-(P*RAD)**2/3.0D0
     2  -AZ**4/5.0D0+AZ*(W*RAD)**3*(1.0D0+4.5D0*AZ)
     3  -W*RAD*AZ**3/2.0D0+(AZ**6)/20.0D0-3.0D0*AZ**8/8.0D0
     4  -3.0D0*W*RAD*AZ**4/8.0D0+10.0D0*(W*RAD)**2*AZ**6
      IF(E0.GT.1.0D-6*EMAX) GO TO 10
C
C  ************  Energy = 0.
C
C  ****  Positron emitters.
      IF(IZ.LE.0) THEN
        FERMI=0.0D0
        RETURN
      ENDIF
C  ****  Negatron emitters.
      ARGIN=2.0D0*GAM+1.0D0
      YY=EXPN1*LOG(AZ)-2.0D0*GAMMAL(ARGIN)
      FERMI=8.0D0*PI*CTE*(EMAX/511.0D3)**2*RL0
     1  *PROHIB(IPRO,W,W0)*EXP(YY)
      RETURN
C
C  ************  Energy .GT. 0.
C
 10   CONTINUE
      YY=AZ*W/P
      XX=EXPN1*LOG(P)+PI*YY+RATIOG(GAM,YY)
      FAFERM=4.0D0*RL0*CTE*EXP(XX)
      FERMI=(W0-W)**2*W*PROHIB(IPRO,W,W0)*FAFERM
      RETURN
      END
C  *********************************************************************
C                       FUNCTION PROHIB
C  *********************************************************************
      FUNCTION PROHIB(IPRO,W,W0)
C
C  Computes the shape factor from the following expressions :
C  - Allowed and first forbidden transitions (IPRO=0),
C            F = 1
C  - Second and unique-first forbidden (IPRO=1),
C            F = p^2 + q^2
C  - Third and unique-second forbidden (IPRO=2),
C            F = p^4 + q^4 + 10(p^2*q^2)/3.
C  - Fourth and unique-third forbidden (IPRO=3),
C            F = p^6 + q^6 + 7 p^2 q^2 (p^2+q^2)
C
C  Improved shape factors for some nuclides that do not follow the
C  standard expressions are defined explicitly.
C
C  Version 2.0                          E.Garca-Torao, April 2016.
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      COMMON/BETAS/EMAX,IMASS,IZ,IPRO2
C
      P2=W**2-1.0D0
      Q2=(W0-W)**2
C
C  ****  Nuclides with special shape factors.
C
C C-11
      IF(IZ.EQ.-5.AND.IMASS.EQ.11) THEN
        PROHIB=1.0D0-0.0074D0*W
        RETURN
      ENDIF
C N-13
      IF(IZ.EQ.-6.AND.IMASS.EQ.13) THEN
        PROHIB=1.0D0+0.0014D0*W
        RETURN
      ENDIF
C F-18
      IF(IZ.EQ.-8.AND.IMASS.EQ.18) THEN
        PROHIB=1.0D0+0.0034D0*W
         RETURN
      ENDIF
C NA-22
      IF(IZ.EQ.-10.AND.IMASS.EQ.22) THEN
        PROHIB=1.0D0-0.005D0*W
        RETURN
      ENDIF
C NA-24
      IF(IZ.EQ.12.AND.IMASS.EQ.24) THEN
        PROHIB=1.0D0-0.011D0*W
        RETURN
      ENDIF
C CL-36
      IF(IZ.EQ.18.AND.IMASS.EQ.36) THEN
        PROHIB=1.0D0-0.282715D0*W-0.045988D0/W-0.491739D0*W**2
     1    +0.438731D0*W**3
        RETURN
      ENDIF
C P-32
      IF(IZ.EQ.16.AND.IMASS.EQ.32) THEN
        PROHIB=1.0D0-0.019D0*W
        RETURN
      ENDIF
C K-40
      IF(IZ.EQ.20.AND.IMASS.EQ.40) THEN
        PROHIB=1.05D0*Q2**3+6.3D0*P2*Q2**2+6.25D0*Q2*P2**2
     1    +0.95D0*P2**3
        RETURN
      ENDIF
C SC-49
      IF(IZ.EQ.22.AND.IMASS.EQ.49) THEN
        PROHIB=1.0D0-0.008D0*W
        RETURN
      ENDIF
C NI-63
      IF(IZ.EQ.22.AND.IMASS.EQ.63) THEN
        PROHIB=1.0D0+1213.49D0*W-232.009D0/W
     1    -1517.16D0*W**2+532.69D0*W**3
        RETURN
      ENDIF
C GA-68
      IF(IZ.EQ.-30.AND.IMASS.EQ.68) THEN
        PROHIB=1.0D0-0.01D0*W
        RETURN
      ENDIF
C SR-89
      IF(IZ.EQ.39.AND.IMASS.EQ.89) THEN
        PROHIB=(P2+Q2)*(1.0D0-0.0112D0*W)
        RETURN
      ENDIF
C SR-90
      IF(IZ.EQ.39.AND.IMASS.EQ.90) THEN
        PROHIB=(P2+Q2)*(1.0D0-0.054D0*W)
        RETURN
      ENDIF
C Y-90
      IF(IZ.EQ.40.AND.IMASS.EQ.90) THEN
        PROHIB=(P2+Q2)*(1.0D0-0.0114D0*W)
        RETURN
      ENDIF
C ZR-89
      IF(IZ.EQ.-39.AND.IMASS.EQ.89) THEN
        PROHIB=1.0D0-0.39D0*W+0.09D0*W**2
        RETURN
      ENDIF
C TC-99
      IF(IZ.EQ.44.AND.IMASS.EQ.99) THEN
        PROHIB=0.529D0*P2+Q2
        RETURN
      ENDIF
C I-129
      IF(IZ.EQ.54.AND.IMASS.EQ.129) THEN
        PROHIB=P2+3.16D0*Q2
        RETURN
      ENDIF
C LA-138
      IF(IZ.EQ.58.AND.IMASS.EQ.138) THEN
        PROHIB=1.0D0+407.71D0*W-50.695D0/W-583.794D0*W**2
     1    +246.279D0*W**3
        RETURN
      ENDIF
C CE-141
      IF(IZ.EQ.59.AND.IMASS.EQ.141) THEN
        PROHIB=1.0D0-0.28D0*W
        RETURN
      ENDIF
C RE-186
      IF(IZ.EQ.56.AND.IMASS.EQ.186) THEN
        PROHIB=1.0D0+0.038D0*(P2+Q2)
        RETURN
      ENDIF
C TL-204
      IF(IZ.EQ.82.AND.IMASS.EQ.204) THEN
        PROHIB=1.097D0*Q2+P2
        RETURN
      ENDIF
C BI-210
      IF(IZ.EQ.84.AND.IMASS.EQ.210) THEN
        PROHIB=1.0D0-0.47D0*W+0.065D0*W**2
        RETURN
      ENDIF
C PU-241
      IF(IZ.EQ.95.AND.IMASS.EQ.241) THEN
        PROHIB=1.0D0-1.9582D0*W+0.96078D0*W**2
        RETURN
      ENDIF
C
C  ****  Nuclides with standard shape factors.
C
      IF(IPRO.EQ.0)THEN
        PROHIB=1.0D0
      ELSE IF(IPRO.EQ.1) THEN
        PROHIB=P2+Q2
      ELSE IF(IPRO.EQ.2) THEN
        PROHIB=P2**2+Q2**2+(10.0D0*P2*Q2)/3.
      ELSE IF(IPRO.EQ.3) THEN
        PROHIB=P2**3+Q2**3+7.0D0*P2*Q2*(P2+Q2)
      ELSE
        PROHIB=1.0D0
      ENDIF
C
      RETURN
      END
C  *********************************************************************
C                       FUNCTION RATIOG
C  *********************************************************************
      FUNCTION RATIOG(G,DNU)
C
C  This function calculates the square modulus of the ratio
C              GAMMA(G+CI*DNU))/GAMMA(2*G+1),
C  where GAMMA stands for the complex gamma function and CI is the
C  imaginary unit.
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      PARAMETER (PI=3.1415926535897932D0)
      RATIOG=0.0D0
      A=6.0D0/(G+5.0D0)
      DO J=1,5
        Y1=A*DNU
        S1=(Y1**2+(J-1)**2)/((G+J-1)**2+DNU**2)
        RATIOG=RATIOG+LOG(S1)
      ENDDO
      Y1=A*DNU
      IF((PI*Y1).GT.30.0D0) THEN
        S1=LOG(2.0D0*PI/Y1)-PI*Y1
      ELSE
        S1=LOG(PI/(Y1*DSINH(PI*Y1)))
      ENDIF
      S2=25.0D0+Y1**2
      RATIOG=RATIOG+S1+LOG(S2)
      S1=(5.0D0+G)**2+DNU**2
      DCOC=DNU/(5.0D0+G)
      S1=2.0D0-LOG(S1)+2.0D0*DNU*ATAN(DCOC)/(5.0D0+G)
     1  +1.0D0/(S1*6.0D0*A)
      RATIOG=RATIOG+(1.0D0-G)*S1-11.0D0*LOG(A)-2.0D0*GAMMAL(G+G+1.0D0)
      RETURN
      END
C  *********************************************************************
C                       FUNCTION GAMMAL
C  *********************************************************************
      FUNCTION GAMMAL(R)
C
C  This function gives LOG(GAMMA(R)) for real arguments R.
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
      GAMMAL=80.5D0
      RA=R
      FAC=1.0D0
      IF(ABS(RA).GT.10.0D0) GO TO 2
    1 CONTINUE
      FAC=FAC/RA
      RA=RA+1.0D0
      IF(ABS(RA).LT.1.0D-35) RETURN
      IF(ABS(RA).LT.10.0D0) GO TO 1
C  ****  Stirling expansion.
    2 CONTINUE
      RI2=1.0D0/(RA*RA)
      RS=(43867.0D0/244188.0D0)*RI2
      RS=(RS-3617.0D0/122400.0D0)*RI2
      RS=(RS+1.0D0/156.0D0)*RI2
      RS=(RS-691.0D0/360360.0D0)*RI2
      RS=(RS+1.0D0/1188.0D0)*RI2
      RS=(RS-1.0D0/1680.0D0)*RI2
      RS=(RS+1.0D0/1260.0D0)*RI2
      RS=(RS-1.0D0/360.0D0)*RI2
      RS=(RS+1.0D0/12.0D0)/RA
      GAMMAL=((RA-0.5D0)*LOG(RA)-RA+9.1893853320467274D-1+RS)+LOG(FAC)
      RETURN
      END


!!>>>> penEasy: The following routines, commented out, are already in PENELOPE
!! CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!! CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!! C
!! C  The following function RAND and the subroutines IRND0, RITA0, and
!! C  RITAI0 are from the RITA subroutine package of the PENELOPE code
!! C  system.
!! C
!! C  *********************************************************************
!! C                         FUNCTION RAND (Random number generator)
!! C  *********************************************************************
!!       FUNCTION RAND(DUMMY)
!! C
!! C  This is an adapted version of subroutine RANECU written by F. James
!! C  (Comput. Phys. Commun. 60 (1990) 329-344), which has been modified to
!! C  give a single random number at each call.
!! C
!! C  The 'seeds' ISEED1 and ISEED2 must be initialised in the main program
!! C  and transferred through the named common block /RSEED/.
!! C
!! C  Some compilers incorporate an intrinsic random number generator with
!! C  the same name (but with different argument lists). To avoid conflict,
!! C  it is advisable to declare RAND as an external function in all sub-
!! C  programs that call it.
!! C
!!       IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
!!       PARAMETER (USCALE=1.0D0/2.147483563D9)
!!       COMMON/RSEED/ISEED1,ISEED2
!! C
!!       I1=ISEED1/53668
!!       ISEED1=40014*(ISEED1-I1*53668)-I1*12211
!!       IF(ISEED1.LT.0) ISEED1=ISEED1+2147483563
!! C
!!       I2=ISEED2/52774
!!       ISEED2=40692*(ISEED2-I2*52774)-I2*3791
!!       IF(ISEED2.LT.0) ISEED2=ISEED2+2147483399
!! C
!!       IZ=ISEED1-ISEED2
!!       IF(IZ.LT.1) IZ=IZ+2147483562
!!       RAND=IZ*USCALE
!! C
!!       RETURN
!!       END
!! C  *********************************************************************
!! C                       SUBROUTINE IRND0
!! C  *********************************************************************
!!       SUBROUTINE IRND0(W,F,K,N)
!! C
!! C  Initialisation of Walker's aliasing algorithm for random sampling
!! C  from discrete probability distributions.
!! C
!! C  Input arguments:
!! C    N ........ number of different values of the random variable.
!! C    W(1:N) ... corresponding point probabilities (not necessarily
!! C               normalised to unity).
!! C  Output arguments:
!! C    F(1:N) ... cutoff values.
!! C    K(1:N) ... alias values.
!! C
!!       IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
!!       DIMENSION W(N),F(N),K(N)
!!       COMMON/CRITAN/CNORM  ! Normalising constant, output.
!! C  ****  Renormalisation.
!!       CNORM=0.0D0
!!       DO I=1,N
!!         CNORM=CNORM+MAX(W(I),0.0D0)
!!       ENDDO
!!       CNORM=1.0D0/CNORM
!!       FACT=DBLE(N)*CNORM
!!       DO I=1,N
!!         K(I)=I
!!         F(I)=MAX(W(I),0.0D0)*FACT
!!       ENDDO
!!       IF(N.EQ.1) RETURN
!! C  ****  Cutoff and alias values.
!!       DO I=1,N-1
!!         HLOW=1.0D0
!!         HIGH=1.0D0
!!         ILOW=0
!!         IHIGH=0
!!         DO J=1,N
!!           IF(K(J).EQ.J) THEN
!!             IF(F(J).LT.HLOW) THEN
!!               HLOW=F(J)
!!               ILOW=J
!!             ELSE IF(F(J).GT.HIGH) THEN
!!               HIGH=F(J)
!!               IHIGH=J
!!             ENDIF
!!           ENDIF
!!         ENDDO
!!         IF(ILOW.EQ.0.OR.IHIGH.EQ.0) RETURN
!!         K(ILOW)=IHIGH
!!         F(IHIGH)=HIGH+HLOW-1.0D0
!!       ENDDO
!!       RETURN
!!       END
!! C  *********************************************************************
!! C                       SUBROUTINE RITA0
!! C  *********************************************************************
!!       SUBROUTINE RITA0(PDF,XLOW,XHIGH,N,NU,ERRM,IWR,IER)
!! C
!! C  Initialisation of the RITA algorithm for random sampling of a
!! C  continuous random variable X from a probability distribution function
!! C  PDF(X) defined in the interval (XLOW,XHIGH). The external function
!! C  PDF(X) --not necessarily normalised-- must be provided by the user.
!! C  N is the number of points in the sampling grid. These points are
!! C  determined by means of an adaptive strategy that minimizes local
!! C  interpolation errors. The first NU grid points are uniformly spaced
!! C  in (XLOW,XHIGH); when NU is negative, the initial grid consists of
!! C  -NU points logarithmically spaced (in this case, XLOW must be
!! C  non-negative).
!! C
!! C  ERRM is a measure of the interpolation error (the largest value of
!! C  the absolute error of the rational interpolation integrated over each
!! C  grid interval).
!! C
!! C  IER is an error code. An output value different from zero indicates
!! C  that the initialization procedure has not succeeded; the reason is
!! C  written in the standard output unit (6).
!! C
!! C  ****  Interpolation coefficients and PDF tables are printed on
!! C        separate files (UNIT=IWR) if IWR is greater than zero.
!! C
!! C  Other subprograms needed: EXTERNAL function PDF,
!! C                            subroutines RITAI0 and IRND0.
!! C
!!
!!       IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
!!       PARAMETER (NM=512)
!!       COMMON/CRITA/XT(NM),PAC(NM),DPAC(NM),A(NM),B(NM),IL(NM),IU(NM),
!!      1  NPM1
!!       COMMON/CRITAA/XA(NM),AA(NM),BA(NM),FA(NM),IA(NM),NPM1A
!!       COMMON/CRITAN/CNORM  ! Normalising constant, output.
!!       EXTERNAL PDF
!! C  ****  Initialisation of the RITA algorithm.
!!       IER=0
!!       CALL RITAI0(PDF,XLOW,XHIGH,N,NU,ERRM,IWR,IER)
!!       IF(IER.NE.0) RETURN
!! C  ****  Walker's aliasing; cutoff and alias values.
!!       NPM1A=NPM1
!!       DO I=1,NPM1
!!         XA(I)=XT(I)
!!         AA(I)=A(I)
!!         BA(I)=B(I)
!!       ENDDO
!!       XA(NPM1+1)=XT(NPM1+1)
!!       SAVE=CNORM
!!       CALL IRND0(DPAC,FA,IA,NPM1A)
!!       CNORM=SAVE
!!       FA(NPM1+1)=1.0D0
!!       IA(NPM1+1)=NPM1A
!!       RETURN
!!       END
!! C  *********************************************************************
!! C                       SUBROUTINE RITAI0
!! C  *********************************************************************
!!       SUBROUTINE RITAI0(PDF,XLOW,XHIGH,N,NU,ERRM,IWR,IER)
!! C
!! C  Initialisation of the RITA algorithm for random sampling of a
!! C  continuous random variable X from a probability distribution function
!! C  PDF(X) defined in the interval (XLOW,XHIGH). The external function
!! C  PDF(X) --not necessarily normalised-- must be provided by the user.
!! C  N is the number of points in the sampling grid. These points are
!! C  determined by means of an adaptive strategy that minimizes local
!! C  interpolation errors. The first NU grid points are uniformly spaced
!! C  in (XLOW,XHIGH); when NU is negative, the initial grid consists of
!! C  -NU points logarithmically spaced (in this case, XLOW must be
!! C  nonnegative).
!! C
!! C  ERRM is a measure of the interpolation error (the largest value of
!! C  the absolute error of the rational interpolation integrated over each
!! C  grid interval).
!! C
!! C  IER is an error code. An output value different from zero indicates
!! C  that the initialization procedure has not succeeded; the reason is
!! C  written in the standard output unit (6).
!! C
!! C  ****  Interpolation coefficients and PDF tables are printed on
!! C        separate files (UNIT=IWR) if IWR is greater than zero.
!! C
!! C  Other subprograms needed: EXTERNAL function PDF.
!! C
!!       IMPLICIT DOUBLE PRECISION (A-H,O-Z), INTEGER*4 (I-N)
!!       PARAMETER (EPS=1.0D-10, ZERO=1.0D-75, ZEROT=0.1D0*ZERO)
!!       PARAMETER (NM=512)
!! C
!! C  The information used by the sampling function RITAI is exported
!! C  through the following common block,
!!       COMMON/CRITA/XT(NM),PAC(NM),DPAC(NM),A(NM),B(NM),IL(NM),IU(NM),
!!      1  NPM1
!! C  where
!! C    XT(I) ..... grid points, in increasing order.
!! C    PAC(I) .... value of the cumulative pdf at XT(I).
!! C    DPAC(I) ... probability of the I-th interval.
!! C    A(I), B(I) ... rational inverse cumulative distribution parameters.
!! C    IL(I) .... largest J for which PAC(J) < (I-1)/(NP-1).
!! C    IU(I) .... smallest J for which PAC(J) > I/(NP-1).
!! C    NPM1 ..... numner of grid points minus one, NP-1 (8.LE.NP.LE.NM).
!! C
!!       COMMON/CRITAN/CNORM
!! C    CNORM .... normalising constant of the external PDF, output.
!! C
!!       DIMENSION ERR(NM),C(NM)
!!       PARAMETER (NIP=51)
!!       DIMENSION XS(NIP),YS(NIP),SUMI(NIP)
!!       EXTERNAL PDF
!! C
!!       IF(N.LT.9) THEN
!!         WRITE(6,'('' Error in RITAI0: N must be larger than 8.'',
!!      1    /,'' N='',I11)') N
!!         IER=1
!!         RETURN
!!       ENDIF
!!       IF(N.GT.NM) THEN
!!         WRITE(6,'('' Error in RITAI0: N must be less than NM=512.'',
!!      1    /,'' N='',I11)') N
!!         IER=2
!!         RETURN
!!       ENDIF
!!       IF(XLOW.GT.XHIGH-EPS) THEN
!!         WRITE(6,'('' Error in RITAI0: XLOW must be larger than XHIGH.'',
!!      1    /,'' XLOW='',1P,E13.6,'', XHIGH ='',E13.6)') XLOW,XHIGH
!!         IER=3
!!         RETURN
!!       ENDIF
!!       IER=0
!! C
!! C  ****  We start with a grid of NUNIF points uniformly spaced in the
!! C        interval (XLOW,XHIGH).
!! C
!!       IF(NU.GE.0) THEN
!!         NUNIF=MIN(MAX(8,NU),N/2)
!!         NP=NUNIF
!!         DX=(XHIGH-XLOW)/DBLE(NP-1)
!!         XT(1)=XLOW
!!         DO I=1,NP-1
!!           XT(I+1)=XLOW+I*DX
!!         ENDDO
!!         XT(NP)=XHIGH
!!       ELSE
!! C  ****  If NU.LT.0,the NUNIF points are logarithmically spaced.
!! C        XLOW must be greater than or equal to zero.
!!         NUNIF=MIN(MAX(8,-NU),N/2)
!!         NP=NUNIF
!!         IF(XLOW.LT.0.0D0) THEN
!!           WRITE(6,'('' Error in RITAI0: XLOW and NU are negative.'',
!!      1    /,'' XLOW='',1P,E14.7, '',  NU='',I11)') XLOW,NU
!!           IER=4
!!           RETURN
!!         ENDIF
!!         XT(1)=XLOW
!!         IF(XLOW.LT.1.0D-16) THEN
!!           XT(2)=XLOW+1.0D-6*(XHIGH-XLOW)
!!           I1=2
!!         ELSE
!!           I1=1
!!         ENDIF
!!         FX=EXP(LOG(XHIGH/XT(I1))/DBLE(NP-I1))
!!         DO I=I1,NP-1
!!           XT(I+1)=XT(I)*FX
!!         ENDDO
!!         XT(NP)=XHIGH
!!       ENDIF
!! C
!!       DO I=1,NP-1
!!         DX=XT(I+1)-XT(I)
!!         DXI=DX/DBLE(NIP-1)
!!         PDFMAX=0.0D0
!!         DO K=1,NIP
!!           XS(K)=XT(I)+DBLE(K-1)*DXI
!!           YS(K)=MAX(PDF(XS(K)),ZEROT)
!!           PDFMAX=MAX(PDFMAX,YS(K))
!!         ENDDO
!! C  ****  Simpson's integration.
!!         CONS=DXI*3.3333333333333333D-1*0.5D0
!!         SUMI(1)=0.0D0
!!         DO K=2,NIP
!!           XIH=XS(K)-0.5D0*DXI
!!           YSH=MAX(PDF(XIH),ZEROT)
!!           PDFMAX=MAX(PDFMAX,YSH)
!!           SUMI(K)=SUMI(K-1)+CONS*(YS(K-1)+4.0D0*YSH+YS(K))
!!         ENDDO
!! C
!!         DPAC(I)=SUMI(NIP)
!!         FACT=1.0D0/DPAC(I)
!!         DO K=1,NIP
!!           SUMI(K)=FACT*SUMI(K)
!!         ENDDO
!! C  ****  When the PDF vanishes at one of the interval end points, its
!! C        value is modified.
!!         IF(YS(1).LT.ZERO) YS(1)=1.0D-5*PDFMAX
!!         IF(YS(NIP).LT.ZERO) YS(NIP)=1.0D-5*PDFMAX
!! C
!!         PLI=YS(1)*FACT
!!         PUI=YS(NIP)*FACT
!!         B(I)=1.0D0-1.0D0/(PLI*PUI*DX*DX)
!!         A(I)=(1.0D0/(PLI*DX))-1.0D0-B(I)
!!         C(I)=1.0D0+A(I)+B(I)
!!         IF(C(I).LT.ZERO) THEN
!!           A(I)=0.0D0
!!           B(I)=0.0D0
!!           C(I)=1.0D0
!!         ENDIF
!! C
!! C  ****  ERR(I) is the integral of the absolute difference between the
!! C  rational interpolation and the true PDF, extended over the interval
!! C  (XT(I),XT(I+1)). Calculated using the trapezoidal rule.
!! C
!!         ICASE=1
!!   100   CONTINUE
!!         ERR(I)=0.0D0
!!         DO K=1,NIP
!!           RR=SUMI(K)
!!           PAP=DPAC(I)*(1.0D0+(A(I)+B(I)*RR)*RR)**2/
!!      1       ((1.0D0-B(I)*RR*RR)*C(I)*(XT(I+1)-XT(I)))
!!           IF(K.EQ.1.OR.K.EQ.NIP) THEN
!!             ERR(I)=ERR(I)+0.5D0*ABS(PAP-YS(K))
!!           ELSE
!!             ERR(I)=ERR(I)+ABS(PAP-YS(K))
!!           ENDIF
!!         ENDDO
!!         ERR(I)=ERR(I)*DXI
!! C  ****  If ERR(I) is too large, the PDF is approximated by a uniform
!! C        distribution.
!!         IF(ERR(I).GT.0.10D0*DPAC(I).AND.ICASE.EQ.1) THEN
!!           B(I)=0.0D0
!!           A(I)=0.0D0
!!           C(I)=1.0D0
!!           ICASE=2
!!           GO TO 100
!!         ENDIF
!!       ENDDO
!!       XT(NP)=XHIGH
!!       A(NP)=0.0D0
!!       B(NP)=0.0D0
!!       C(NP)=0.0D0
!!       ERR(NP)=0.0D0
!!       DPAC(NP)=0.0D0
!! C
!! C  ****  New grid points are added by halving the subinterval with the
!! C        largest absolute error.
!! C
!!   200 CONTINUE
!!       ERRM=0.0D0
!!       LMAX=1
!!       DO I=1,NP-1
!! C  ****  ERRM is the largest of the interval errors ERR(I).
!!         IF(ERR(I).GT.ERRM) THEN
!!           ERRM=ERR(I)
!!           LMAX=I
!!         ENDIF
!!       ENDDO
!! C
!!       NP=NP+1
!!       DO I=NP,LMAX+1,-1
!!         XT(I)=XT(I-1)
!!         A(I)=A(I-1)
!!         B(I)=B(I-1)
!!         C(I)=C(I-1)
!!         ERR(I)=ERR(I-1)
!!         DPAC(I)=DPAC(I-1)
!!       ENDDO
!!       XT(LMAX+1)=0.5D0*(XT(LMAX)+XT(LMAX+2))
!!       DO I=LMAX,LMAX+1
!!         DX=XT(I+1)-XT(I)
!!         DXI=(XT(I+1)-XT(I))/DBLE(NIP-1)
!!         PDFMAX=0.0D0
!!         DO K=1,NIP
!!           XS(K)=XT(I)+DBLE(K-1)*DXI
!!           YS(K)=MAX(PDF(XS(K)),ZEROT)
!!           PDFMAX=MAX(PDFMAX,YS(K))
!!         ENDDO
!! C  ****  Simpson's integration.
!!         CONS=DXI*3.3333333333333333D-1*0.5D0
!!         SUMI(1)=0.0D0
!!         DO K=2,NIP
!!           XIH=XS(K)-0.5D0*DXI
!!           YSH=MAX(PDF(XIH),ZEROT)
!!           SUMI(K)=SUMI(K-1)+CONS*(YS(K-1)+4.0D0*YSH+YS(K))
!!         ENDDO
!! C
!!         DPAC(I)=SUMI(NIP)
!!         FACT=1.0D0/DPAC(I)
!!         DO K=1,NIP
!!           SUMI(K)=FACT*SUMI(K)
!!         ENDDO
!! C
!!         IF(YS(1).LT.ZERO) YS(1)=1.0D-5*PDFMAX
!!         IF(YS(NIP).LT.ZERO) YS(NIP)=1.0D-5*PDFMAX
!!         PLI=YS(1)*FACT
!!         PUI=YS(NIP)*FACT
!!         B(I)=1.0D0-1.0D0/(PLI*PUI*DX*DX)
!!         A(I)=(1.0D0/(PLI*DX))-1.0D0-B(I)
!!         C(I)=1.0D0+A(I)+B(I)
!!         IF(C(I).LT.ZERO) THEN
!!           A(I)=0.0D0
!!           B(I)=0.0D0
!!           C(I)=1.0D0
!!         ENDIF
!! C
!!         ICASE=1
!!   300   CONTINUE
!!         ERR(I)=0.0D0
!!         DO K=1,NIP
!!           RR=SUMI(K)
!!           PAP=DPAC(I)*(1.0D0+(A(I)+B(I)*RR)*RR)**2/
!!      1       ((1.0D0-B(I)*RR*RR)*C(I)*(XT(I+1)-XT(I)))
!!           IF(K.EQ.1.OR.K.EQ.NIP) THEN
!!             ERR(I)=ERR(I)+0.5D0*ABS(PAP-YS(K))
!!           ELSE
!!             ERR(I)=ERR(I)+ABS(PAP-YS(K))
!!           ENDIF
!!         ENDDO
!!         ERR(I)=ERR(I)*DXI
!! C
!!         IF(ERR(I).GT.0.10D0*DPAC(I).AND.ICASE.EQ.1) THEN
!!           B(I)=0.0D0
!!           A(I)=0.0D0
!!           C(I)=1.0D0
!!           ICASE=2
!!           GO TO 300
!!         ENDIF
!!       ENDDO
!! C
!!       IF(NP.LT.N) GO TO 200
!!       NPM1=NP-1
!! C
!! C  ****  Renormalisation.
!! C
!!       CNORM=0.0D0
!!       DO I=1,NPM1
!!         CNORM=CNORM+DPAC(I)
!!       ENDDO
!!       CNORM=1.0D0/CNORM
!!       ERRM=0.0D0
!!       DO I=1,NPM1
!!         DPAC(I)=DPAC(I)*CNORM
!!         ERR(I)=ERR(I)*CNORM
!!         ERRM=MAX(ERRM,ERR(I))
!!       ENDDO
!! C
!!       PAC(1)=0.0D0
!!       DO I=1,NPM1
!!         PAC(I+1)=PAC(I)+DPAC(I)
!!       ENDDO
!!       PAC(NP)=1.0D0
!! C
!! C  ****  Pre-calculated limits for the initial binary search in
!! C        subroutine RITAI.
!! C
!!       BIN=1.0D0/DBLE(NPM1)
!!       IL(1)=1
!!       DO I=2,NPM1
!!         PTST=(I-1)*BIN
!!         DO J=IL(I-1),NP
!!           IF(PAC(J).GT.PTST) THEN
!!             IL(I)=J-1
!!             IU(I-1)=J
!!             GO TO 400
!!           ENDIF
!!         ENDDO
!!   400   CONTINUE
!!       ENDDO
!!       IU(NPM1)=NP
!!       IL(NP)=NP-1
!!       IU(NP)=NP
!! C
!! C  ****  Print interpolation tables (only when IWR.GT.0).
!! C
!!       IF(IWR.GT.0) THEN
!!         OPEN(IWR,FILE='param.dat')
!!         WRITE(IWR,1000) ERRM
!!  1000   FORMAT(1X,'#  Interpolation error =',1P,E15.7)
!!         WRITE(IWR,1001) CNORM
!!  1001   FORMAT(1X,'# Normalising constant =',1P,E15.7)
!!         WRITE(IWR,1002)
!!  1002   FORMAT(1X,'#',5X,'X',11X,'PDF(X)',10X,'A',13X,'B',13X,'C',
!!      1    11X,'error')
!!         DO I=1,NPM1
!!           PDFE=MAX(PDF(XT(I)),ZEROT)*CNORM
!!           WRITE(IWR,'(1P,7E14.6)') XT(I),PDFE,A(I),B(I),C(I),ERR(I)
!!         ENDDO
!!         CLOSE(IWR)
!! C
!!         OPEN(IWR,FILE='table.dat')
!!         WRITE(IWR,1000) ERRM
!!         WRITE(IWR,1001) CNORM
!!         WRITE(IWR,2000)
!!  2000   FORMAT(1X,'#',6X,'X',13X,'PDF_ex',10X,'PDF_ap',11X,'err')
!!         DO I=1,NPM1
!!           DX=(XT(I+1)-XT(I))/DBLE(NIP-1)
!!           DO K=1,NIP,5
!!             XTAU=XT(I)+(K-1)*DX
!!             P1=MAX(PDF(XTAU)*CNORM,ZEROT)
!! C  ****  Rational interpolation.
!!             TAU=(XTAU-XT(I))/(XT(I+1)-XT(I))
!!             CON1=2.0D0*B(I)*TAU
!!             CON2=C(I)-A(I)*TAU
!!             IF(ABS(CON1).GT.1.0D-10*ABS(CON2)) THEN
!!               ETA=CON2*(1.0D0-SQRT(1.0D0-2.0D0*TAU*CON1/CON2**2))/CON1
!!             ELSE
!!               ETA=TAU/CON2
!!             ENDIF
!!             P2=DPAC(I)*(1.0D0+(A(I)+B(I)*ETA)*ETA)**2
!!      1        /((1.0D0-B(I)*ETA*ETA)*C(I)*(XT(I+1)-XT(I)))
!!             WRITE(IWR,'(1P,5E16.8)') XTAU,P1,P2,(P1-P2)/P1
!!           ENDDO
!!         ENDDO
!!         CLOSE(IWR)
!! C
!!         OPEN(IWR,FILE='limits.dat')
!!         WRITE(IWR,3000)
!!  3000   FORMAT(1X,'#  I',6X,'PAC(ITL)',9X,'(I-1)/NPM1',11X,'I/NPM1',
!!      1    12X,'PAC(ITU)',12X,'PAC(I)')
!!         DO I=1,NPM1
!!           WRITE(IWR,'(I5,1P,5E19.11)') I,PAC(IL(I)),(I-1)*BIN,I*BIN,
!!      1      PAC(IU(I)),PAC(I)
!!           IF(PAC(IL(I)).GT.(I-1)*BIN+EPS.OR.
!!      1      PAC(IU(I)).LT.I*BIN-EPS) THEN
!!             WRITE(IWR,3001)
!!  3001       FORMAT(' #  WARNING: The first four values should be in in',
!!      1        'creasing order.')
!!           ENDIF
!!         ENDDO
!!         CLOSE(IWR)
!!       ENDIF
!! C
!!       RETURN
!!       END
!!>>>> penEasy ends

